      subroutine BICGSTAB(n, matrix, rhs, solution, aquracy)
      
c     Bicojugate Gradient Stabilized solver
c     Solves dense linear system : matrix * solution = rhs
c
c     where :
c     n         - the matrix & vectors dimensions
c     matrix    - the dense n x n matrix
c     rhs       - the right hand side vector
c     solution  - the vector linear system solution to place in
c                 (may contain an initial guess)
c     aquracy   - the solver aquracy
      
      implicit none
      
      integer :: n
      real(8) :: matrix(1:n, 1:n), rhs(1:n), solution(1:n), aquracy
      
c     Internal variables      
      real(8) :: residual(1:n)
      real(8) :: rtilde(1:n), p(1:n), v(1:n), s(1:n), t(1:n)
      real(8) :: alpha, beta, omega, rho_1, rho_2
      integer :: iterationsNumber
      
c     BODY

c     Setup the residual vector.
	residual = rhs - matmul(matrix, solution)

c     Setup rtilde, p and v.
      rtilde = residual
      p = residual
      v = residual

c     Initialize parameters.
	alpha = 0; omega = 0; rho_2 = 0; 

c     Repeat until converged with the specified accuracy.
      iterationsNumber = 0
	do while (maxval(rhs - matmul(matrix, solution)) .GE. aquracy)
	  print*, 'residual', maxval(rhs - matmul(matrix, solution))
	  iterationsNumber = iterationsNumber + 1

	  rho_1 = dot_product(rtilde, residual)

	  if (iterationsNumber .NE. 1) then

            beta = (rho_1 / rho_2) * (alpha / omega)
            
            p = residual + beta * (p - omega * v)
            
        endif
            
        v = matmul(matrix, p)

		alpha = rho_1 / dot_product(rtilde, v)

		s = residual - alpha * v

        t = matmul(matrix, s)

		omega = dot_product(t, s) / dot_product(t, t)

		solution = solution + alpha * p + omega * s
		residual = s - omega * t

		rho_2 = rho_1
		
      enddo

      end subroutine BICGSTAB