
      SUBROUTINE ASSIM_lst_sq
!     Specification of the initial profiles using least-square assimilation
      use assim_var
      implicit none
      
!      real(8) b_mod1(3,3),h_obs1(2,3),RpHBHT1(3,2)
      real(8) BHT(size_mod,size_obs),Cm1d(size_obs),accuracy,
     & increment(1:size_mod),
     & RpHBHT_Tdep_obs(1:size_obs),
     & RpHBHT_TRpHBHT(1:size_obs,1:size_obs)
      real(8) eigenvalues(1:size_mod),
     &        eigenvectors(1:size_mod,1:size_mod),
     &        eigenvalues1(1:size_obs),
     &        eigenvectors1(1:size_obs,1:size_obs)
      real(8) work1(1:size_obs,1:size_obs),
     &        work(1:size_mod,1:size_mod) 
      integer i,j,ierr
      
!      data b_mod1(1,1:3) /1.,2.,3./
!      data b_mod1(2,1:3) /2.,2.,2./
!      data b_mod1(3,1:3) /3.,3.,3./
      
!      data h_obs1(1,1:3) /1.,2.,3./
!      data h_obs1(2,1:3) /1.,1.,1./
      
      open (701,file='results/B_matrix.dat',status='old',err=7010)
      open (702,file='results/R_matrix.dat',status='old',err=7020)
      
      call readgrd_lake(701,b_mod,1,size_mod,1,size_mod)
      call readgrd_lake(702,r_obs,1,size_obs,1,size_obs)
      
      call obs_oper
           
      BHT = matmul(b_mod,transpose(h_obs))
      RpHBHT = matmul(h_obs,BHT)
           
      RpHBHT = RpHBHT + r_obs
      RpHBHT_TRpHBHT = matmul(transpose(RpHBHT),RpHBHT)
      
      do i=1,size_obs
       dep_obs(1,i) = y_obs(1,i)-sum(h_obs(i,:)*x_mod(1,:))
      enddo
      
      RpHBHT_Tdep_obs = matmul(transpose(RpHBHT),
     & dep_obs(1,1:size_obs))
     
!      do i=1,size_mod
!       do j=1,size_mod
!        if (b_mod(i,j)/=b_mod(j,i)) then
!         print*, 'Non-symmetric'
!         print*, i,j,b_mod(i,j),b_mod(j,i)
!        endif
!       enddo
!      enddo
!      read*
     
!     Checking, if the r_obs and b_mod matrizes are positively defined
      call  AEH1d (size_mod, b_mod, eigenvalues, eigenvectors, 
     & work, ierr) 
     
!      do i = 1, size_mod
!       print*, 'Residual error of eigenvalues problem solving', i,
!     &  eigenvalues(i) 
!       print*, matmul(b_mod,eigenvectors(:,i))-
!     &  eigenvalues(i)*eigenvectors(:,i)
!       read*
!      enddo
     
      print*, 'Eigenvalues of b_mod', ierr
      print*, eigenvalues
      read*
      
!     BICONJUGATE GRADIENT METHOD is intended to be used here to solve
!     the set of equations with
!     r.h.s.:  dep_obs (1,1:size_obs) (stored in the module assim_var)
!     Matrix:  RpHBHT  (1:size_obs,1:size_obs) (stored in the module assim_var)
 
      Cm1d = 0.
      accuracy = 1.d-4
      call BICGSTAB(
     &  size_obs,                ! the linear system dimension (n)
     &  RpHBHT,                  ! the dense n x n matrix
     &  dep_obs(1,1:size_obs),   ! the 1 x n right hand side vector
     &  Cm1d,                    ! the 1 x n solution vector
     &  accuracy)                ! the method aquracy
      
      print *, 'BICGSTAB solution is :'
      print *, Cm1d   
      read *
      
      print*, 'residual'
      print*, matmul(RpHBHT,Cm1d)-dep_obs(1,:)
      read*
      
!      work = RpHBHT
      print*, 'RpHBHT'
      print*, maxval(RpHBHT)
      read*
      
      print*, 'dep_obs'
      print*, maxval(dep_obs)
      read*
      
      work1 = RpHBHT
!      call ASG0D(work,dep_obs(1,1:size_obs),Cm1d,size_obs,1)
      call ASG0D(work1,dep_obs(1,1:size_obs),Cm1d,size_obs,1)
      
      print*, 'residual of precise method'
      print*, matmul(RpHBHT,Cm1d)-dep_obs(1,:)
      read*
      
      print *, 'Eulerian elimination method solution is :'
      print *, Cm1d   
      read *
            
      increment = matmul(BHT,Cm1d)
      
      print*, 'increment=', increment
      
      print*, 'The least square assimilation is not operational:
     & STOP'
      STOP
      
      RETURN
      
7010  print*, 'Error reading file B_matrix.dat. The file may not exist'
      STOP
7020  print*, 'Error reading file R_matrix.dat. The file may not exist' 
      STOP
      
      END SUBROUTINE ASSIM_lst_sq
     
      SUBROUTINE HL_method
      use assim_var
      use driving_params
      
!     The program implements Hollingworth-Lonnberg (H-L) algorithm for
!     evaluation of background error covariance matrix and 
!     observation error covariance matrix for the case of 1D (physical) space.
!     It could be used for 1D assimilation purposes in the fields
!     of e.g. atmospheric radiation, soil, water bodies modeling etc.

!     Additional assumptions
!     1) Observations are performed at the points of the model grid;
!     2) In each gridpoint there is the same set of observed parameters; 
!        the number of observed parameters is n_obsvar;
!     3) In each gridpoint there is the same set of calculated parameters
!        the number of calculated parameters is n_modvar;
!     3) The observation operator is constant in time;

!     Moscow State University,
!     Scientific Research Computing Center.
!     Developed by Victor Stepanenko,
!     tel.: +7 (495) 939-23-53,
!     E-mail: stepanen@srcc.msu.ru

      implicit none
      real(8) pi, ntr, Mr
      integer i, j, k, L, n, ii, jj, im1, ip1
     
      pi = 4.*datan(1.d0)
     
      Mr = float(M)
      ntr = float(ntim)

      open(701,file='results/B_matrix.dat',status='unknown')
      open(702,file='results/R_matrix.dat',status='unknown')
                     
      call obs_oper()
      
!      do i = 1, size_obs
!       write(*,'(25f3.0)') h_obs(i,1:25)
!      enddo 
      
!     Generation of the time series for observed and modeled fields
!      do i = 1, ntim
!       do j = 1, M+1
!        ir = float(i); jr = float(j)
!     The case, when all the parameters coincide        
!        do k = 1, n_obsvar
!         ii = (j-1)*n_obsvar + k
!         kr = 0. !float(k)
!         y_obs(i,ii) = dsin(10*pi*ir/ntr*(1+kr/10.) )*
!     &    (1.-jr/(Mr+1.))
!        enddo
!        do k = 1, n_modvar
!         jj = (j-1)*n_modvar + k
!         kr = 0. !float(k)
!         x_mod(i,jj) = dsin(10*pi*ir/ntr*(1+kr/10.)+pi/5.)*
!     &    (1.-jr/(Mr+1.))
!        enddo
!       enddo
!      enddo      
      
!      print*, 'The HL_method is working in debug mode'
!      do k = 1, ntim
!       do i = 1, M+1
!        j = (i-1)*n_modvar
!        x_mod(k,j+1)= y_obs(k,i)+0.1
!        x_mod(k,j+2)= 0.
!       enddo 
!      enddo

!     Calculating mean profile, measured and calculated
      y_obs_mean = 0.
      x_mod_mean = 0.
      do i = 1, ntim
       y_obs_mean(:) = y_obs_mean(:) + y_obs(i,:)
       x_mod_mean(:) = x_mod_mean(:) + x_mod(i,:)
      enddo 
      y_obs_mean = y_obs_mean/ntr
      x_mod_mean = x_mod_mean/ntr
      
      print*, 'y_obs_mean', y_obs_mean
      print*, 'x_mod_mean', x_mod_mean
      print*, 'delta is'  , y_obs_mean - matmul(h_obs,x_mod_mean)
      print*, 'Attention: the shift between measurements and the model
     & now will be substracted from the vector of measurements'
      read*
            
      delta = y_obs_mean - matmul(h_obs,x_mod_mean)
      do i = 1, ntim
       y_obs(i,:) = y_obs(i,:) - delta(:)
      enddo
      
!      y_obs_mean = 0.
!      do i = 1, ntim
!       y_obs_mean(:) = y_obs_mean(:) + y_obs(i,:)
!      enddo 
!      y_obs_mean = y_obs_mean/ntr
            
!      print*, 'delta2 is'  , y_obs_mean - matmul(h_obs,x_mod_mean)
!      read*
      
!     Calculating covar(i,j,:,:) 
      covar = 0.
      do i = 1, M+1          
       do j = 1, M+1
        do k = 1, n_obsvar
         do L = 1, n_obsvar
          ii = (i-1)*n_obsvar + k
          jj = (j-1)*n_obsvar + L
          do n = 1, ntim
            covar(i,j,k,L) = covar(i,j,k,L) + 
     &       (y_obs(n,ii)-dot_product(h_obs(ii,:),x_mod(n,:)) )*
     &       (y_obs(n,jj)-dot_product(h_obs(jj,:),x_mod(n,:)) )
          enddo
         enddo
        enddo 
       enddo
      enddo 
      covar = covar/ntr
      
      do i = 1, M+1
       do j = 1, M+1
        if (j/=i) then
         k = (i-1)*n_modvar
         L = (j-1)*n_modvar
         b_mod(k+1:k+n_obsvar,L+1:L+n_obsvar) = 
     &    covar(i,j,1:n_obsvar,1:n_obsvar)
        endif
       enddo
      enddo
      
      do i=2,M
       im1 = i-1
       ip1 = i+1
       k = (i-1)*n_modvar
       b_mod(k+1:k+n_obsvar,k+1:k+n_obsvar)=
     &  (covar(i,im1,1:n_obsvar,1:n_obsvar)+
     &  covar(i,ip1,1:n_obsvar,1:n_obsvar))/2.
      enddo 
      
      i = 1
      k = 0
      ip1 = 2
      b_mod(k+1:k+n_obsvar,k+1:k+n_obsvar)=
     & covar(i,ip1,1:n_obsvar,1:n_obsvar)
     
      i = M+1
      k = (i-1)*n_modvar
      im1 = M
      b_mod(k+1:k+n_obsvar,k+1:k+n_obsvar)=
     & covar(i,im1,1:n_obsvar,1:n_obsvar)
     
      do i = 1, M+1
       k = (i-1)*n_modvar
       L = (i-1)*n_obsvar
       r_obs(L+1:L+n_obsvar,L+1:L+n_obsvar) =
     &  covar(i,i,1:n_obsvar,1:n_obsvar)-
     &  b_mod(k+1:k+n_obsvar,k+1:k+n_obsvar)
      enddo 
     
      print*, 'Matrix of covariances of model errors is calculated'
!      do i=1,size_mod
!       print*, (b_mod(i,j), j=1,size_mod)
!       read*
!      enddo
      call wrigrd_lake(701,b_mod,1.d0,1.d0,1,size_mod,1,size_mod)
      
      print*, 'Matrix of covariances of observational errors 
     & is calculated'
      call wrigrd_lake(702,r_obs,1.d0,1.d0,1,size_obs,1,size_obs)
            
      close (701)
      close (702)

      END SUBROUTINE HL_method
            
      SUBROUTINE obs_oper()
      use assim_var
      use driving_params
!     Specifies the constant in time observation operator 
!     The observations are assumed to be taken in the points
!     of model grid, but the number of observed parameters is
!     less or equal to number of model parameters: n_obsvar<=n_modvar
      implicit none
      integer i, j, k, n
      
      h_obs = 0.     
      do k = 1, M+1
       do n = 1, n_obsvar
        i = (k-1)*n_obsvar+n
        j = (k-1)*n_modvar+n
        h_obs(i,j) = 1.
       enddo  
      enddo 
                 
      END SUBROUTINE obs_oper

 
      SUBROUTINE measured_data_reader
      use driving_params
      use assim_var
      implicit none
!      Variables
      real(8) Tsurf_meas,Tbot_meas,h_ML_meas,z_depth
      real(8) h1,l1,hs1,ls1
      real(8), external:: dzeta
      integer i,k,j,ii
      character*10, buf(15)
            
      common /layers/ h1,l1,hs1,ls1
     
      open (700,file='data/data_Vend_jl03_for_LAKE.dat',status='old')

      do k=1,2

!     Reading the header of the data file          
      do i=1,20
       read (700,'(a)')
      enddo
           
      i=0
      rddata: do while (.not. eof(700))
       read(700,*) (buf(ii),ii=1,13)
       read (buf(10),'(f)') Tsurf_meas
       read (buf(11),'(f)') h_ML_meas
       read (buf(12),'(f)') Tbot_meas
       if (Tsurf_meas==-9999.or.h_ML_meas==-9999.or.Tbot_meas==-9999)
     &   exit rddata
       i=i+1
       if (allocated(y_obs)) then
        do j = 1, M+1
         z_depth = dzeta(dble(j))*h1
         continue
         if (z_depth<=h_ML_meas) then
!        Only temperature is measured in the lake
          y_obs(i,j) = Tsurf_meas
         else
          y_obs(i,j) = Tsurf_meas+(z_depth-h_ML_meas)*
     &     (Tbot_meas-Tsurf_meas)/(h1-h_ML_meas) 
         endif 
        enddo       
       endif 
      enddo rddata
      
      if (.not.allocated(y_obs)) then
       ntim = i
       call alloc_assim_var
       rewind(700)
      endif 
     
      enddo
     
      return
    
      END SUBROUTINE measured_data_reader

