!-----ONE-DIMENSIONAL LAKE MODEL Lake1.0beta
!-----Moscow State University, Scientific Research Computing Center
!-----Moscow, 119992, GSP-2, Vorobievy Gory, SRCC MSU	
!-----Victor M. Stepanenko, stepanen@srcc.msu.ru
!-----Release: May 2007
	
	PROGRAM CALL_LAKE
!     This program calls the LAKE model.
!     It is needed only in stand-alone runs; in interactive regime with
!     atmospheric model it must be cut from the code.
      use arrays
      use driving_params
      implicit none 
      integer(4) ix,iy,nx,ny,i_year
      real(8) ta,qa,pa,ua,va,atm_rad,ti,tsurf,wind,
     & solar_rad,precip,zref,ts,sens_flux,latent_flux,exch_coef,dt
      real(8) t_Flake,t_Lake,t1,t2
      real(8) h1,l1,hs1,ls1
      real(8) year,month,day,hour
      logical flag_assim
      common /layers/ h1,l1,hs1,ls1
      data ti,i_year,t_Lake,t_Flake,tsurf/0.,0,0.,0.,293./
      data flag_assim /.false./

!     NAMELIST /atmlist/ ta,qa,pa,ua,va,atm_rad,solar_rad,precip
!     NAMELIST /fluxes/  sens_flux,latent_flux,exch_coef
!     NAMELIST /fluxesF/ sens_fluxF,latent_fluxF,exch_coefF                

      call DEFINE_driving_params
      
      dt = dtdr
! Transforming tinteg units from days to seconds
      tinteg = tinteg*24.*60.*60.
      zref = height_T_q

      ix = 1
      iy = 1
      nx = 1
      ny = 1
      call init_lake(nx,ny,'sam_map')
      
      do while (ti<tinteg+spinup_day*24.*60*60)
       ti=ti+dt
!       if(tsurf==0) tsurf=293.
       call METEODATA(ti,ta,qa,pa,ua,va,atm_rad,solar_rad,
     &  precip,tsurf,year,month,day,hour,dt)
     
!_var_in: ta - air temperature at z=zref, K  
!_var_in: qa - specific humidity at z=zref, kg/kg
!_var_in: pa - atmospheric pressure, Pa
!_var_in: ua - wind speed at z=zref, m/s
!_var_in: atm_rad - downward atmospheric radiation, W/m**2
!_var_in: solar_rad - net solar radiation, W/m**2
!_var_in: precip - precipitation rate in terms of liquid water, m/sec
!_var_in: zref - first atmospheric model level, or level of measurements, m
!_var_out: ts - surface temperature, K
!_var_out: sens_flux - sensible heat flux, W/m**2
!_var_out: latent_flux - latent heat flux, W/m**2
!_var_out: exch_coef - exchange coefficient (tau=ro*exch_coef*ua) W/m**2
!_var_in: dt - time step, s
!_var_in: ix - x-number of current model grid point at the surface
!_var_in: iy - y-number of current model grid point at the surface
!_var_in: nx - number of grid points in x direction 
!_var_in: ny - number of grid points in y direction 
! CPU time counter

       call cpu_time(t1)
       
!      If the spinup IS NOT USED, 
!      data assimilation implemented once at the first timestep
       if (assim==1.and.spinup_day==0.and.ti==dt) flag_assim=.true.
!      If the spinup IS USED, 
!      data assimilation is implemented once at the end of spinup period
       if (assim==1.and.spinup_day/=0.and.
     &  ti-spinup_day*24.*60*60<=dt.and.ti-spinup_day*24.*60*60>0) 
     &  flag_assim=.true.
     
       call LAKE(ta,qa,pa,ua,va,atm_rad,
     &  solar_rad,precip,zref,ts,sens_flux,latent_flux,exch_coef,
     &  trib_in,dt,ix,iy,nx,ny,year,month,day,hour,flag_assim)
       call cpu_time(t2)
       t_Lake=t_Lake+t2-t1

       wind=dsqrt(ua**2+va**2)
       if (flag_assim) flag_assim=.false.
! CPU time counter
!      call cpu_time(t1)
!      call Flake1(ta,qa,pa,wind,atm_rad,&
!       solar_rad,precip,zref,tsF,sens_fluxF,latent_fluxF,exch_coefF,dt,ix,iy,nx,ny)
!      call cpu_time(t2)
!      t_Flake=t_Flake+t2-t1

!      tsurf=(ts+tsF)/2.
!      tsurf=tsF
       tsurf=ts

       if (int(ti/(60*60*24*30*12))>i_year) then
        print*, i_year+1, 'years left'
        i_year=i_year+1
       endif
      enddo

      print*, 'time_Lake =', t_Lake, 'sec  ', 
     & 'time_Flake =', t_Flake, 'sec  '

700   format (f10.1, 3f9.2)
770   format (f6.1, 4f9.1, 2e10.2)
  
      END PROGRAM CALL_LAKE


      SUBROUTINE METEODATA(ti,ta,qa,pa,ua,va,atm_rad,solar_rad,
     & precip,tsurf,year,month,day,hour,dt)
!     Subroutine METEODATA reads the file with data on atmospheric variables
!     in surface air layer. 
!     It is needed only in stand-alone runs; in interactive regime with
!     atmospheric model it must be cut from the code.      
      use driving_params
      implicit none
      real(8) ti,ta,qa,pa,ua,va,atm_rad,solar_rad,precip,tsurf,
     & Rad_bal,cloud,dt
      real(8) taf,qaf,paf,uaf,vaf,atm_radf,solar_radf,precipf,
     & Rad_balf,cloudf
      real(8) ta_old,qa_old,pa_old,ua_old,va_old,atm_rad_old,
     & solar_rad_old,xx,Rad_bal_old,precip_old,
     & cloud_old
      real(8) z0_wind,year,month,day,hour,sigma,tim,hour0
      integer(4) ti_int,ti_int_old,ispin,ti_old
      character*60 datafile,outpath
      logical firstcallM1
      data firstcallM1, ti_int_old, ispin, tim /.true.,-1,1,0./

      common /cloud/ cloud
      common /out/   outpath

      SAVE

      z0_wind = 0.001
      sigma = 5.67e-8

      if (firstcallM1) then
       print*, 'The lake is  ', dataname
       datafile = dataname(1:len_trim(dataname))//'.dat'
       datafile = 'data/'//datafile
       outpath = 'results/'//dataname
       outpath = outpath(1:len_trim(outpath))//'/'
       open (unit=2233,file=datafile(1:len_trim(datafile)), 
     &  status='old')

       if (form==1) then
!     READ (UNIT=2233,NML=old1)
        READ(2233,'(2f7.1,e14.3,3f10.3,f10.1,e12.3)') 
     &   solar_rad_old,xx,precip_old,ta_old, 
     &   ua_old,va_old,pa_old,qa_old 
       elseif (form==2) then
!     READ (UNIT=2233,NML=old2)
        READ(2233,'(4f7.1,2f7.1,e14.3,3f10.3,f10.1,e12.3)') 
     &   year,month,day,hour0,solar_rad_old,xx,precip_old,ta_old, 
     &   ua_old,va_old,pa_old,qa_old
       elseif (form==3) then
        READ(2233,'(4f7.1,2f7.1,e14.3,3f10.3,f10.1,e12.3,f5.1)') 
     &   year,month,day,hour0,solar_rad_old,xx,precip_old,ta_old, 
     &   ua_old,va_old,pa_old,qa_old,cloud_old
       endif
       if (rad==1) then
        atm_rad_old = xx
       elseif (rad==2) then
        Rad_bal_old = xx
       endif
       taf = ta_old 
       paf = pa_old 
       qaf = qa_old 
       uaf = ua_old 
       vaf = va_old 
       solar_radf = solar_rad_old
       atm_radf = atm_rad_old
       Rad_balf = Rad_bal_old 
       precipf = precip_old
       firstcallM1=.false.
      endif

      if (int4(ti/(24*60*60))>=ispin.and.ispin<=spinup_day) then
       ispin=ispin+1
       rewind(2233) 
      endif  
      
      tim = tim + dt
      hour = hour0 + tim/(60*60)

      ti_int = int4(ti/(60*60*interval))
      if (ti_int - ti_int_old /= 0) then
       ti_int_old    = ti_int
       ti_old        = ti_int_old*60*60*interval
       ta_old        = taf
       pa_old        = paf
       qa_old        = qaf
       ua_old        = uaf
       va_old        = vaf
       solar_rad_old = solar_radf
       atm_rad_old   = atm_radf 
       Rad_bal_old   = Rad_balf
       precip_old    = precipf
       if (form==1) then
!       READ (UNIT=2233,NML=new1)
        READ(2233,'(2f7.1,e14.3,3f10.3,f10.1,e12.3)') 
     &    solar_radf,xx,precipf,taf,uaf,vaf,paf,qaf 
       elseif (form==2) then
!       READ (UNIT=2233,NML=new2)
        READ(2233,'(4f7.1,2f7.1,e14.3,3f10.3,f10.1,e12.3)') 
     &   year,month,day,hour0,solar_radf,xx,precipf,
     &   taf,uaf,vaf,paf,qaf
       elseif (form==3) then
        READ(2233,'(4f7.1,2f7.1,e14.3,3f10.3,f10.1,e12.3,f5.1)') 
     &   year,month,day,hour0,solar_radf,xx,precipf,
     &   taf,uaf,vaf,paf,qaf,cloudf
       endif
!     READ(2233,'(2f7.1,e14.3,3f10.3,f10.1,e12.3)') &
!     solar_radf,xx,precipf,taf,uaf,vaf,paf,qaf
       if (rad==1) then
        atm_radf = xx
       elseif (rad==2) then
        Rad_balf = xx
       endif
       tim = 0.
      end if

      ta = ta_old+(ti - ti_old)*(taf-ta_old)/(60*60*interval)
      pa = pa_old+(ti - ti_old)*(paf-pa_old)/(60*60*interval)
      qa = qa_old+(ti - ti_old)*(qaf-qa_old)/(60*60*interval)
      ua = ua_old+(ti - ti_old)*(uaf-ua_old)/(60*60*interval)
      va = va_old+(ti - ti_old)*(vaf-va_old)/(60*60*interval)
      solar_rad = solar_rad_old+(ti - ti_old)*
     &       (solar_radf-solar_rad_old)/(60*60*interval)
      atm_rad   = atm_rad_old+  (ti - ti_old)*
     &       (atm_radf-atm_rad_old)    /(60*60*interval)
      Rad_bal   = Rad_bal_old+  (ti - ti_old)*
     &       (Rad_balf-Rad_bal_old)    /(60*60*interval)
!     cloud     = cloud_old+    (ti - ti_old)*&
!            (cloudf-cloud_old)        /(60*60*interval)
!     Converting precipitation rate from mm/min to m/sec
!     Be careful about units of precipitation rate!
      precip = precipf/1000./60. !               /(60*60*interval)
!     precip = precipf/1000./60./60./3.

      if (height_T_q /= height_u) then
       ua=ua*dLOG(height_T_q/Z0_WIND)/dLOG(height_u/Z0_WIND)
      endif

      if (rad==2) then
!      For water surface
       atm_rad = Rad_bal - solar_rad*0.93 + 0.95*sigma*tsurf**4.
      endif

      END SUBROUTINE METEODATA 


      SUBROUTINE INIT_LAKE(nx,ny,fnmap)
!      (tempair1,humair1,pressure1,uwind1,vwind1,
!     &longwave1,sw,prec,zref1,tsw,hw1,xlew1,cdmw1,trib_inflow1,
!     &dtl,ix,iy,nx,ny)
     
!     ENTRY SUBROUTINE OF THE MODEL 

      USE phys_parameters
	USE phys_constants 
	USE numeric_params
	USE phys_constants2
	use driving_params
	use atmos
	use arrays
	implicit none
	
      real(8) x1
	integer(4) i,j,nx,ny
	logical uniform_depth
	character, external:: fext2*80
	character fnmap*80, path2*80 
	data uniform_depth /.true./

	call DEFINE_phys_parameters
	call DEFINE_phys_constants
	call DEFINE_phys_constants2
	call DEFINE_driving_params
	call allocarrays(nx,ny)
	call pbldat
	call comsoilforlake
	
	if (error_cov==1) then
	 if (runmode==2) then
	  print*, 'The error covariance estimation algorithm works
     & only in standalone runs: STOP'
        STOP
	 endif
	 if (spinup_day/=0) then
	  print*, 'No spinup is allowed when calculating 
     & error covariances: STOP'
        STOP
       endif 
       if (assim == 1) then
        print*, 'assim = 1 and error_cov = 1: these regimes could not
     & be turned on simultaneously: STOP'
        STOP
       endif
       if (nx>1.or.ny>1) then
        print*, 'Error covariance calculation is currently adjusted 
     & only for one-point stand-alone runs of the lake model: STOP'
        STOP
       endif
	endif
	if (assim==1.and.(nx>1.or.ny>1)) then
	 print*, 'Data assimilation is currently adjusted 
     & only for one-point stand-alone lake model: STOP'
       STOP
	endif
	
	if (dt_out<dtdr/(60*60)) then
       print*, 'dt_out must be larger or equal to timestep: STOP'
       STOP
      endif
		 	 
	if (runmode == 2.and.uniform_depth) then
	 path2 = fext2(fnmap,'dep')
	 open (999, file=path(1:len_trim(path))//
     & path2(1:len_trim(path2)), status='old')
	 call readgrd_lake (999,dep_2d,0,nx-1,0,ny-1)
	 close (999)
	 dep_av=0.;x1=0.
	 do i=1,nx-1
	  do j=1,ny-1
         if (dep_2d(i,j)>0.) then
	   dep_av=dep_av+dep_2d(i,j)
	    x1=x1+1.
         endif
	  enddo
       enddo
	 dep_av=dep_av/x1
	endif

	print*, 'Lake model is initialized'

	END SUBROUTINE INIT_LAKE


      SUBROUTINE LAKE 
     &(tempair1,humair1,pressure1,uwind1,vwind1,
     &longwave1,sw,prec,zref1,tsw,hw1,xlew1,cdmw1,trib_inflow1,
     &dtl,ix,iy,nx,ny,year,month,day,hour,flag_assim)
      
!     MAIN SUBROUTINE OF THE MODEL 
!     Calculates all the parameters of state of a lake (temperature, currents,
!     eddy diffusivity coefficients, snow, ice, soil characteristics, etc.) 
!     at the next timestep (i.e. implements transition of variables 
!     from i-th time step to (i+1)-th )      

!     In atmospheric model must be called once each timestep,
!     or once each N timesteps, where N = dt_lake/dt_atmos 
!     (dt_lake - timestep in the lake model, dt_atmos - timestep the atmospheric model) 

!     INPUT VARIABLES:
!     tempair1      --- air temperature, K;
!     humair1       --- specific humidity of air, kg/kg;
!     pressure1     --- air pressure, Pa;
!     uwind1        --- x-component of wind, m/s;
!     vwind1        --- y-component of wind, m/s;
!     longwave1     --- longwave downward radiation, W/m**2;
!     sw            --- net solar radiation, W/m**2;
!     prec          --- precipitation intensity, m/s;
!     xref1         --- the height (m) of level in atmosphere, where temperature, 
!                       humidity and wind are measured (calculated by atmospheric model);
!     trib_inflow1  --- tributary inflow rate, m/s;
!     dtl           --- timestep (for stability must be not larger than 10-15 min), s;
!     ix            --- current number of grid point in X-direction;
!     iy            --- current number of grid point in Y-direction;
!     nx            --- total number of grid points in X-direction of atmospheric model;
!     ny            --- total number of grid points in Y-direction of atmospheric model;

!     OUTPUT VARIABLES:   
!     tsw           --- surface temperature of a lake, K;
!     hw1           --- sensible heat flux from lake, W/m**2; 
!     xlew1         --- latent heat flux from lake, W/m**2; 
!     cdmw          --- exchange coefficient, m/s ( == wind_speed*exchange_nondimensional_coeficient)
      
      use numeric_params    
	use driving_params
	use atmos
	use phys_constants2
	use arrays
	use assim_var
	implicit none


!------------------------ MAIN VARIABLES ------------------------
!  arrays:    Tw1 and Tw2 - temperature of water, C
!             Ti1 and Ti2 - temperature of ice, C
!             T - temperature of snow, C
!  functions of time:h1 and h2 - thickness of water, m
!             l1 and l2 - thickness of ice, m
!             hs1 - thickness of snow, m                    
!             flag_ice_surf - shows if ice is melting on the top surface, n/d        
!  constants: cw - specific heat of water, J/(kg*K) 
!             ci - specific heat of ice, J/(kg*K)
!             row0 - density of water, kg/m**3
!             roi - density of ice, kg/m**3
!             lamw - thermal conductivity of water, J*sec/(m**3*K)
!             lami - thermal conductivity of ice, J*sec/(m**3*K)
!             L - specific heat of melting, J/kg
!             Lwv - specific heat of evaporation, J/kg
!             Liv - specific heat of sublimation, J/kg    
!             ddz - size of grid element (space), m
!             dt - size of grid element (time), sec
!             kstratwater - coefficient for linear initial conditions in water
!             kstratice - coefficient for linear initial conditions in ice     	     
!  boundary conditions:
!             eFlux - total heat flux on the upper surface,
!      		  shortwave(1-A)-T**4+longwave-H-LE, J/(m**2*sec)
!             Elatent - latent heat flux, J/(m**2*sec) 
!             Erad - shortwave radiation, penetrated below a surface, J/(m**2*sec)
!             tempair - air temperature (2 m), C
!             precip - precipitation, m/sec
!            (M) number of layers in water and snow 	
	            
      real(8) l2,h2,h1,dhwhigh,dhwlow,dhwp,dhwe,
     & dhihigh,dhilow,dhip,
     & totalevap,dhis,totalprecip,totalmelt,
     & totalpen,totalwat,dhif,
     & ls1, ls2, snowmass, snowmass_init,
     & totalerad,totalhflux,totalpen1,
     & Tf_old1,pi,
     & year_sec,h_ML0zv,
     & wl_max,unfrwat,wi_max,dhwls,year,month,day,hour,
     & nyear,nmonth,nday,nhour	 
	real(8), dimension(1:350)::a,b,c,d 
	
	real(8) lams(ms) 
	real(8) dz
      real(8) ST,PGR,TGROLD,QGROLD,RADIAT,PRECIP,WSOLD,SNOLD,
     & ZS,thsoil,whsoil,bOLD,RF1,RF2,SNMELT,HSNOW,
     & HS,ES,TGRNEW,QGRNEW,WSNEW,SNNEW,RUNOFF,
     & ElatOld,HFold,PRSold,shortwave,extinct
      real(8) AL,DLT,DVT,ALLL,DL,
     & ALV,DV,Z,T,WL,WV,WI,dens
      real(8) dt,l1,hs1
      real(8) q(110)
	real(8) roughness,emissivity,albedo,aM,bM,relhums
      real(8), dimension(1:350):: Temp
     	real(8) sw,prec,dtl
!	real(8) x1,x2,x3,x4,x5
	
	real(8), external:: meltpnt,dzeta
	
	real(8), intent(in):: tempair1,humair1,pressure1,uwind1,vwind1,
     & longwave1,zref1,trib_inflow1
      real(8), intent(out):: hw1,xlew1,cdmw1,tsw

      integer(4) i,j,i_ML,
     & flag_snow,itop,mon, 
     & nmonth_old, nhour_old,iter,
     & xday,xhour,ix,iy,nx,ny,nstep_meas,
     & n_1cm,n_5cm,layer_case
     
	     
      logical soil_ts_ext_iter,firstcall,uniform_depth,flag_assim

      common /watericesnowarr/ lams,q
	common /SOILDAT/ dz(ms),itop
      common /bL/ ST,PGR,TGROLD,QGROLD,RADIAT,WSOLD,SNOLD,
     & ZS,thsoil,whsoil,bOLD,RF1,RF2,SNMELT,HSNOW,
     & HS,ES,TGRNEW,QGRNEW,WSNEW,SNNEW,RUNOFF,
     & ElatOld,HFold,PRSold,extinct(ms)
      common /SOILSOL/ AL(ML),DLT(ML),DVT(ML),ALLL(ML),DL(ML),
     & ALV(ML),DV(ML),Z(ML),T(ML),WL(ML),WV(ML),WI(ML),
     & dens(ms)
	common /ymdh/ nyear,nmonth,nday,nhour
	common /ts_ext_iter/ Tf_old1,soil_ts_ext_iter 
	common /layers/ h1,l1,hs1,ls1
	common /atmos2/ shortwave, precip
	common /surface/ roughness,emissivity,albedo,aM,bM,relhums
	
	data uniform_depth /.true./
	data nstep_meas /0/
	 
      SAVE
  
!BODY OF PROGRAM

	rosoil = 1200.	
!	extwat = 1. !0.7 !0.9 !0.524  
	extice = 2.
	pi = 4.*atan(1.)
	
	nyear = year; nmonth = month; nday = day; nhour = hour

	year_sec  = 60*60*24*365
		
	tempair = tempair1-273.2
	humair = humair1
	pressure = pressure1
	trib_inflow = trib_inflow1
	uwind = uwind1
	if (uwind==0) uwind=0.1
	vwind = vwind1
      if (vwind==0) vwind=0.1
	wind = dsqrt(uwind**2+vwind**2)
	longwave = longwave1
	zref = zref1
!     Control of the input atmosferic forcing
	if (tempair>60.or.tempair<-90) then
	 print*, 'The air temperature ', tempair, 'is illegal: STOP'
	 STOP
	elseif (dabs(humair)>1.) then
	 print*, 'The air humidity ', humair, 'is illegal: STOP'
	 STOP
	elseif (dabs(pressure)>110000.) then
	 print*, 'The air pressure ', pressure, 'is illegal: STOP'
	 STOP
	elseif (dabs(uwind)>200.) then
	 print*, 'The x-component of wind ', uwind, 'is illegal: STOP'
	 STOP
	elseif (dabs(vwind)>200.) then
       print*, 'The y-component of wind ', vwind, 'is illegal: STOP'
	 STOP
      elseif (dabs(longwave)>1000.) then
       print*, 'The longwave radiation ',longwave,'is illegal: STOP'
	 STOP
      elseif (dabs(sw)>1400.) then
       print*, 'The shortwave radiation ', sw, 'is illegal: STOP'
	 STOP
      elseif (dabs(prec)>1.) then
       print*, 'The atmospheric precipitation ',prec,
     & 'is illegal: STOP'
	 STOP
      endif     
	
	shortwave=sw
c converting of precipitation units: from kg/m**s/sec to m/sec
	precip=prec !*row

	Erad = shortwave*(1-albedo)
	dt=dtl
	      
	if (init(ix,iy)==0) then
!     Specification of initial profiles	
       h1 = h10
	 l1 = l10
	 hs1 = hs10
	 ls1 = ls10
       do i=1,M+1
        u1(i)=0.1*(1-float(i)/float(M+1))
        v1(i)=0.1*(1-float(i)/float(M+1))
       enddo 
       do i=1,M 
	  E1(i) = 1.d-5 + float(i)/float(M)*(1.d-8  -  1.d-5)
	  eps1(i)= 1.d-9 + float(i)/float(M)*(1.d-12  -  1.d-9)
	 enddo
	 do i=1,ns
	  if (i<0.7*ns) then 
         Tsoil1(i) = Tb0 + float(i-1)/(0.7*float(ns))*(Tbb0-Tb0)
        else
	   Tsoil1(i) = Tbb0
	  endif 
	  Sals1(i) = 1.d-8
        if (Tsoil1(i)>Meltpnt(Sals1(i))) then
	   wi1(i) = 0.
	   wl1(i) = wl_max(por(i),rosoil,0.d0)-0.01
	  else
	   wl1(i) = unfrwat(Tsoil1(i),i)
         wi1(i) = wi_max(por(i),rosoil)-wl1(i)-0.01
	  endif
       enddo
		 
	 i_ML = int4(M*h_ML0/h1)
       Sal1(1:max(i_ML,1)) = Sals0
	 do i=max(i_ML+1,2),M+1
	  Sal1(i) = Sals0 + (Salb0-Sals0)*float(i-i_ML)/float(M+1-i_ML)
	 enddo 
	 
	 if (init_T==2) then
	  h_ML0zv = (Tm-0.5*(Tb0+Ts0))/(Ts0/h1-0.5*(Ts0+Tb0)/h1)
	  i_ML = int4(M*h_ML0zv/h1)	 
	 endif 
	 Tw1(1:max(i_ML,1)) = Ts0
	 do i=max(i_ML+1,2),M+1
	  Tw1(i) = Ts0 + (Tb0-Ts0)*float(i-i_ML)/float(M+1-i_ML)
	 enddo 
	 Ti1 = 0.
	 Tis1 = 0. 

	 if (l1/=0) then
	  do i=1,M+1
         Ti1(i)=-5.+5.*float(i-1)/float(M)
	  enddo
	 endif
	  		
	 if (hs1==0.) then	 
	  flag_snow = 0
	  flag_snow_init = 1
       else
	  flag_snow = 1
	  flag_snow_init = 0
        hs1 = dmax1(2.d-2, hs1)
	  hs1 = int4(hs1/0.01)*0.01
	  n_5cm = int4(hs1/0.05)
	  n_1cm = int4((hs1-n_5cm*0.05)/0.01)
	  if (n_1cm==0) then
	   n_1cm=1; hs1=hs1+0.01
        endif
	  itop = ms - (n_1cm+n_5cm)
 	  do i=itop,itop+n_1cm-1
         dz(i) = 0.01
         T(i) = -5.
	   wl(i) = 0
	   dens(i) = 150.
        enddo
	  do i=itop+n_1cm,ms-1
	   dz(i) = 0.05
         T(i) = -5.
	   wl(i) = 0
	   dens(i) = 150.
        enddo
       endif
  
        cdmw2 = 1.d-3 

	  totalevap = 0 
	  totalmelt = 0 
	  totalprecip = 0. 
	  totalwat = 0
	  totalpen = 0
	  totalpen1 = 0
	  time = 0
	  dhwfsoil = 0.

	  par = 1
        mon = 1  
	  iter = 0 
	  nmonth_old = 1
	  nhour_old = 21
	  xday=1
        xhour=0
       if (runmode==2.and.uniform_depth==.false.) then
	  if (ix<=nx-1.and.iy<=ny-1) then
	   h1 = dep_2d(ix,iy)
        else
	   h1 = dep_av
	  endif  
	  if (h1<=0) then
	   print*, 'Negative or zero initial depth at ix=',ix,
     &   'iy=',iy,'h1=',h1,':terminating program'
         STOP 
        endif
	 endif
	 if (assim == 1.or.error_cov == 1) then
	  call measured_data_reader
	 endif
	else
!ALIGNING VALUES FROM PREVIOUS TIME STEP (time) IN CURRENT POINT (ix,iy)
       l1 = l1_2d(ix,iy) 
       h1 = h1_2d(ix,iy)
       ls1 = ls1_2d(ix,iy) 

	 do i=1,M+1
	  u1(i)=u_2d(i,ix,iy)
	  v1(i)=v_2d(i,ix,iy)
	 enddo

	 do i=1,M+1
        E1(i)=E_2d(i,ix,iy)
	  eps1(i)=eps_2d(i,ix,iy)
	 enddo
	  	            
       do i=1,ns 
        Tsoil1(i) = Tsoil1_2d(i,ix,iy)
        Sals1(i) = Sals1_2d(i,ix,iy)
        wi1(i)=wi1_2d(i,ix,iy)
	  wl1(i)=wl1_2d(i,ix,iy)
       enddo
		 	  
	 do i=1,M+1 
	  Tw1(i) = Tw1_2d(i,ix,iy)
	  Sal1(i) = Sal1_2d(i,ix,iy)
	  Ti1(i) = Ti1_2d(i,ix,iy)
	  Tis1(i) = Tis1_2d(i,ix,iy)
       enddo 
      			 
	 flag_snow = fl_sn_2d(ix,iy)
	 flag_snow_init = fl_sn_init_2d(ix,iy)

       itop = itop_2d(ix,iy)
	 hs1 = hs1_2d(ix,iy)

       do i=max(1,itop),ms-1
        dz(i) = dz_2d(i,ix,iy)
        T(i) = T_2d(i,ix,iy)
	  wl(i) = wl_2d(i,ix,iy)
	  dens(i) = dens_2d(i,ix,iy)
       enddo
  
        cdmw2 = cdm2(ix,iy)

	  totalevap = 0 
	  totalmelt = 0 
	  totalprecip = 0. 
	  totalwat = 0
	  totalpen = 0
	  totalpen1 = 0
	  
        time = time_2d(ix,iy)
	  dhwfsoil = dhwfsoil_2d(ix,iy)
	  Elatent = Elatent_2d(ix,iy)
	  dhw = dhw_2d(ix,iy)
	  dhw0 = dhw0_2d(ix,iy)
	  velfrict_prev = velfrict_2d(ix,iy)
 	endif
 	
 	if (flag_assim==.true.) then
!      Correcting model fields at the current timestep 
!      using least-square data assimilation	
       do i = 1, M+1
        j = (i-1)*n_modvar
        x_mod(1,j+1)= Tw1(i)
!        x_mod(1,j+2)= E1(i)
       enddo
       call assim_lst_sq 
	endif
         
      time = time + dt
      nstep = nstep + 1
      
      layer_case = 1
      if (h1 > 0  .and. l1 > 0) layer_case = 2 
      if (h1 ==0  .and. l1 > 0) layer_case = 3
      if (h1 ==0  .and. l1 ==0) layer_case = 4
      
      if (layer_case == 1.and.
     & Tw1(1)<Meltpnt(Sal1(1))-0.1.and.h1-0.01*roi/row0>0.01) 
     & layer_case = 2
     
      if (layer_case == 3.and.
     & Ti1(M+1)>Meltpnt(0.d0)+0.1.and.l1-0.01*row0/roi>0.01) 
     & layer_case = 2
      
	if (layer_case == 3.and.
     & Ti1(1)>Meltpnt(0.d0)+0.1.and.l1-0.01*row0/roi>0.01.
     & and.flag_snow==0) then
	 h1 = 0.01
	 ls1 = l1 - 0.01*row0/roi
	 Tis1 = Ti1
	 l2 = 0
	 Ti2 = 0
	 if (ls1<0.009999) print*, 'Thin layer! - ls'
	 layer_case = 1 
      endif
  
      if1: IF (layer_case == 1) THEN 
      
!---------------------------- CASE 1: LIQUID WATER LAYER AT THE TOP ("Summer")-----------------------

      if (Tw1(M+1)<Meltpnt(Sal1(M+1))-0.1.and.ls1 == 0.and.
     & h1-0.01*roi/row0>0.01) then
	 ls1 = 0.01
	 h1 = h1 - 0.01*roi/row0
	endif	
      
	do i=2,M
	 if (Tw1(i)<Meltpnt(Sal1(i))) Tw1(i)=Meltpnt(Sal1(i))
	enddo
	
      call KT_eq(dt)
	do i=1,M+1
	 lamw(i)=lamw0+KT(i) !+KC(i) !+KLengT(i)
	enddo
	lamsal = lamw/(cw*row0) * alsal
      
	do i=1,M
!       SR(i)=Erad*(1-sabs)*dexp(-extwat*(i-0.5)*ddz*h1)                         VS, 06.2007
       SR(i)=Erad*(1-sabs)*dexp(-extwat*dzeta(dble(i+0.5))*h1)                   VS, 06.2007
	enddo
	do i=2,M
	 RS(i)=-(SR(i)-SR(i-1))/(0.5*(ddz(i-1)+ddz(i))*h1)
      enddo
	if (ls1==0) then
	 ice=0;snow=0;water=1;deepice=0
       dhwp = precip*dt
	 dhwe = - Elatent/(Lwv*row0)*dt
	 dhw = dhwp + dhwe + dhwfsoil
	 dhw0 = dhwp + dhwe
	 dls = 0.
	else
	 ice=0;snow=0;water=1;deepice=1
	 dhwls = -lamw(M)*dt/(ddz(M)*h1*row0*Lwi)*(Tw1(M+1) - Tw1(M))+
     & lami*dt/(ddz(1)*ls1*row0*Lwi)*(Tis1(2) - Tis1(1))+
     & (1-albedoofice)*Erad*(1-sabs)*dexp(-extwat*h1)/(row0*Lwi)*dt 
       dls = -dhwls*row0/roi
	 dls0 = dls
	 dhwp = precip*dt
	 dhwe = - Elatent/(Lwv*row0)*dt
	 dhw = dhwp + dhwe + dhwfsoil + dhwls
	 dhw0 = dhwp + dhwe
	endif
	
	call soilforlake(dt, h1, l1, ls1, hs1,
     & a,b,c,d,Temp)
						
	h2 = h1 + dhw
	ls2 = ls1 + dls
	
      ENDIF if1               

      if2: IF (layer_case==2) THEN
      
!---------------------------CASE 2: WATER AND ICE ("Winter")------------------------------

       if (l1 == 0.) then
       l1 = 0.01 
	 h1 = h1-0.01*roi/row0
	 if (tempair<0.) then
	  do i=1, M+1
	   Ti1(i) = tempair - float(i-1)/float(M)*tempair
	  enddo 
       else 
	  Ti1=0.
       endif
	end if 
	
	do i=2,M
	 if (Tw1(i)<Meltpnt(Sal1(i))) Tw1(i)=Meltpnt(Sal1(i))+1.E-5
	enddo
		  
	if (h1 == 0.) then
	 h1 = 0.01
	 l1 = l1-0.01*row0/roi
	end if

	!if (h1<0.009999.or.l1<0.009999) print*, 'Thin layer!-l or h'
		  
	if (Tw1(M+1)<Meltpnt(Sal1(M+1))-0.1.and.ls1 == 0.and.
     &	h1-0.01*roi/row0>0.01) then
	 ls1 = 0.01
	 h1 = h1 - 0.01*roi/row0
	endif

      if (Turbpar/=1) then
    	call KT_eq(dt)
    	 do i=1,M+1
	  lamw(i)=lamw0+KT(i) !+KC(i)
	 enddo	
      else
	 lamw=lamw0*3.
      endif
      lamsal = lamw/(cw*row0) * alsal
   
!CASE 2.1: WATER, ICE AND SNOW 

	if (flag_snow == 1) then
	 SR=0;RS=0
       dhwlow = lamw(1)*dt/(ddz(1)*h1*row0*Lwi)*(Tw1(2) - Tw1(1))               VS,06.2007
     &  -lami*dt/(ddz(M)*l1*row0*Lwi)*(Ti1(M+1) - Ti1(M))                       VS,06.2007 
       dhilow = -dhwlow*row0/roi
	 if (Ti1(1)>Meltpnt(0.d0)+0.1) then
        dhwhigh = (Ti1(1)-Meltpnt(0.d0))*ci*roi*l1*ddz(1)/2/(Lwi*row0)          VS,06.2007
	  dhihigh = -dhwhigh*row0/roi
       else
	  dhwhigh = 0
	  dhihigh = 0
       endif
	 if (ls1/=0) then
	  ice=1;snow=1;water=1;deepice=1
        dhwls=-lamw(M)*dt/(ddz(M)*h1*row0*Lwi)*(Tw1(M+1) - Tw1(M))              VS,06.2007
     &   +lami*dt/(ddz(1)*ls1*row0*Lwi)*(Tis1(2) - Tis1(1))                      VS,06.2007
        dls = -dhwls*row0/roi
	  dhw = dhwhigh + dhwlow + snmelt*dt + dhwfsoil + dhwls
	  dhw0 = dhwlow+dhwhigh+snmelt*dt
	  dls0 = dls
	 else
	  ice=1;snow=1;water=1;deepice=0
        dls = 0.
	  dhwls = 0.
	  dhw = dhwhigh + dhwlow + snmelt*dt + dhwfsoil
	  dhw0 = dhwlow+dhwhigh+snmelt*dt
       endif
	 dhi = dhilow + dhihigh 
	 dhi0 = dhihigh
	 
	 call snowtemp(snowmass,snowmass_init,a,b,c,d,Temp,dt)

       h2 = h1 + dhw
	 l2 = l1 + dhi 
	 ls2 = ls1 + dls
	 
      else
	     
!CASE 2.2: WATER AND ICE WITHOUT SNOW

       do i=1,M
        SR(i)=Erad*dexp(-extice*l1)*dexp(-extwat*dzeta(dble(i+0.5))*h1)         VS,06.2007
	 enddo
	 do i=2,M
	  RS(i)=-(SR(i)-SR(i-1))/(0.5*(ddz(i-1)+ddz(i))*h1)                       VS,06.2007
       enddo
	 if (Ti1(1)>Meltpnt(0.d0)+1.E-5) then
        dhwhigh = (Ti1(1)-Meltpnt(0.d0))*ci*roi*ddz(1)*l1/2/(Lwi*row0)           VS,06.2007
!	  Ti1(1) = -1.E-5
!	  do i=2,M
!	   if (Ti1(i)>1.E-5) then
!	    dhwhigh = dhwhigh+Ti1(i)*ci*roi*l1*ddz/(Lwi*row)
!          Ti1(i)=-1.E-5
!         endif
!	  enddo
	  dhis = 0. 
	  dhihigh = -dhwhigh*row0/roi
       endif	
       if (precip>0.and. tempair<0.and.flag_snow /= 1.and.l1>0.05) then
	  flag_snow = 1
	  hs1 = 0.02
	  itop = ms-2
	  dhip = 0
       else
	  dhip = precip*row0/roi*dt
	 end if      
       dhwlow = lamw(1)*dt/(ddz(1)*h1*row0*Lwi)*(Tw1(2) - Tw1(1))                VS,06.2007
     &  -lami*dt/(ddz(M)*l1*row0*Lwi)*(Ti1(M+1) - Ti1(M))                        VS,06.2007
       dhilow = -dhwlow*row0/roi
       if (Ti1(1)<=Meltpnt(0.d0)+0.1) then
	  dhis = - Elatent/(roi*Liv)*dt
	  dhwhigh = 0
	  dhihigh = 0
       endif
	 dhi = dhihigh + dhilow + dhis + dhip
	 dhi0 = dhis + dhihigh + dhip
	 if (ls1/=0) then
	  ice=1;snow=0;water=1;deepice=1 
        dhwls =-lamw(M)*dt/(ddz(M)*h1*row0*Lwi)*(Tw1(M+1) - Tw1(M))              VS,06.2007
     &   +lami*dt/(ddz(1)*ls1*row0*Lwi)*(Tis1(2) - Tis1(1))                      VS,06.2007
        dls = -dhwls*row0/roi
	  dls0 = dls
	  dhw = dhwhigh + dhwfsoil + dhwls + dhwlow
	  dhw0 = dhwhigh+dhwlow
	 else
	  ice=1;snow=0;water=1;deepice=0
        dhw = dhwhigh + dhwfsoil + dhwlow
	  dhw0 = dhwhigh+dhwlow
	  dls = 0.
	 endif
	
	 call soilforlake(dt, h1, l1, ls1, hs1,
     &  a,b,c,d,Temp)
    	
       h2 = h1 + dhw
	 l2 = l1 + dhi
	 ls2 = ls1 + dls
	 
       endif
       
       ENDIF if2

!CASE 3: ICE WITHOUT WATER

      if3: IF (layer_case==3) THEN
          
!CASE 3.1: ICE WITH SNOW
      
	if (flag_snow == 1) then
	 if (Ti1(1)>Meltpnt(0.d0)+0.1) then
        dhwhigh = (Ti1(1)-Meltpnt(0.d0))*ci*roi*l1*ddz(1)/2/(Lwi*row0)           VS,06.2007
	  dhif = -dhwhigh*row0/roi
       else
	  dhwhigh = 0
	  dhif = 0
       endif
	 dhw = snmelt*dt + dhwhigh
	 dhi = dhif
	 dhi0 = dhi
	 ice=1;snow=1;water=0;deepice=0
	 call snowtemp(snowmass,snowmass_init,a,b,c,d,Temp,dt)
       l2 = l1 + dhi 
	 h2 = h1 + dhw
      
      else
!CASE 3.2: ICE WITHOUT SNOW !
       if (precip>0.and. tempair<0) then
	  flag_snow = 1
	  hs1 = 0.02
        dhip = 0
       else
	  dhip = precip*row0/roi*dt
	 end if
	 if (Ti1(1)>Meltpnt(0.d0)+0.1) then
        dhw = (Ti1(1)-Meltpnt(0.d0))*ci*roi*l1*ddz(1)/2/(Lwi*row0)               VS,06.2007
	  dhihigh = - dhw*row0/roi
	  dhis = 0
       else
        dhis = - Elatent/(roi*Liv)*dt
	  dhw = 0.
	  dhihigh = 0
       endif
	 dhi = dhis + dhihigh + dhip
	 dhi0 = dhi
	 ice=1;snow=0;water=0;deepice=0 
	 call soilforlake(dt, h1, l1, ls1, hs1,
     &  a,b,c,d,Temp)
	 			        	 		   	 				
	 l2 = l1 + dhi
	 h2 = h1 + dhw
	 
      endif
      
      ENDIF if3 

! CASE 4:SNOW ANS SOIL WITHOUT ICE AND WATER
			  
 5    if4: IF (layer_case==4) THEN
       print*, 'The non-operational case:                                       VS,02.10.07
     &  there is no water and ice layers: STOP'                                 VS,02.10.07
       STOP                                                                     VS,02.10.07
       if (flag_snow==1) then
    	  ice=0;snow=1;water=0;deepice=0
	  call snowtemp(snowmass,snowmass_init,a,b,c,d,Temp,dt)
       else
	  ice=0;snow=0;water=0;deepice=0
	  call soilforlake(dt, h1, l1, ls1, hs1,
     &   a,b,c,d,Temp)
	 endif 
	ENDIF if4

!     TRIBUTARY INFLOW to a lake
      h2 = h2 + trib_inflow/year_sec*dt 
	      
	if (h2 < 0.005.and.ls2/=0.and.l2==0) then
       ls2 = ls2 + h2*row0/roi
	 if (ls2 < 0.005) then 
	  ls2 = 0; Tis2 = 0
	 endif
	 h2 = 0.
	 Tw2 = 0.
	 Sal2 = 0.
      endif

	if ((h2<0.005.and.h2>0).and.l2 /= 0) then
	 l2 = l2 + h2*row0/roi
	 if (l2 < 0.005) then 
	  l2 = 0; Ti2 = 0
	 endif
	 h2 = 0.
	 Tw2 = 0.
	 Sal2 = 0. 
	end if
			  
	if (h2 < 0.005.and.l2/=0.and.ls2/=0) then
       Ti2 = Ti2*l2/(l2+ls2) + Ti2*ls2/(l2+ls2)                                 VS,06.2007
	 l2 = l2 + h2*row0/roi + ls2 
!	 addt=0                                                                   VS,06.2007
!	 do i=1,M+1                                                               VS,06.2007
!	  if ((i==1).or.(i==M+1)) then                                            VS,06.2007
!         addt = addt + Tis2(i)*ls2/l2*ddz/2+Ti2(i)*(l1/l2-1)*ddz/2             VS,06.2007
!	  else                                                                    VS,06.2007 
!	   addt = addt + Tis2(i)*ls2/l2*ddz+Ti2(i)*(l1/l2-1)*ddz                  VS,06.2007 
!	  endif                                                                   VS,06.2007
!       enddo                                                                   VS,06.2007
!	 Ti2 = Ti2 + addt                                                         VS,06.2007
	 ls2 = 0.
	 Tis2 = 0.
	 h2 = 0 
	 Tw2 = 0.
	 Sal2 = 0.
      endif
			 
      if (l2 < 0.005.and.h2 /= 0) then
	 h2 = h2 + l2*roi/row0
	 if (h2 < 0.005) then 
	  h2 = 0; Tw2 = 0; Sal2 = 0.
	 endif
	 l2 = 0.
	 Ti2 = 0.
	end if

	if (h2 < 0.005.and. l2 == 0) then
	 h2 = 0.
	 Tw2 = 0.
	 Sal2 = 0. 
	end if
	
	if (l2 < 0.005.and. h2 == 0) then
	 l2 = 0.
	 Ti2 = 0.
	end if
			 
	if ((hs1 < 0.005.and.hs1>0).or.(hs1>0.and.l2==0.and.h2/=0)) then
	 h2 = h2 + (totalprecips-totalevaps-totalmelts)
	 if (h2 < 0.005) then
	  l2 = l2 + row0*h2/roi
	  h2 = 0; Tw2 = 0; Sal2 = 0.
	  if (l2 < 0.005) then 
	   l2 = 0; Ti2 = 0
	  endif
	 endif
	 hs1 = 0.
	 flag_snow = 0. 
	 flag_snow_init = 1
      end if
		 			         
	if (ls2 < 0.005.and.h2/=0) then
	 h2 = h2 + ls2*roi/row0
	 if (h2 < 0.005) then 
	  h2 = 0; Tw2 = 0; Sal2 = 0.
	 endif
	 ls2 = 0
	 Tis2 = 0.
      endif 

	if (ls2>0.and.h2 == 0) then
	 l2 = ls2
	 Ti2 = Tis2
	 if (l2 < 0.005) then 
	  l2 = 0; Ti2 = 0
	 endif
	 ls2 = 0
	 Tis2 = 0.
      endif
		      
	h1 = h2
	l1 = l2
	ls1 = ls2
	Tw1 = Tw2
	Ti1 = Ti2
	Tis1 = Tis2
	Sal1 = Sal2
	Sals1 = Sals2
	   	
	! CALCULATION OF SUMMARY FLUXES !			
      totalpen = totalpen+dhwfsoil 
	totalpen1 = totalpen1+dhwfsoil
	totalevap = totalevap+Elatent/(row0*Lwv)*dt
	totalprecip = totalprecip+precip*dt
!	totalsurfrad = totalsurfrad + surfrad*dt
!	totallongwave = totallongwave + longwave*dt
	totalhflux = totalhflux + hflux*dt
	totalerad = totalerad + erad*dt
	
	if (l1==0) tsw=Tw1(1)+273.2
	if (l1/=0.and.flag_snow==0) tsw=Ti1(1)+273.2
	if (flag_snow==1) tsw=T(itop)+273.2
	
	if (init(ix,iy)==0) then
	 init(ix,iy)=1	
      endif
	
! VALUES AT NEXT TIME STEP (time + dt) IN CURRENT POINT (ix,iy)
      
      l1_2d(ix,iy)  = l1 
	h1_2d(ix,iy)  = h1
      ls1_2d(ix,iy)  = ls1

	do i=1,M+1
	 u_2d(i,ix,iy)=u1(i)
	 v_2d(i,ix,iy)=v1(i)
	enddo

	do i=1,M+1
	 E_2d(i,ix,iy)=E1(i)
	 eps_2d(i,ix,iy)=eps1(i)
	enddo
	  	            
      do i=1,ns 
       Tsoil1_2d(i,ix,iy)=Tsoil1(i)
       Sals1_2d(i,ix,iy) = Sals1(i) 
       wi1_2d(i,ix,iy)=wi1(i)
	 wl1_2d(i,ix,iy)=wl1(i)
      enddo
		 	  
	do i=1,M+1 
	 Tw1_2d(i,ix,iy) = Tw1(i)
	 Sal1_2d(i,ix,iy) = Sal1(i) 
       Ti1_2d(i,ix,iy) = Ti1(i)
	 Tis1_2d(i,ix,iy) = Tis1(i)
      enddo 
      			 
	fl_sn_2d(ix,iy) = flag_snow
	fl_sn_init_2d(ix,iy) = flag_snow_init

      itop_2d(ix,iy) = itop
	hs1_2d(ix,iy) = hs1

      do i=max(1,itop),ms-1
       dz_2d(i,ix,iy) = dz(i)
       T_2d(i,ix,iy) = T(i)
	 wl_2d(i,ix,iy) = wl(i)
	 dens_2d(i,ix,iy) = dens(i)
      enddo
      cdm2(ix,iy) = cdmw
      time_2d(ix,iy) = time
      dhwfsoil_2d(ix,iy) = dhwfsoil
      Elatent_2d(ix,iy) = Elatent
      dhw_2d(ix,iy) = dhw
	dhw0_2d(ix,iy) = dhw0
	velfrict_2d(ix,iy) = velfrict
      
      hw1=hw;xlew1=xlew;cdmw1=cdmw
      
      if (error_cov==1) then
       if (int(nstep*dt/(interval*60*60))>=nstep_meas) then
        nstep_meas = nstep_meas + 1
        if (nstep_meas<=ntim) then
         do i = 1, M+1
          j = (i-1)*n_modvar
          x_mod(nstep_meas,j+1)= Tw2(i)
!          x_mod(nstep_meas,j+2)= E2(i)
         enddo 
        endif
       endif
      endif 
      if (error_cov==1.and.nstep_meas==ntim) then
	 call HL_method
	 error_cov = 0
	endif

	if (ix==10.and.iy==10) print*, 'Lake',
     & 'T1=',Tw1(1),'T2=',Tw1(2),'h1=',h1,'S=',shortwave,
     & 'hh=',hflux,'LE=',Elatent,
     & 'eflux=',eflux, 'Longwave=', longwave,
     & 'Bal=',(shortwave*(1-albedoofwater)*(sabs)+longwave-
     &  surfrad-hflux-Elatent-eflux)/(cw*row0*ddz(1)*h1/2)*dt
     
!-----OUTPUTS IN STANDALONE RUN-----
      if (runmode==1) then
       if (time>spinup_day*24*60*60) then
        if (monthly_out ==1) call mon_out(dt) 
	  if (daily_out   ==1) call day_out(dt) 
	  if (hourly_out  ==1) call hour_out(dt)
	  if (everystep   ==1) call everystep_out
	  if (time_series ==1) call series_out(tsw)
	 endif
	 if (mod(nstep,nscreen)==0) then
        write (*,'(2a25)') 'timestep', 'Surface temperature'
        write (*,'(i25,f25.1)') nstep, tsw
        write (*,'(4a8)') 'Year', 'Month', 'Day', 'Hour'
        i = int(nyear); if (nyear<1000) i = int(nyear+1000)
        write (*,'(4i8)') i        ,      int(nmonth),
     &                    int(nday),      int(nhour)
	 endif
	endif
	      	      
	!FORMATS!
  7   format (f7.3, 4i5,37f7.3) 
  60  format (f5.2, 2e16.5, 2f8.3, e15.5, f11.5, f7.1)
  !61  format (i5, f6.2, 2e16.5, 2f8.3, e15.5, f11.5, f7.1)
  61  format (f6.2, f5.2, 2e16.5, 2f8.3, e15.5, f11.5, f7.1)
  62  format (f6.2, f5.2, 2f12.3, 2f8.3, f11.5, f7.1,3f12.4,2f7.2,3f7.1)
  80  format (f7.3, 10f9.2)
  90  format (f7.3, 12f9.2)   
  100 format (a5, 2a16, 2a8, a15, a11, a7) 
      !FINISHING PROGRAM!       
  
      END SUBROUTINE LAKE

!-----------------------------------------------------------------------------------
!                  TEMPERATURE EQUATION SOLVER
!-----------------------------------------------------------------------------------

      SUBROUTINE T_SOLVER(dt)
      
!     T_SOLVER implements iterations to find water surface temperature 
!     at the next time step    

	use driving_params
	use arrays
      implicit none

	real(8) dt_scan,Tsurf1,Tsurf2,Tsurf3,Bal1,Bal2,Bal3,dt,
     & Tsurf_prev
	integer(4) iter,maxiter,surftyp_prev,surftyp
	logical firstcall
	data firstcall /.true./

	SAVE
      
      if (firstcall) then
       open (666,file=path(1:len_trim(path))//
     &  'results/debug/iter_T.dat')
       maxiter=10
       dt_scan = 5.
      endif
      
      if (snow==1) then
	 surftyp=3
      elseif (ice==1) then
	 surftyp=2
      elseif (water==1) then
	 surftyp=4
	else
	 surftyp=1
      endif
      call SURF_CHAR(surftyp)
       
      if (.not.firstcall.and.runmode==1
     & .and.surftyp==surftyp_prev) then
       Tsurf2 = Tsurf_prev - 10.
      else
       Tsurf2=-90.
      endif
      
!     SCANNING INTERVAL OF SURFACE TEMPERATURE [-90 C, ...]
      
	Bal1=1.;Bal2=1.
	do while (Bal1*Bal2>=0)
	 Tsurf1=Tsurf2
	 call Bal_calc(Bal1,Tsurf1,dt,surftyp)  
	 Tsurf2=Tsurf1+dt_scan
	 if (Tsurf2>50.) then
	  print*, 'Temperature limit in scan process is exceeded: STOP'
	  do while (Tsurf2>50.)
	  enddo
	  STOP
       endif
       call Bal_calc(Bal2,Tsurf2,dt,surftyp)
	 continue
      enddo
      
	iter=0
	Bal3=1.
	
!     CHORDE METHOD TO FIND SURFACE TEMPERATURE

      cy1:do while (dabs(Bal3)>0.1)
	 iter=iter+1
	 call Bal_calc(Bal1,Tsurf1,dt,surftyp)
	 call Bal_calc(Bal2,Tsurf2,dt,surftyp)    
	 Tsurf3=(Tsurf1*Bal2-Tsurf2*Bal1)/(Bal2-Bal1)
	 call Bal_calc(Bal3,Tsurf3,dt,surftyp)
	 if (iter>maxiter) then
	  write (666,*) Bal3
	  exit cy1
	 endif
	 if (Bal1*Bal3<0.) then
	  Tsurf2=Tsurf3
       elseif (Bal2*Bal3<0.) then
	  Tsurf1=Tsurf3
       endif
	enddo cy1     

	call T_diff(0,Tsurf3,dt)
	Tsurf_prev = Tsurf3
	surftyp_prev = surftyp
	
	if (firstcall) firstcall = .false.
      END SUBROUTINE T_SOLVER
      

	SUBROUTINE BAL_CALC(Bal,Tsurf,dt,surftyp)
	
!     BAL_CALC calculates the components of energy balance at the water surface	

	use atmos
	use driving_params
	use numeric_params
	use phys_constants2
	use arrays
	implicit none

	real(8) Bal,Tsurf,dt

      real(8), dimension(1:ms):: Tsn,cs,lams,q
	real(8) shortwave,precip 
	real(8) TET2,TET1,ro,esatsurf,humsurf,bx(7),bix(11),
     & c_u,c_t,ksw,hwave,SHF,LHF,dHdt
      real(8) h1,l1,hs1,ls1,year,month,day,hour
	real(8) wind10,urel,vrel,u,v,wr
	real(8) Hm, LEm
	real(8) roughness,emissivity,albedo,aM,bM,relhums
	real(8) AL,DLT,DVT,ALLL,DL,
     & ALV,DV,Z,T,WL,WV,WI,dens,dz
     
	integer(4) itdrag,surftyp,itop

	common /layers/ h1,l1,hs1,ls1
	common /wind/ wind10,urel,vrel,u,v
	common /measflux/ Hm, LEm
	common /surface/ roughness,emissivity,albedo,aM,bM,relhums
	common /snow_char/ Tsn,cs
	common /atmos2/ shortwave, precip
	common /SOILSOL/ AL(ML),DLT(ML),DVT(ML),ALLL(ML),DL(ML),
     & ALV(ML),DV(ML),Z(ML),T(ML),WL(ML),WV(ML),WI(ML),dens(ms)
	common /SOILDAT/ dz(ms),itop
	common /watericesnowarr/ lams,q
	common /ymdh/ year,month,day,hour
	
	SAVE

	call T_diff(1,Tsurf,dt)
	SELECT CASE (surftyp)   
	 case (1)
	  eflux = -(lamsoil(1)+lamsoil(2))/2.*(Tsoil2(2)-Tsoil2(1))/dzs(1)
	 case (2)
	  eflux = -lami*(Ti2(2)-Ti2(1))/(ddz(1)*l1) 
	  !if (l1<0.1) eflux = -lami*(Ti2(2)-Ti2(1))/(ddz*0.1) 
	 case (3)
	  eflux = -(lams(itop)+lams(itop+1))/2.*(Tsn(itop+1)-Tsn(itop))
     &  /dz(itop)
       case (4)  
	  eflux = -lamw(1)*(Tw2(2)-Tw2(1))/(ddz(1)*h1) 
	  !if (h1<0.1) eflux = -lamw(1)*(Tw2(2)-Tw2(1))/(ddz*0.1) 
      END SELECT
	
!    (urel,vrel) is wind, relative to lake currents

      if (relwind==2) then
	 urel=uwind-u
	 vrel=vwind-v
	elseif (relwind==1) then
       urel=uwind
	 vrel=vwind
      endif
	wr=dmax1(dsqrt(urel**2+vrel**2),0.1d0)
!      wr=dmax1(dsqrt(urel**2+vrel**2),1.d0)
      wind10 = wr*log(10./0.01)/log(2./0.01) !0.01 m is water surface roughness
      TET2 = (tempair+273.15)*(100000./pressure)**0.286
	ro = pressure/(Rd*(tempair+273.15))  
      TET1 = (Tsurf+273.15)*(100000./pressure)**0.286  
	esatsurf = 610.7*10.**(aM*Tsurf/(bM+Tsurf))
	humsurf = 0.622/pressure*esatsurf*relhums
	if (PBLpar==1.or.PBLpar==11) then
	
c     Businger-Dayer, Beljaars parameterization for exchange coefficients

       itdrag=10
       bx(1) = wr !wind
	 bx(2) = TET2
	 bx(3) = TET1
	 bx(4) = humair
	 bx(5) = humsurf
	 bx(6) = zref
	 bx(7) = roughness 
       call dragvl (bx, bix, itdrag)
	 c_u = bix(8)
	 c_t = bix(9)
	 hflux = -cp*ro*c_u*c_t*wr*(TET2-TET1)
	 Elatent =  -Lwv*ro*c_u*c_t*wr*(humair-humsurf) 
	 tau = ro*c_u**2*wr**2
	endif

c     Parameterization of exchange coefficients acording to (Louis, 1979)

	if (PBLpar==2.or.PBLpar==21) 
     &call RichPBL(hflux,Elatent,tau,Tsurf,humsurf,roughness)
	
c     SHALLOW WATER EFFECT ON HEAT, MOISTURE AND MOMENTUM FLUXES (PANIN ET. AL., 2006)

      hwave = 0.07*wind10**2*(g*h1/dMAX1(wind10,1.d0)**2)**(3./5.)/g
	if (l1==0.and.h1/=0.and.(PBLpar==11.or.PBLpar==21)) then
	 ksw=2. 
	 hflux=hflux+hflux*ksw*hwave/h1
	 Elatent=Elatent+Elatent*ksw*hwave/h1
	 tau=tau+tau*ksw*hwave/h1
      endif
	cdmw=tau/(ro*wr);xlew=Elatent;hw=hflux
		
	velfrict = dsqrt(tau/ro)	
      if (l1 /= 0) Elatent = -Liv*ro*c_u*c_t*wr*(humair-humsurf) 

c CALCULATED sensible and latent heat fluxes are used in heat balance

	SHF=hflux
	LHF=Elatent
	
c MEASURED sensible and latent heat fluxes are used in heat balance
	!SHF = Hm
	!LHF = LEm
	
	if (snow==1) then
       Radbal = longwave-emissivity*sigma*(Tsurf+273.15)**4
      elseif (ice==0.and.water==1) then
	 Radbal = shortwave*(1-albedo)*(1-(1-sabs)*
     & exp(-extwat*0.5*ddz(1)*h1))+longwave-emissivity*sigma*
     &(Tsurf+273.15)**4
	elseif (ice==1.and.snow==0) then
	 Radbal = shortwave*(1-albedo)*(1-exp(-extice*l1))+longwave-
     & emissivity*sigma*(Tsurf+273.15)**4
	else
       Radbal = shortwave*(1-albedo)+longwave-emissivity*sigma*
     &(Tsurf+273.15)**4  
	endif 
	
      SELECT CASE (surftyp)   
	 case (1)
	  dHdt=csoil(1)*rosoil(1)*dzs(1)/2.*(Tsoil2(1)-Tsoil1(1))/dt
	 case (2)
	  dHdt=ci*roi*ddz(1)*l1/2.*(Ti2(1)-Ti1(1))/dt
	 case (3)
	  dHdt=cs(1)*dens(1)*dz(1)/2.*(Tsn(1)-T(1))/dt
       case (4)  
        dHdt=cw*row0*ddz(1)*h1/2.*(Tw2(1)-Tw1(1))/dt
	END SELECT

	surfrad = emissivity*sigma*(Tsurf+273.15)**4  
	botflux = -lamw(M)*(Tw2(M+1)-Tw2(M))/(ddz(M)*h1)

      Bal=dHdt-(Radbal-SHF-LHF)+eflux
	
	END SUBROUTINE BAL_CALC


      SUBROUTINE T_diff(surf,Tsurf,dt)
      
!     T_diff calculates temperature profile in water, soil, ice and snow,
!     with known temparature at the surface 

	use numeric_params
	use phys_constants2
	use driving_params
	use arrays
      implicit none

	real(8) dt,Tsurf,Hflow
	real(8), dimension(1:ms):: Tsn,lams,q,cs
	real(8), dimension(1:350):: a,b,c,d,Temp
	real(8) h1,l1,hs1,ls1
	real(8) AL,DLT,DVT,ALLL,DL,ALV,DV,Z,T,WL,WV,WI,dens
	real(8) dz
	real(8), external:: meltpnt
	integer(4) i,j,itop,surf
   

      common /snow_char/ Tsn,cs
	common /layers/ h1,l1,hs1,ls1
	common /SOILSOL/ AL(ML),DLT(ML),DVT(ML),ALLL(ML),DL(ML),
     & ALV(ML),DV(ML),Z(ML),T(ML),WL(ML),WV(ML),WI(ML),dens(ms)
	common /SOILDAT/ dz(ms),itop
	common /watericesnowarr/ lams,q
!	data num /1/
	
	SAVE

	Hflow=0.
!     Meltpnt - Melting point temperature, C degrees 
!	Meltpnt=0.
!	ddz=1./float(M)                                                           VS,06.2007

	if (surf==1) then
	 if (snow==1) then
	  c(itop)=1.
	  b(itop)=0.
	  d(itop)=Tsurf
	  if (itop<=ms-2) then
	   call diff_coef(a,b,c,d,itop+1,ms-1,itop+1,ms-1,3,dt) 
	  endif
	  if (ice==1) then
!----------------ICE-SNOW INTERFACE
	   a(ms)=-(lams(ms-1)+lams(ms))/(2.*dz(ms-1))
	   b(ms)=-lami/(ddz(1)*l1)                                                VS,06.2007
	   c(ms)=a(ms)+b(ms)-(cs(ms)*dens(ms)*dz(ms)/(2.*dt)+
     &   ci*roi*ddz(1)*l1/(2.*dt))                                              VS,06.2007
	   d(ms)=-Ti1(1)*(cs(ms)*dens(ms)*dz(ms)/(2.*dt)+
     &   ci*roi*ddz(1)*l1/(2.*dt))                                              VS,06.2007
!-----------------------------------
	   call diff_coef(a,b,c,d,2,M,ms+1,ms+M-1,2,dt) 
	   if (water==1) then
	    a(ms+M)=0.
	    c(ms+M)=1.
	    d(ms+M)=Meltpnt(Sal2(1))
	    call progonka2 (a,b,c,d,Temp,itop,ms+M)
          do i=itop, ms
	     Tsn(i)=Temp(i)
          enddo
	    do i=ms, ms+M
	     Ti2(i-ms+1)=Temp(i)
	    enddo 
         else
!----------------ICE-SOIL INTERFACE-------------------------
	    a(ms+M)=-lami/(ddz(M)*l1)                                             VS,06.2007
	    b(ms+M)=-(lamsoil(1)+lamsoil(2))/(2*dzs(1))
	    c(ms+M)=a(ms+M)+b(ms+M)-(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &    ci*roi*ddz(M)*l1/(2.*dt))                                             VS,06.2007
	    d(ms+M)=-Ti1(M+1)*(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &    ci*roi*ddz(M)*l1/(2.*dt))                                             VS,06.2007
!-----------------------------------------------------------       
          call diff_coef(a,b,c,d,2,ns-1,ms+M+1,ms+M+ns-2,1,dt) 
	    c(ms+M+ns-1)=1.
	    a(ms+M+ns-1)=1.
	    d(ms+M+ns-1)=Hflow*dzs(ns-1)/(lamsoil(ns-1)+lamsoil(ns))
	    call progonka2 (a,b,c,d,Temp,itop,ms+M+ns-1)
	    do i=itop, ms
	     Tsn(i)=Temp(i)
          enddo
	    do i=ms, ms+M
	     Ti2(i-ms+1)=Temp(i)
	    enddo 
	    do i=ms+M, ms+M+ns-1
	     Tsoil2(i-ms-M+1)=Temp(i)
          enddo
         endif
        else
!-------------------SNOW-SOIL INTERFACE-------------------------
         a(ms)=-(lams(ms-1)+lams(ms))/(2.*dz(ms-1))
	   b(ms)=-(lamsoil(1)+lamsoil(2))/(2*dzs(1))
	   c(ms)=a(ms)+b(ms)-(cs(ms)*dens(ms)*dz(ms)/(2.*dt)+
     &   csoil(1)*rosoil(1)*dzs(1)/(2.*dt))
	   d(ms)=-Tsoil1(1)*(cs(ms)*dens(ms)*dz(ms)/(2.*dt)+
     &   csoil(1)*rosoil(1)*dzs(1)/(2.*dt))
!---------------------------------------------------------------
         call diff_coef(a,b,c,d,2,ns-1,ms+1,ms+ns-2,1,dt) 
	   c(ms+ns-1)=1.
	   a(ms+ns-1)=1.
	   d(ms+ns-1)=Hflow*dzs(ns-1)/(lamsoil(ns-1)+lamsoil(ns))
	   call progonka2 (a,b,c,d,Temp,itop,ms+ns-1)
         do i=itop, ms
	    Tsn(i)=Temp(i)
         enddo
	   do i=ms, ms+ns-1
	    Tsoil2(i-ms+1)=Temp(i)
         enddo
        endif
       elseif (ice==1) then
	  b(1)=0.
	  c(1)=1.
	  d(1)=Tsurf
        call diff_coef(a,b,c,d,2,M,2,M,2,dt)  
	  if (water==1) then
	   a(M+1)=0.
	   c(M+1)=1.
	   d(M+1)=Meltpnt(Sal2(1))
	   call progonka2 (a,b,c,d,Temp,1,M+1)
	   do i=1,M+1
	    Ti2(i)=Temp(i)
         enddo
        else
!----------------ICE-SOIL INTERFACE-------------------------
	   a(M+1)=-lami/(ddz(M)*l1)                                               VS,06.2007
	   b(M+1)=-(lamsoil(1)+lamsoil(2))/(2*dzs(1))
	   c(M+1)=a(M+1)+b(M+1)-(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &   ci*roi*ddz(M)*l1/(2.*dt))                                              VS,06.2007
	   d(M+1)=-Ti1(M+1)*(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &   ci*roi*ddz(M)*l1/(2.*dt))                                              VS,06.2007
!-----------------------------------------------------------
         call diff_coef(a,b,c,d,2,ns-1,M+2,M+ns-1,1,dt) 
	   c(M+ns)=1.
	   a(M+ns)=1.
	   d(M+ns)=Hflow*dzs(ns-1)/(lamsoil(ns-1)+lamsoil(ns))
         call progonka2 (a,b,c,d,Temp,1,M+ns)
	   do i=1,M+1
	    Ti2(i)=Temp(i)
         enddo
	   do i=M+1,M+ns
	    Tsoil2(i-M)=Temp(i)
         enddo
	  endif
	 elseif (water==1) then  
	  b(1)=0.
	  c(1)=1.
	  d(1)=Tsurf
	  call diff_coef(a,b,c,d,2,M,2,M,4,dt) 
	  if (deepice==1) then
	   a(M+1)=0.
	   c(M+1)=1.
	   d(M+1)=Meltpnt(Sal2(M+1))
	   call progonka2(a,b,c,d,Temp,1,M+1) 
	   do i=1,M+1
	    Tw2(i)=Temp(i)
         enddo
        else
!---------------------WATER-SOIL INTERFACE------------------
	   a(M+1)=-lamw(M)/(ddz(M)*h1)                                            VS,06.2007
	   b(M+1)=-(lamsoil(1)+lamsoil(2))/(2*dzs(1))
	   c(M+1)=a(M+1)+b(M+1)-(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &   cw*row0*ddz(M)*h1/(2.*dt))                                              VS,06.2007 
	   d(M+1)=-Tw1(M+1)*(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &   cw*row0*ddz(M)*h1/(2.*dt))-SR(M)                                        VS,06.2007
!-----------------------------------------------------------
	   call diff_coef(a,b,c,d,2,ns-1,M+2,M+ns-1,1,dt)
	   c(M+ns)=1.
	   a(M+ns)=1.
	   d(M+ns)=Hflow*dzs(ns-1)/(lamsoil(ns-1)+lamsoil(ns))
	   call progonka2 (a,b,c,d,Temp,1,M+ns)
	   do i=1,M+1
	    Tw2(i)=Temp(i)
         enddo
	   do i=M+1,M+ns
	    Tsoil2(i-M)=Temp(i)
         enddo
        endif
	 else
	  b(1)=0.
	  c(1)=1.
	  d(1)=Tsurf  
        call diff_coef(a,b,c,d,2,ns-1,2,ns-1,1,dt) 
	  c(ns)=1.
	  a(ns)=1.
	  d(ns)=Hflow*dzs(ns-1)/(lamsoil(ns-1)+lamsoil(ns))
	  call progonka2 (a,b,c,d,Temp,1,ns)
	  do i=1,ns
	   Tsoil2(i)=Temp(i)
        enddo
       endif
	else   
       if (water==0) then
	  RETURN
       else
	  if (ice==1) then
	   b(1)=0.
	   c(1)=1.
	   d(1)=Meltpnt(Sal2(1))
         call diff_coef(a,b,c,d,2,M,2,M,4,dt)  
	   if (deepice==1) then
	    a(M+1)=0.
	    c(M+1)=1.
	    d(M+1)=Meltpnt(Sal2(M+1))
          call progonka2 (a,b,c,d,Temp,1,M+1)
          do i=1,M+1
	     Tw2(i)=Temp(i)
          enddo
          b(1)=0.
	    c(1)=1.
	    d(1)=Meltpnt(Sal2(M+1))
	    call diff_coef(a,b,c,d,2,M,2,M,5,dt)
!----------------DEEPICE-SOIL INTERFACE-------------------------
          a(M+1)=-lami/(ddz(M)*ls1)                                             VS,06.2007  
	    b(M+1)=-(lamsoil(1)+lamsoil(2))/(2*dzs(1))
	    c(M+1)=a(M+1)+b(M+1)-(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &    ci*roi*ddz(M)*ls1/(2.*dt))                                            VS,06.2007
	    d(M+1)=-Tis1(M+1)*(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &    ci*roi*ddz(M)*ls1/(2.*dt))                                            VS,06.2007
!-----------------------------------------------------------	    
          call diff_coef(a,b,c,d,2,ns-1,M+2,M+ns-1,1,dt)  
	    c(M+ns)=1.
	    a(M+ns)=1.
	    d(M+ns)=Hflow*dzs(ns-1)/(lamsoil(ns-1)+lamsoil(ns))
	    call progonka2 (a,b,c,d,Temp,1,M+ns)
	    do i=1,M+1
	     Tis2(i)=Temp(i)
	    enddo
	    do i=M+1,M+ns
	     Tsoil2(i-M)=Temp(i)
	    enddo
         else
!---------------------WATER-SOIL INTERFACE------------------
          a(M+1)=-lamw(M)/(ddz(M)*h1)                                            VS,06.2007
	    b(M+1)=-(lamsoil(1)+lamsoil(2))/(2*dzs(1))
	    c(M+1)=a(M+1)+b(M+1)-(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &    cw*row0*ddz(M)*h1/(2.*dt))                                              VS,06.2007
	    d(M+1)=-Tw1(M+1)*(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &    cw*row0*ddz(M)*h1/(2.*dt))                                              VS,06.2007 
!-----------------------------------------------------------
          call diff_coef(a,b,c,d,2,ns-1,M+2,M+ns-1,1,dt)
	    c(M+ns)=1.
	    a(M+ns)=1.
	    d(M+ns)=Hflow*dzs(ns-1)/(lamsoil(ns-1)+lamsoil(ns))
	    call progonka2 (a,b,c,d,Temp,1,M+ns)
	    do i=1,M+1
	     Tw2(i)=Temp(i)
          enddo
	    do i=M+1,M+ns
	     Tsoil2(i-M)=Temp(i)
          enddo
         endif
	  elseif (deepice==1) then   	    
         b(1)=0.
	   c(1)=1.
	   d(1)=Meltpnt(Sal2(M+1))
	   call diff_coef(a,b,c,d,2,M,2,M,5,dt)
!----------------DEEPICE-SOIL INTERFACE-------------------------
	   a(M+1)=-lami/(ddz(M)*ls1)                                               VS,06.2007
	   b(M+1)=-(lamsoil(1)+lamsoil(2))/(2*dzs(1))
	   c(M+1)=a(M+1)+b(M+1)-(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &   ci*roi*ddz(M)*ls1/(2.*dt))                                              VS,06.2007
	   d(M+1)=-Tis1(M+1)*(csoil(1)*rosoil(1)*dzs(1)/(2.*dt)+
     &   ci*roi*ddz(M)*ls1/(2.*dt))                                              VS,06.2007
!-----------------------------------------------------------	    
         call diff_coef(a,b,c,d,2,ns-1,M+2,M+ns-1,1,dt)  
	   c(M+ns)=1.
	   a(M+ns)=1.
	   d(M+ns)=Hflow*dzs(ns-1)/(lamsoil(ns-1)+lamsoil(ns))
	   call progonka2 (a,b,c,d,Temp,1,M+ns)
	   do i=1,M+1
	    Tis2(i)=Temp(i)
	   enddo
	   do i=M+1,M+ns
	    Tsoil2(i-M)=Temp(i)
	   enddo
        endif
       endif
	endif

	END SUBROUTINE T_diff

	
	SUBROUTINE S_diff(dt)
	use arrays
	use driving_params
	implicit none
	
!     S_diff solves salinity (mineralization) diffusion equation	

      real(8) dt,Sflux1,Sflux_soil_bot
      real(8), dimension(1:350):: a,b,c,d,Sal
      real(8) h1,l1,hs1,ls1
      
!     Sflux1 --- salinity flux at the bottom boundary       
!     Sflux0 --- salinity flux at the top boundary       

!     Sflux0 = -dh0/dt*Sal2(1)

      common /layers/ h1,l1,hs1,ls1
      
!      ddz = 1./float(M)
      Sflux1 = 0.
      Sflux_soil_bot = 0.                                                       VS,03.10.07
            
!     c(1)   = 1. + dhw0*ddz(1)*h1/(dt*lamsal(1))                               VS,06.2007
!	b(1)   = 1.
!	d(1)   = 0. ! Sflux0*ddz*h1/lamsal(1)
!	c(M+1) = 1.
!	a(M+1) = 1.
!     d(M+1) = Sflux1*ddz(M)*h1/lamsal(M)                                       VS,06.2007
      if (water==1) then                                                        VS,01.10.07
       call diff_coef(a,b,c,d,2,M,2,M,6,dt)                                     VS,01.10.07
       c(1)   = -lamsal(1)/(h1*ddz(1))-dhw/dt+dhw0/(2*dt)-                      VS,23.09.07
     &  ddz(1)*h1/(2*dt)                                                        VS,23.09.07
       b(1)   = -lamsal(1)/(h1*ddz(1))+dhw0/(2*dt)                              VS,23.09.07
       d(1)   = -ddz(1)*h1*Sal1(1)/(2*dt)                                       VS,23.09.07
       if (deepice==1) then                                                     VS,01.10.07
!       Case water,deepice and soil; upper ice and snow are allowed       
!       Salinity diffusion in water       
        c(M+1) = -lamsal(M)/(ddz(M)*h1)-ddz(M)*h1/(2*dt)+                       VS,23.09.07
     &   (dhw-dhw0)/(2*dt)                                                      VS,23.09.07
        a(M+1) = -lamsal(M)/(ddz(M)*h1)+(dhw-dhw0)/(2*dt)                       VS,23.09.07
        d(M+1) = -Sal1(M+1)*ddz(M)*h1/(2*dt)+Sflux1                             VS,01.10.07
        call progonka2 (a,b,c,d,Sal,1,M+1)                                      VS,01.10.07
        Sal(:)=dmax1(Sal(:),0.d0)                                               VS,04.10.07
        Sal2(1:M+1)=Sal(1:M+1)                                                  VS,01.10.07
!       Salinity diffusion in soil
        call diff_coef(a,b,c,d,2,ns-1,2,ns-1,7,dt)                              VS,03.10.07
        c(1) = -1-dt*wsoil(1)/dzs(1)                                            VS,03.10.07
        b(1) = dt*wsoil(1)/dzs(1)                                               VS,03.10.07
        d(1) = -Sals1(1)                                                        VS,03.10.07
        c(ns) = -1+wsoil(ns-1)*dt/(dzs(ns-1))                                   VS,03.10.07
        a(ns) = -wsoil(ns-1)*dt/(dzs(ns-1))                                     VS,03.10.07
        d(ns) = -Sals1(ns)+2*dt*Sflux_soil_bot/dzs(ns-1)                        VS,03.10.07
        call progonka2 (a,b,c,d,Sal,1,ns)                                       VS,01.10.07
        Sal(:)=dmax1(Sal(:),0.d0)                                               VS,04.10.07        
        Sals2(1:ns)=Sal(1:ns)                                                   VS,01.10.07
       else                                                                     VS,01.10.07
!      Case water, soil; upper ice and snow are allowed       
!------WATER-SOIL INTERFACE-------------------------
        a(M+1) = 0.5*(dhw-dhw0)/dt - lamsal(M)/(ddz(M)*h1)                      VS,03.10.07
        b(M+1) = 0.5*wsoil(1)                                                   VS,03.10.07
        c(M+1) = -( 0.5*(dzs(1)+ddz(M)*h1)/dt + 0.5*wsoil(1) +                  VS,03.10.07
     &   lamsal(M)/(ddz(M)*h1)+0.5*(dhw0-dhw)/dt )                              VS,03.10.07
        d(M+1) = -Sal1(M+1)*0.5*(dzs(1)+ddz(M)*h1)/dt                           VS,03.10.07
        call diff_coef(a,b,c,d,2,ns-1,M+2,M+ns-1,7,dt)                          VS,03.10.07
        c(M+ns) = -1+wsoil(ns-1)*dt/(dzs(ns-1))                                 VS,03.10.07
        a(M+ns) = -wsoil(ns-1)*dt/(dzs(ns-1))                                   VS,03.10.07
        d(M+ns) = -Sals1(ns)+2*dt*Sflux_soil_bot/dzs(ns-1)                      VS,03.10.07
        call progonka2 (a,b,c,d,Sal,1,M+ns)                                     VS,01.10.07
        Sal(:)=dmax1(Sal(:),0.d0)                                               VS,04.10.07        
        Sals2(1:ns)=Sal(M+1:M+ns)                                               VS,01.10.07
        Sal2(1:M+1)=Sal(1:M+1)
       endif
      else
!     Case of the soil and the ice above     
       call diff_coef(a,b,c,d,2,ns-1,2,ns-1,7,dt)                               VS,03.10.07
        c(1) = -1-dt*wsoil(1)/dzs(1)                                            VS,03.10.07
        b(1) = dt*wsoil(1)/dzs(1)                                               VS,03.10.07
        d(1) = -Sals1(1)                                                        VS,03.10.07
        c(ns) = -1+wsoil(ns-1)*dt/(dzs(ns-1))                                   VS,03.10.07
        a(ns) = -wsoil(ns-1)*dt/(dzs(ns-1))                                     VS,03.10.07
        d(ns) = -Sals1(ns)+2*dt*Sflux_soil_bot/dzs(ns-1)                        VS,03.10.07
        call progonka2 (a,b,c,d,Sal,1,ns)                                       VS,01.10.07
        Sal(:)=dmax1(Sal(:),0.d0)                                               VS,04.10.07        
        Sals2(1:ns)=Sal(1:ns)                                                   VS,01.10.07
      endif 
		
	END SUBROUTINE S_diff
	
	
!---------------------------------------------------------------------------------
      

	SUBROUTINE diff_coef(a,b,c,d,n0,n1,m0,m1,substr,dt)
	
!-------------------DEFINES COEFFICIENTS FOR SOLVING A THERMAL DIFFUSIVITY-----------------
!-------------------EQUATION BY FACTORIXATION METHOD----------------------------------------

      use numeric_params
	use phys_constants2	 
	use driving_params
	use arrays
	implicit none

	real(8), dimension(1:ms):: Tsn,lams,cs,q
	real(8), dimension(1:350):: a,b,c,d
	real(8) lam1,lam2,dzmean,dt
	real(8) h1,l1,hs1,ls1
	real(8) AL,DLT,DVT,ALLL,DL,ALV,DV,Z,T,WL,WV,WI,dens
	real(8) dz
	real(8), external:: dzeta
	integer(4) i,j,n0,n1,m0,m1,substr,itop

      common /layers/ h1,l1,hs1,ls1
	common /snow_char/ Tsn,cs
	common /SOILSOL/ AL(ML),DLT(ML),DVT(ML),ALLL(ML),DL(ML),
     & ALV(ML),DV(ML),Z(ML),T(ML),WL(ML),WV(ML),WI(ML),dens(ms)
	common /SOILDAT/ dz(ms),itop
	common /watericesnowarr/ lams,q

	SAVE

!      ddz=1./float(M)                                                          VS,06.2007
      if (m1-m0/=n1-n0) then
	 print*, 'Error: diff_coef'
	 STOP
      endif

!     substr is a type of media : 1 - soil
!                                 2 - ice
!                                 3 - snow
!                                 4 - water
!                                 5 - deepice

!                                 6 - salinity in water

	SELECT CASE (substr)
!     CASE 1: SOIL
	 case (1)
       do i=m0,m1
	  j=i-m0+n0
	  lam2 = (lamsoil(j)+lamsoil(j+1))/2
	  lam1 = (lamsoil(j-1)+lamsoil(j))/2
	  dzmean = (dzs(j-1)+dzs(j))/2
	  a(i)=-lam1/dzs(j-1)
	  b(i)=-lam2/dzs(j)
	  c(i)=-(lam1/dzs(j-1)+lam2/dzs(j))-
     &   1./dt*(csoil(j)*rosoil(j)*dzmean)
	  d(i)=-Tsoil1(j)/dt*(csoil(j)*rosoil(j)*dzmean)
     	 enddo 
!      CASE 2: ICE
	 case (2)
       do i=m0,m1
	  j=i-m0+n0
	  c(i)=-(lami/ddz(j-1)+lami/ddz(j))/l1-                                   VS,06.2007
     &   1./dt*0.5*(ddz(j)+ddz(j-1))*l1*ci*roi                                  VS,06.2007
	  a(i)=-lami/(ddz(j-1)*l1)+ci*roi*dhi*dzeta(dble(j))/(2*dt)-              VS,06.2007 
     &  ci*roi*dhi0/(2*dt)  
        b(i)=-lami/(ddz(j)*l1)-ci*roi*dhi*dzeta(dble(j))/(2*dt)+                VS,06.2007
     &  ci*roi*dhi0/(2*dt)  
	  d(i)=-Ti1(j)/dt*0.5*(ddz(j)+ddz(j-1))*l1*ci*roi                         VS,06.2007
	  !dt*Erad*(1-sabs)*extwat/(cw*row)*
     	 enddo
!      CASE	3: SNOW  
	 case (3)
       do i=m0,m1
	  j=i-m0+n0
	  lam2 = (lams(j)+lams(j+1))/2
	  lam1 = (lams(j-1)+lams(j))/2
	  dzmean = (dz(j-1)+dz(j))/2
	  a(i)=-lam1/dz(j-1)
	  b(i)=-lam2/dz(j)
	  c(i)=-(lam1/dz(j-1)+lam2/dz(j))-
     &   1./dt*(cs(j)*dens(j)*dzmean)
	  d(i)=-T(j)/dt*(cs(j)*dens(j)*dzmean)
     	 enddo 
!      CASE 4: WATER
	 case (4)
	 do i=m0,m1
	  j=i-m0+n0
        c(i)=-(lamw(j)/ddz(j)+lamw(j-1)/ddz(j-1))/h1                            VS,06.2007
     &   -1./dt*0.5*(ddz(j-1)+ddz(j))*h1*cw*row0                                 VS,06.2007
        a(i)=-lamw(j-1)/(ddz(j-1)*h1)+cw*row0*dhw*dzeta(dble(j))/                VS,24.06.2007
     &   (2*dt)-cw*row0*dhw0/(2*dt)  
        b(i)=-lamw(j)/(ddz(j)*h1)-cw*row0*dhw*dzeta(dble(j))/(2*dt)+             VS,06.2007
     &   cw*row0*dhw0/(2*dt)  
        d(i)=-Tw1(j)/dt*0.5*(ddz(j)+ddz(j-1))*h1*cw*row0                         VS,06.2007
     &	 -RS(j)*0.5*(ddz(j)+ddz(j-1))*h1                                         VS,06.2007 
	  !dt*Erad*(1-sabs)*extwat/(cw*row)*
     	 enddo 
!      CASE 5: DEEP ICE
       case (5)
	 do i=m0,m1
	  j=i-m0+n0
        c(i)=-(lami/ddz(j)+lami/ddz(j-1))/ls1                                   VS,06.2007 
     &   -1./dt*0.5*(ddz(j)+ddz(j-1))*ls1*ci*roi                                VS,06.2007 
        a(i)=-lami/(ddz(j-1)*ls1)+ci*roi*dls*dzeta(dble(j))/(2*dt)-             VS,06.2007
     &  ci*roi*dls0/(2*dt)  
        b(i)=-lami/(ddz(j)*ls1)-ci*roi*dls*dzeta(dble(j))/(2*dt)+               VS,06.2007
     &  ci*roi*dls0/(2*dt)  
        d(i)=-Tis1(j)/dt*0.5*(ddz(j)+ddz(j-1))*ls1*ci*roi                       VS,06.2007
	  !dt*Erad*(1-sabs)*extwat/(cw*row)*
     	 enddo
!      CASE 6: SALINITY IN WATER
	 case (6)
	 do i=m0,m1
	  j=i-m0+n0
        c(i)=-(lamsal(j)/ddz(j)+lamsal(j-1)/ddz(j-1))/h1                        VS,06.2007
     &   -1./dt*0.5*(ddz(j)+ddz(j-1))*h1                                        VS,06.2007
        a(i)=-lamsal(j-1)/(ddz(j-1)*h1)+dhw*dzeta(dble(j))/(2*dt)               VS,06.2007
     &   -dhw0/(2*dt)         
        b(i)=-lamsal(j)/(ddz(j)*h1)-dhw*dzeta(dble(j))/(2*dt)                   VS,06.2007
     &   +dhw0/(2*dt)             
        d(i)=-Sal1(j)/dt*0.5*(ddz(j)+ddz(j-1))*h1                               VS,06.2007
	  !-RS(j)*ddz*h1 !dt*Erad*(1-sabs)*extwat/(cw*row)*
     	 enddo 
     	 case (7)                                                                 VS,01.10.2007
!      CASE 7: SALINITY IN THE SOIL     	 
     	 do i = m0, m1                                                            VS,01.10.2007
     	  j = i-m0+n0                                                             VS,01.10.2007
     	  a(i) = -wsoil(j-1)*dt/(dzs(j-1)+dzs(j))                                 VS,01.10.2007
     	  b(i) = wsoil(j)*dt/(dzs(j-1)+dzs(j))                                    VS,01.10.2007
     	  c(i) = -1-wsoil(j)*dt/(dzs(j-1)+dzs(j))+                                VS,01.10.2007
     &	 wsoil(j-1)*dt/(dzs(j-1)+dzs(j))                                         VS,01.10.2007
        d(i) = -Sals1(j)                                                       VS,01.10.2007
     	 enddo                                                                    VS,01.10.2007
      END SELECT
  
	END SUBROUTINE diff_coef


      SUBROUTINE SURF_CHAR(surftyp)
	use driving_params
	use phys_constants2
	use arrays
	
!----------------DEFINES SURFACE CHARACTERISTICS:-------------------
!----------------roughness,emissivity,albedo,relative humidity,-----
!----------------coefficients in Magnus formula---------------------

	implicit none
      real(8) roughness,emissivity,albedo,aM,bM,niu,relhums,
     & sinh0
      integer(4) surftyp
      common /surface/ roughness,emissivity,albedo,aM,bM,relhums
       
	SAVE 
	 	
!     niu is air molecular viscosity
      niu = 1.5d-5 !1.007E-6
      select case (surftyp)
	 case (1)
        albedo = albedoofsoil
	  roughness = 0.05
	  aM = aMagw 
	  bM = bMagw
	  emissivity = emissivityofsoil
	  relhums = 0.5
       case (2) 
        albedo = albedoofice
	  emissivity = emissivityofice
	  aM = aMagi
	  bM = bMagi 
	  roughness = 0.00001
	  relhums = 0.7
	 case (3)
	  albedo = albedoofsnow
	  emissivity = emissivityofsnow
        aM = aMagi
	  bM = bMagi 
	  roughness = 0.001
	  relhums = 0.7
       case (4)
	  velfrict_prev = dmax1(velfrict_prev, 1.d-2)            
	  roughness = dmin1(dmax1(0.111*niu/velfrict_prev+
     &  0.0144*velfrict_prev**2/g, 1.d-5),1.1d-5)
!        print*, roughness
	  aM = aMagw
	  bM = bMagw
	  if (varalb==0) then
	   albedo = albedoofwater
	  elseif (varalb==1) then
	   albedo = 0.05/(sinh0()+0.15)   
	   !albedo = (dirdif()*0.05/(sinh0()+0.15)+0.05)/
     &   !(1+dirdif())
	  endif
	  emissivity = emissivityofwater 
	  relhums = 1.
	end select	

	END SUBROUTINE SURF_CHAR

      REAL(8) FUNCTION dirdif()
      implicit none
	real(8) cloud
	real(8) dirdif0,b,S0,shortwave,precip,sinh0
	common /cloud/ cloud
	common /atmos2/ shortwave,precip
!	data cloud /0./

	SAVE

      b=1./3.
	S0=1367.
	cloud = 0.
    
      dirdif0 = (shortwave-b*S0*sinh0())/(b*(S0*sinh0()-shortwave))
	dirdif = dirdif0*(1.-sngl(cloud))
	dirdif = dmax1(dirdif,0.d0)
      
	END FUNCTION dirdif

	REAL(8) FUNCTION sinh0()
	use driving_params
!     sinh0 is sine of solar height (or cosine of zenith angle)	
	
      implicit none
	
	real(8) delta,theta,pi
	real(8) hour,phi_rad
	real(8) year,month,day,nday
	
      common /ymdh/ year,month,day,hour

      pi=dble(4.*atan(1.))		

	nday = day+(month-1)*30
	delta = 23.5*pi/180.*dCOS(2*pi*(nday-173)/365.)
	theta = pi*(hour-12.)/12.
	phi_rad = phi_g*pi/180.
	!lam_rad = lam_g*pi/180.
      sinh0 = dSIN(phi_rad)*dSIN(delta)+
     & dCOS(phi_rad)*dCOS(delta)*dCOS(theta)
	sinh0=dmax1(sinh0,dble(0.d0)) 

	END FUNCTION sinh0

c_____________________________________________________________________________

	SUBROUTINE RichPBL(h,le,tau,Tsurf1,humsurf,z0)
	
!     RichPBL calculates heat fluxes according to Louis parameterization (Louis, 1079)	
	
	use atmos
	implicit none
      
	real(8),parameter:: gg = 9.80665
      real(8),parameter:: ra = 287.05
      real(8),parameter:: rv = 461.51
      real(8),parameter:: cp = 1005.46
      real(8),parameter:: cpv = 1869.46
      real(8),parameter:: xl = 2500800.
      real(8),parameter:: stefan = 5.6697d-08
      real(8),parameter:: xpi = 3.141592654
	real(8),parameter:: too = 273.16
      real(8),parameter:: xpoo = 100000.
      real(8),parameter:: vkarmn = .4
      real(8),parameter:: xli = 2834500.
      real(8),parameter:: tcdil = 86400.
      real(8),parameter:: rascp = ra / cp
      real(8),parameter:: xlsg = xl / gg
      real(8),parameter:: xlscp = xl / cp
      real(8),parameter:: cpsl = cp / xl
      real(8),parameter:: cpsg = cp / gg
      real(8),parameter:: gscp = gg / cp
      real(8),parameter:: unsg = 1. / gg
      real(8),parameter:: gsra = gg / ra
      real(8),parameter:: rasl = ra / xl
      real(8),parameter:: rascp2 = ra / (2.*cp)
      real(8),parameter:: rasrv = ra / rv
      real(8),parameter:: rvsra = rv / ra
      real(8),parameter:: etv = rvsra - 1.
      real(8),parameter:: ecph = cpv / cp - 1.
      real(8),parameter:: xlsrv = xl / rv
      real(8),parameter:: unscp = 1. / cp
      real(8),parameter:: cpvmcp = cpv - cp
      real(8),parameter:: etvq = 1. - rasrv
      real(8),parameter:: xlf = xli - xl
      real(8),parameter:: xliscp = xli / cp
      real(8),parameter:: xlisg = xli / gg

	real(8) z0hz0,Tsurf1,humsurf,z,xmu,xfh,ts,h,le,leg,
     & urel,vrel,ta,qa,ps,ua,z0h,z0,hu,veg,ztvi,rhoa,zdepl,zua,ztvis,Ri,
     & valnon,zeps,zch,zsta,iyra,zdi,rra,zdim,cdm,zds,chstar2,qs,ph2,
     & cdh,cmstar2,pm2,zdsm,xhu,tau,wind10,u,v
	common /wind/ wind10,urel,vrel,u,v

	SAVE
      
      ta=tempair+273.15
	qa=humair
	ps=pressure
	ua=dsqrt(urel**2+vrel**2)
	ts=Tsurf1+273.15
	qs=humsurf

      z0hz0=0.1
	z0h=z0*z0hz0

c
c**********************************************************************
c           delta calculatlon
c
!      delta=0.
!      if (veg.gt.0.) delta=(wr(li)/wrmax)**(2./3.)
c
c**********************************************************************
c                         hv calculation
c                         ______________
c
      ! hv = 1. - dmax1(0.d0,dsign(1.d0,qsat(ts(li), ps) - qa)) * rs *
      !&      (1. - delta) / (rra + rs )

	  hu=1. !air above water is saturated
	  veg=0. !no vegetation on the lake
c
c?    rra  a resistencia aerodinamica = 1/ChVa  -volta
c        a ser calculada  frente (correctamente

c        Implicit resolution of the surface temperature equation
c        linearization of Ts

        ztvi = ta * (1. + etv * qa)
        rhoa = ps / (ra * ztvi)
        !qs = (hu * (1. - veg) + veg * hv) * qsat (ts(li),ps) +
     &  !     (1. - hv) * veg * qa
c
c**********************************************************************
c                CMWF, 1981/11/25-27, pp. 59-79)
c**********************************************************************
c
        zdepl = 0.
        !write(*,*) 'nhli9904zref',zref
        zua=zref + zdepl
        ztvis = ts * (1. + etv * qs)
c______________________________________________________________________
c       calculo do numero de richardson e da resistencia aerodinamica
c
      !write(*,*) 'zref',zref
      ua=dmax1(1.d0,ua)
      ri=2.*gg*zua*zua*(ztvi-ztvis+gscp*zref)/(ztvi+ztvis)/(ua*ua)/zref
c
c       calculo do comprimento de Monin -Obukhov,de ustar,tstar
c
        valnon=-999.
c        call calclmon(z,zO,zOh,ua,ztvi,ztvls,xlmon,ustar,tstar,valnon)
        zeps= 1.e-6
        !write(* ,*) 'zua,z0',zua,z0
        zch=(vkarmn/dlog(dabs(zua/z0)))**2
        !write(*,*) 'zch',zch
        zsta=ri*ua*ua
c
c  iyra escolhe entre a nova e a velha versao da rotina yra
c         de NP89. Quando iyra (ficheiro de entrada(?!)) e maior
c         do que 0 utiliza-se a nova versao que destingue entre
c         os comprimentos de rugosidade
c
        iyra=1
        if(iyra.eq.0) then
c
          if(ri.lt.0.) then
           zdi=1./(ua+75.*zch*dsqrt(-zref*zsta/z0h))
           rra=zch*(ua-15.*zsta*zdi)
           zdim=1./(1.+75.*zch*dsqrt(-zref*ri/z0))
           cdm=zch*(1.-10.*ri*zdim)*ua
          else
           zds=dsqrt(ua * ua + 5. * zsta + zeps)
           rra = zch*ua/(1.+15.*zsta*zds/ua/ua/ua)
           zdim=(1.+5.*ri)**2.
           cdm=zch*ua/zdim
          endif
c
        else
c
c nova rotina yramng
          ! write(*,*) 'estez0h',z0h
          !write(*,*) 'z,zref',z,zref
           z=zref
           xmu=dlog(z0/z0h)
           xfh=dlog(z/z0)/dlog(z/z0h)
          ! write(*,*) '--------------------'
          ! write(*,*) 'z0,z0h',z0,z0h
           if(ri.lt.0.) then
                zdi=1./(ua+chstar2(xmu)*zch*15.*(z/z0h)**
     &		  ph2(xmu)*xfh
     s              *dsqrt(-zsta))
                rra=zch*(ua-15.*zsta*zdi)*xfh
c
c alteracoes para o calculo do coeficiente de dragg_______________
c        cdm=Cd*uvsuf
c
                cdh=rra
                zdim=1./(ua+cmstar2(xmu)*zch*10.*(z/z0)**
     &		  pm2(xmu)
     s              *dsqrt(-zsta))
                cdm=zch*(ua-10.*zsta*zdim)
c_______________________________________________________________________
           else
                zds = dsqrt(ua * ua + 5. * zsta + zeps)
                rra = zch*ua/(1.+15.*zsta*zds/ua/ua/ua)*xfh
c
                cdh=rra
                zdsm= dsqrt(ua * ua + 5. * zsta + zeps)
                cdm = zch*ua/(1.+10.*zsta/zdsm/ua)
	
           endif
c
        endif
c
        rra=1./rra
c----------------------------------------------------------------------
c
!      zrsra = rhoa / rra
!      if (iclay.ge.0) then
!         ct = 1. / ((1. - veg) / cso + veg / cv)
!         hv=1.-dmax1(0.d0,dsign(1.d0,qsat(ts(li), ps) - qa)) * rs*
!     &    (1. - delta) / (rra + rs)
c
!         za=1. / dt + ct * (4. * emis * stefan * (ts(li)**3) +
!     &      zrsra*xl*dqsat(ts(li),ps)*(veg*hv+(1.-veg)*hu*xhu)
!     &      +zrsra*cp)+2.*xpi/tcdil
!         zb=1./dt+ ct*(3. * emis * stefan * (ts(li)** 3) +
!     &      zrsra*xl*dqsat(ts(li),ps)*(veg*hv+(1.-veg)*hu*xhu))
!         zc=2.*xpi*t2(li)/tcdil+ct*(zrsra*cp*ta+rg*
!     &      (1.-alb)+emis*rat-zrsra*xl*(veg*hv*(qsat(ts(li),
!     &      ps)-qa)+(1.-veg)*xhu*(hu*qsat(ts(li),ps)-qa)))
c
!         ts(lf) = (ts(li) * zb + zc) / za
c
c       resolution of t2 equation
c
!         t2(lf) = (t2(li) + dt * ts(lf) / tcdil) / (1. + dt / tcdil)
!      else
!          ts(lf)=tsfunc
!          t2(lf)=t2(li)
!      endif
c
c**********************************************************************
c
      xhu=1.
      !hqs2=hu*qsat(ts(lf),ps)
      !rn = rg * (1. - alb) + emis * (rat - stefan * (ts(lf)**4))
      h = rhoa * cp * (ts - ta) / rra
      leg = rhoa * xl*(1.-veg) *xhu*(qs - qa)/ rra 
      !lev = rhoa * xl * veg * hv * (qsat(ts(lf), ps) - qa) / rra
      !zzhv = dmax1(0.d0,dsign(1.d0,qsat(ts(li), ps) - qa))
      !letr = zzhv * (1. - delta) * rhoa * xl * veg * (qsat(ts(lf), ps)
     &!       - qa) / (rra + rs)
	le = leg !+ lev
	tau = rhoa*cdm*ua
      !hw=h;xlew=le;cdmw=cdm

      return
	END SUBROUTINE RichPBL
      	
      function chstar2(xmu)

      implicit real*8(a-h,o-z)
      chstar2=3.2165+4.3431*xmu+.536*xmu*xmu-.0781*xmu*xmu*xmu
      return
      end

      function cmstar2(xmu)

      implicit real*8(a-h,o-z)
      cmstar2=6.8741+2.6933*xmu+.3601*xmu*xmu-.0154*xmu*xmu*xmu
      return
      end

      function ph2(xmu)

      implicit real*8(a-h,o-z)
      ph2    =0.5802-0.1571*xmu+.0327*xmu*xmu-.0026*xmu*xmu*xmu
      return
      end
c
c******************************************************************
      function pm2(xmu)
c****************************************************************
c
      implicit real*8(a-h,o-z)
      pm2    =0.5233-0.0815*xmu+.0135*xmu*xmu-.001*xmu*xmu*xmu
      return
      end
	
      SUBROUTINE dragvl(bx2,bix2,itdrag)
      
!     dragvl calculates exchange coefficients in aerodynamic formulas for surface
!     sensible heat, latent heat and momentum fluxes at the surface
!     following Businger-Dayer interpolation formulas, Beljaars parameterization and others      
      
      real bx(7),bix(11)
	real(8) bix2(11),bx2(7)
	
      common /p_drag/ vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     *                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     *                zeta_st,a_m,a_t,b_m,b_t
	real vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     *                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     *                zeta_st,a_m,a_t,b_m,b_t,grav
	real u_a,tpot_a,tpot_s,q_a,q_s,z,z0m,z0t,tvir_a,
     &tvir_s,dtvir,alam,c_u,c_t,ust,hfl,re,xx,zeta,zeta0m,
     &zeta0t,smo,psi_m,psi_t,phi_m,phi_t
      INTEGER iter,itdrag
     
	   
      grav = 9.81
	
	bx=sngl(bx2)  

      u_a = amax1(bx(1),0.1) ! speed 
      tpot_a = bx(2) ! potential temperature - air
      tpot_s = bx(3) ! potential temperature - surface
      q_a = bx(4)
      q_s = bx(5)
      z = bx(6)
      if(bx(7) .gt. 0.) then
		z0m = bx(7)
      else
		z0m = z0min
      end if
      tvir_a = tpot_a * (1. + 0.61 * q_a) ! virtual temp - air
      tvir_s = tpot_s * (1. + 0.61 * q_s) ! virtual temp - surf
      dtvir = tvir_a - tvir_s  ! for smo
      alam = grav/(0.5*(tpot_a + tpot_s))  ! for smo
      c_u = vkc/log(z/z0m) ! for smo
      c_t = c_u
      do iter = 1,itdrag
		ust = c_u * u_a  ! for smo
		hfl =  - c_u * c_t * u_a * dtvir ! for smo Hflux
		if(bx(7) .lt. 0.) z0m = amax1(chc*ust**2/grav, z0min)
		re = ust*z0m/anu
		if(re .le. 0.111) then
			xx = - 2.43
		else
			if(0.111 .lt. re .and. re .le. 16.3) then
				xx = 0.83*log(re) - 0.6
			else
				xx = 0.49 * re**0.45
			end if
		end if
		z0t = amax1(z0m*exp(-xx), z0min)
c         z0t = z0m
		if(abs(dtvir) .lt. 1.e-3) then
			zeta = 1.
			zeta0m = 1.
			zeta0t = 1.
		else
			smo = - ust**3/(alam*vkc*hfl) ! vkc = 0.4
			smo = sign(1.,smo)*amax1(abs(smo),1.e-3) !
			zeta = z/smo
			zeta0m = z0m/smo
			zeta0t = z0t/smo
		end if
		c_u = vkc/amax1( (log(z/z0m) - psi_m(zeta,zeta0m)), 0.1)
		c_t = vkc/amax1( (log(z/z0t) - psi_t(zeta,zeta0t)), 0.1)
      end do
CC         c_t = c_u
      bix(1) = zeta
      bix(2) = alam*z*dtvir/u_a**2
      bix(3) = re
      bix(4) = log(z0m/z0t)
      bix(5) = z0m
      bix(6) = z0t
      bix(7) = 0.
      bix(8) = c_u
      bix(9) = c_t
      bix(10) = vkc*ust*z/phi_m(zeta)
      bix(11) = phi_m(zeta) / phi_t(zeta)

	bix2=dble(bix)

      return
      END                       

      real function phi_m(zeta)
      common /p_drag/ vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     *                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     *                zeta_st,a_m,a_t,b_m,b_t
	real vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     *                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     *                zeta_st,a_m,a_t,b_m,b_t
      real zeta,y,a
	a=1.0

c*   universal function for momentum: businger et al. (1971) formulation
c*   for zeta .ge. zeta_st and large et al. (1994) formulation otherwise
c*   for zeta .ge. 0. : beljaars and holtslag (1991)
      if(zeta .ge. 0.) then
c         y = alpha_m + beta_m*zeta     !  businger et al.
		y = 1. + zeta*(a+0.667*(6.-0.35*zeta)*exp(-0.35*zeta)) ! b&h
      else
		if(zeta. ge. zeta_st) then
			y = (1. - gamma_m*zeta)**(p_m)
		else
			y = (a_m - b_m*zeta)**(q_m)
		end if
      end if
      phi_m = y

      return
      end

      real function phi_t(zeta)
      common /p_drag/ vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     *                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     *                zeta_st,a_m,a_t,b_m,b_t
	real vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     *                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     *                zeta_st,a_m,a_t,b_m,b_t
      real zeta,y,a
	a=1.0

c*   universal function for heat: businger et al. (1971) formulation
c*   for zeta .ge. zeta_st and large et al. (1994) formulation otherwise
c*   for zeta .ge. 0. : beljaars and holtslag (1991)
      if(zeta .ge. 0.) then
c         y = alpha_t + beta_t*zeta
		y = 1. + zeta*(
     &    a*sqrt(1.+2.*zeta*a/3.)+0.667*(6.-0.35*zeta)*exp(-0.35*zeta)) ! b&h
      else
		if(zeta .ge. zeta_st) then
			y = (1. - gamma_t*zeta)**(p_t)
		else
			y = (a_t - b_t*zeta)**(q_t)
		end if
      end if
      phi_t = y

      return
      end

      real function psi_m1(x)

	real x,pi
      pi = 3.1415926
c*   integrated universal function for momentum based on businger et al.
c*   (1971) formulation for zeta_st .le. zeta .le. 0.
c*    p_m = - 1./4. (paulson, 1970)
      psi_m1 = log(0.125*(1. + x**2)*(1. + x)**2) -2.*atan(x) + pi/2.
      return
      end

      real function psi_m2(a,x)
	real x,a
c*   integrated universal function for momentum based on large et al.
c*   (1994) formulation for zeta .le. zeta_st
c*    only for q_m = -1./3.
      psi_m2 = (0.5*log(x**2 + x + 1.)
     &         + sqrt(3.)*atan((2.*x + 1.)/sqrt(3.)))
     &         / sign(1.,a)*abs(a)**(1./3.)
      return
      end

      real function psi_m(zeta,zeta0)
      common /p_drag/ vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     *                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     *                zeta_st,a_m,a_t,b_m,b_t
	real vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     *                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     *                zeta_st,a_m,a_t,b_m,b_t
	real zeta,y,zeta0,x,x_st,psi_m2,psi_m1,x0,x0_st,a
	
	a=1.0
c*   for zeta .ge. 0. : beljaars and holtslag (1991)
      if(zeta .ge. 0.) then
c         y = (1.-alpha_m)*log(zeta/zeta0) - beta_m*(zeta-zeta0)
		y = - (a*zeta+
     &      0.667*(zeta-(5./0.35))*exp(-0.35*zeta)+(0.667*5./0.35))  ! b&h
      else
		if(zeta .ge. zeta_st) then
			x = (1. - gamma_m*zeta)**(- p_m)
			y = psi_m1(x)
		else
			x0 = 1. - (b_m/a_m)*zeta
			x = sign(1.,x0)*abs(x0)**(- q_m)
			x0_st = 1. - (b_m/a_m)*zeta_st
			x_st = sign(1.,x0_st)*abs(x0_st)**(- q_m)
			y = psi_m1(zeta_st) + psi_m2(a_m,x) - psi_m2(a_m,x_st)
		end if
      end if
      psi_m = y
      return
      end

      real function psi_t1(x)
	real x
c*   integrated universal function for heat based on businger et al.
c*   (1971) formulation for zeta_st .le. zeta .le. 0.
c*    p_t = - 1./2. (paulson, 1970)
      psi_t1 = log(0.25*(1. + x)**2) !in (paulsen, 1970) log(0.25*(1. + x**2)**2)!
      return
      end

      real function psi_t(zeta,zeta0)
      common /p_drag/ vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     *                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     *                zeta_st,a_m,a_t,b_m,b_t
	real vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     *                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     *                zeta_st,a_m,a_t,b_m,b_t
	real zeta,zeta0,y,x,x0,psi_t1,x0_st,x_st,psi_m2,a
	a=1.0
c*   for zeta .ge. 0. : beljaars and holtslag (1991)
      if(zeta .ge. 0.) then
c         y = (1.-alpha_t)*log(zeta/zeta0) - beta_t*(zeta-zeta0)
		y = - (a*(1.+(2./3.)*zeta*a)**(1.5)
     &   + 0.667*(zeta-(5./0.35))*exp(-0.35*zeta)+((0.667*5./0.35)-1.))
      else
		if(zeta .ge. zeta_st) then
			x = (1. - gamma_t*zeta)**(- p_t)
			y = psi_t1(x)
		else
			x0 = 1. - (b_t/a_t)*zeta
			x = sign(1.,x0)*abs(x0)**(- q_t)
			x0_st = 1. - (b_t/a_t)*zeta_st
			x_st = sign(1.,x0_st)*abs(x0_st)**(- q_t)
			y = psi_t1(x_st) + psi_m2(a_t,x) - psi_m2(a_t,x_st)
		end if
      end if
      psi_t = y
      return
      end
      
      SUBROUTINE KT_eq(dt)
      
!     KT_eq calculates eddy diffusivity (ED) in water coloumn following parameterizations:

!     1) empirical profile of ED;
!     2) Semi-empirical formulations for ED;
!     2) k-epsilon parameterization for ED, including kolmogorov relation.
      
	use atmos
    	use driving_params
	use arrays
      implicit none
		
	real(8), dimension(350):: a,b,c,d
	real(8), dimension(350,2,2)::am,bm,cm
	real(8), dimension(350,2)::ym,dm
	real(8) alp(10)
!	real(8) sbix(11)
	real(8) nyear,nmonth,nday,nhour
	real(8) dtdz0,KC0,wind10,
     & urel,vrel,u,v,CL1,CL2,roughness,l0,b0,wr
	real(8) h1,l1,hs1,ls1
	real(8) pi,CE,row0,cw,g,lam_eps,ufr,
     & niu,ACC,month,dt,kar,Cs,Cs1,dudz2dvdz2,
     & lam_E,Co,taux,tauy,lam_gen,dE_it,a0,a1,a2,a3,lam_T,
     & deps_it,tau_gr,tau_i,mi,ma,drodt1,drodt2,E_mid,eps_mid,
     & kor2,Cz_gr,coef1,coef2,rp,Cz_i,tau_sbl,tau_wav,
     & xx,yy,tau2,lm,ext_lamw,lamw0,day,hour,ceps1,ceps2,
     & Conuv,al_it,kron,FS,FB,al,alft,GAMUN,zup,lamTM,
     & CEt,sigmae,sigmaeps,urels,vrels,minwind,delta0
   
        
	real(8) AG(5),GRADU,GRADV,GRADT,GAMVN,CON0,TFR,COGRTN,COGRUN,
     &COGRVN,DTDZH,CON1,CON2
	real(8), external:: dzeta
	integer(4) iter,iter_t,
     & nl,i,j,k,numEeps
     	integer(4) psi(10)
	logical indstab,ind_bound,iterat,firstcall
	character outpath*60,tp_name*10,numt*1
	
  	common /ymdh/ nyear,nmonth,nday,nhour
	common /wind/  wind10,urel,vrel,u,v
	common /out/   outpath
	common /layers/ h1,l1,hs1,ls1

	data firstcall /.true./
	data psi/6,3,7,2,5,4,8,1,0,0/ !/3,2,4,1,0,0,0,0,0,0/  

	SAVE

      if (firstcall) then
	 if (turb_out==1) then
	  write (numt,'(i1)') Turbpar
	  open(2114,file=path(1:len_trim(path))//'results/'//
     &                          'err_progon.dat',  status = 'unknown')
	  open(2115,file=path(1:len_trim(path))//'results/'//'E-eps.dat', 
     & 										    status = 'unknown')
 	   open(2116,file=path(1:len_trim(path))//'results/'//'E-eps1.dat',
     & 										    status = 'unknown')
	  open (2117,file=path(1:len_trim(path))//'results/'
     &   //'turb'//numt//'.dat',	    status = 'unknown')
  	  open (2118,file=path(1:len_trim(path))//'results/'
     &   //'temp'//numt//'.dat',	    status = 'unknown')
        open (2119,file=path(1:len_trim(path))//'results/'
     &   //'uv'//numt//'.dat',	    status = 'unknown')
        open (2120,file=path(1:len_trim(path))//'results/'
     &   //'KT'//numt//'.dat',	    status = 'unknown') 
	  open (2121,file=path(1:len_trim(path))//'results/'
     &   //'dE1'//numt//'.dat',	    status = 'unknown') 
	  open (2122,file=path(1:len_trim(path))//'results/'
     &   //'dE2'//numt//'.dat',	    status = 'unknown') 
	  open (2123,file=path(1:len_trim(path))//'results/
     &   err_Eeps_iter.dat',	        status = 'unknown')  
	 endif
	 
	 month = 30.*24.*60.*60.
	 day = 24*60.*60.
	 hour = 60.*60.
	 pi = 4.*dtan(1.d0)
!PHYSICAL CONSTANTS
       row0 = 1000.
	 lamw0 = 0.561
	 cw = 3990.
	 g = 9.814 
	 kar = 0.4
!	 kor = 1.26*10.**(-4) !(-4) !Koriolis parameter for lat=60 deg 
	 niu = 1.007d-6
	 roughness = 0.01 !water roughness	       
!DIMENSIONLESS CONSTANTS IN E-EPS PARAMETERIZATION, according to Goudsmit et al., 2002
       CE = 0.09 !0.09
	 CEt = 0.072
	 ceps1 = 1.44
	 ceps2 = 1.92
	 sigmae = 1.
	 sigmaeps = 1.3
!	Ce_a=1. !4. !4.
!	Ce_i=1. !4. !1.
!	Ce_gr=1. !4. !3.
!	C2_a=Ce_a**(3./2.)  !*CE
!	C2_i=Ce_i**(3./2.)  !*CE
!	C2_gr=Ce_gr**(3./2.) !*CE 
	 lam_E=CE/sigmaE ! 0.73
	 lam_T=1. !0.05
	 lam_eps=CE/sigmaeps ! 0.73
	 lam_gen= 1. 
	 lamTM = 0.14
!Co is according to Satyanarayana et al., 1999
	 Co=1.9 !? 1.9 
!CONSTANTS IN CONVECTION PARAMETERIZATION
	 KC0=0. !1.d+5 !500.
	 dtdz0=2.
!CONSTANTS IN LENGMUIR PARAMETERIZATION
       CL1=60. !120./5. !120.
	 CL2=2./5.
!CONSTANTS OF SIMOES'S PARAMETERIZATION
	 Cs = 1.5  !0.15
	 Cs1 = 100.	
	
!	 lo=0.4*ddz*h1*3.                                                         VS,06.2007
	 L0=0.1
		
!AUXILIARY PHYSICAL CONSTANTS	
	 drodt1=-0.112
	 drodt2=0.027 !-2.*drodt1 !0.02
	 CONUV = 1.
		
!PARAMETERS OF NUMERICAL SCHEME
	 AG(1)=0. !0.
	 AG(2)=0. !0.
	 AG(3)=0. !0.
	 !AG(4)=1. !0.
	 !AG(5)=1. !0.
	 CON0 = 0.05
	 CON1 = 0.11
       CON2 = 1.56
	 al_it=0.2
	 ma=20.5
	 mi=1.
	 ACC=1.d-10
	 knum = 0.
       minwind = 1.0d0
!       minwind = 1.0d-1

	 numEeps = 1
	 delta0 = 2.d-2
	
	 do i=1,8 
	  alp(i)=2*(mi+ma-(ma-mi)*dcos((2*psi(i)-1)/16*pi))**(-1)
       enddo

	 a0=800.969d-7
	 a1=588.194d-7
	 a2=811.465d-8
	 a3=476.600d-10
	 
!FRICTION: VEGETATION

      if (h_veg>0) then
	 print*, 'Vegetation friction parameterization
     & is not operational currently: STOP'
       STOP
	endif
     
      veg_sw = 0.
      do i = 1, M+1
       if (h1-dzeta(dble(i))*h1 < h_veg) veg_sw(i) = 1.
	enddo
	
      endif

	iterat=.true.
	kor2=kor*dt/2

	u=u1(1)
	v=v1(1)
	
	if (relwind==2) then
	 urel=uwind-u
	 vrel=vwind-v
	elseif (relwind==1) then
       urel=uwind
	 vrel=vwind
	else
	 print*, 'relwind=',relwind,' - incorrect meaning: STOP' 
	 stop
      endif
	
	if (dabs(urel)<minwind) then
!	 urels=urel/dabs(urel)*minwind
       urels=sign(minwind,urel)
!      elseif (dabs(urel)>1.) then
!       urels=urel/dabs(urel)*1.
	else
	 urels=urel
      endif

	if (dabs(vrel)<minwind) then
!	 vrels=vrel/dabs(vrel)*minwind
	 vrels=sign(minwind,vrel)
!      elseif (dabs(vrel)>1.) then
!       vrels=vrel/dabs(vrel)*1.
	else   
       vrels=vrel
	endif
	
	wr=dsqrt(urels**2+vrels**2) !,1.d-1)
	tau2=1.273*cdmw2*wr !wind      
      ufr=dsqrt(tau2/row0)	
	do i = 1, M+1
!	 dT4=Tw1(i)-4.
!	 if (dT4/=0.) then 
!	  drodt=(0.5*drodt1*(dabs(dT4)+dT4)+0.5*drodt2*(dT4-dabs(dT4)))
!     &  /dT4
!	 row(i)=1000.+drodt*dT4 
!	 else
!	  row(i)=1000. 
!       endif
!	 T(i)=20.
       row(i) = row0*(1+a0+a1*Tw1(i)-a2*(Tw1(i)**2)+a3*(Tw1(i)**3))
	enddo  

	KC=0                                                                           VS,06.2007
	KCu=0                                                                          VS,06.2007

	!PARAMETERIZATION OF WIND STRESS SPLIT-UP INTO TWO PARTS:
! tau=tau_sbl+tau_wave
      ! if (wind10>5.) then
      !  C10=1.26*10**(-3.)
      ! else
	!  C10=0.0044*wind10**(-1.15)
      ! endif
!a.PARAMETERIZATION USNG SIGNIFICANT WAVE HEIGHT (NB1, pp. 8)
	!co1=4.*kws/sqrt(C10*1000.*g) !1000 m is "average fetch" of lake Syrdakh
	!kw=min(max(co1*wind10,kws),0.75)
!b.PARAMETERIZATION USING WAVE AGE (agw) (NB1, pp. 9)
	!agw=0.14*(g*1000.)**(1./3.)*C10**(1./6.)*wind10**(-2./3.) 
	!kw=min(max(kws*1.15/agw,kws),0.75)
	!KW=0.75
	tau_wav=tau2 !*kw
	tau_sbl=tau2 !*(1.-kw)
!CALCULATION SHEZY COEFFICIENT
	rp=0.1 !height of roughness elements
     	if (l1/=0) then
	 Cz_gr=7.7*dsqrt(g)*(h1/2/rp)**(1./6.) !50. !ground Shezy coefficient
	else
	 Cz_gr=7.7*dsqrt(g)*(h1/rp)**(1./6.)
	endif
	rp=0.01 
      Cz_i=7.7*dsqrt(g)*(h1/2/rp)**(1./6.)

!FRICTION: GROUND
	tau_gr=row0*g*Cz_gr**(-2)*(u1(M)**2+v1(M)**2)
!     tau_grx=row0*g*Cz_gr**(-2)*u1(M)*dsqrt((u1(M)**2+v1(M)**2)) !*0.005
!     tau_gry=row0*g*Cz_gr**(-2)*v1(M)*dsqrt((u1(M)**2+v1(M)**2)) !*0.005
!FRICTION: ICE
	tau_i=row0*g*Cz_i**(-2)*(u1(1)**2+v1(1)**2)
!     tau_ix=row0*g*Cz_i**(-2)*u1(1)*dsqrt(u1(1)**2+v1(1)**2)
!     tau_iy=row0*g*Cz_i**(-2)*v1(1)*dsqrt(u1(1)**2+v1(1)**2)
	coef1=-g*Cz_i**(-2)*dsqrt(u1(1)**2+v1(1)**2)
	coef2=g*Cz_gr**(-2)*dsqrt(u1(M)**2+v1(M)**2) !*0.005 !*row0
      
!________EDDY VISCOSITY PARAMETERIZATION__________________________________
      do i=1,M+1
	 uv(i)=dsqrt(u1(i)**2+v1(i)**2)
	enddo
	do i=1,M
!	 Ri(i)=dmin1(dmax1(g/row0/(((uv(i+1)-uv(i))/
!     & (ddz*h1))**2)*(row(i+1)-row(i))/(ddz*h1),-10.d0),10.d0)
      dudz2dvdz2 = dmax1( ( (u1(i+1)-u1(i))/(ddz(i)*h1) )**2 +
     & ( (v1(i+1)-v1(i))/(ddz(i)*h1) )**2, 1.d-1)
	 Ri(i) = g/row0*(row(i+1)-row(i))/(ddz(i)*h1)/dudz2dvdz2
      enddo

      select case (Turbpar)
	 
!     1. "Empirical" parametrization: Stepanenko, Lykossov (2005)
	 case (1)
	 tp_name='SL_Empir' 
       k2(1) =(lamw0*10.+(wind*2./zref/20.)*(lamw0*1000.-lamw0*10.))/
     & (cw*row0)
	 ext_lamw = dlog(k2(1)*cw*row0/(lamw0*10.))/h1
	 do i=2,M+1
	  k2(i) = k2(1)*dexp(-ext_lamw*dzeta(dble(i+0.5))*h1)+niu
	 enddo
	 k2(1)=k2(1)+niu

!     2. "E-epsilon" parameterization: k=E**2/eps with 
!         prognostic equations for E and eps
	 case (2)
	 do i=1,M+1
	  if (E1(i)<=0) E1(i)=10.**(-16)
	  if (eps1(i)<=0) eps1(i)=10.**(-18)
	 enddo
       do i = 1, M
	  k2(i)=dmax1(CE*E1(i)**2/(eps1(i)+ACC),1.d-6)+niu+knum(i)
	 end do
	 continue

!     3. Nickuradze (NICK) formulation: Rodi (1993)
	 case (3)
	 tp_name='NICK'
	 do i=1,M
	  zup=h1-dzeta(dble(i+0.5))*h1
	  lm=h1*(0.14-0.08*(1-zup/h1)**2-0.06*(1-zup/h1)**4)
	  k2(i)=lm**2*dabs((uv(i+1)-uv(i))/(ddz(i)*h1))*dexp(-Cs*Ri(i))           VS,06.2007   
     &   +niu                                                                   VS,06.2007
	 enddo
	
!     4. Parabolic (PARAB) formulation: Engelund (1976)
	 case (4)
	 tp_name='PARAB'
       do i=1,M
	  zup=h1-dzeta(dble(i+0.5))*h1
	  k2(i)=kar*ufr*zup*(1-zup/h1)*dexp(-Cs*Ri(i))
     &  +niu
	 enddo 

!     5. W2 (used in Version 2 of CE-QUAL-W2 model): Cole and Buchak (1995)	
	 case (5)

!     6. W2N (W2 with mixing length of Nickuradze): Cole and Buchak (1995) and Rodi (1993)
	 case (6)
    
!     7. RNG (re-normalization group) formulation: Simoes (1998)
       case (7)
	 tp_name='RNG'
	 do i=1,M
	  zup=h1-dzeta(dble(i+0.5))*h1
	  k2(i)=niu*(1+dmax1(3*kar*(zup/niu*ufr)**3*
     &  (1-zup/h1)**3-Cs1,0.d0))**(1./3.)*dexp(-Cs*Ri(i))+niu  
       enddo
	end select
	  
	 k2=k2+KCu !+KLengu
	
!(U,V) - EQUATION
       !xx=1+ddz**2*h1*dhw/(k2(1)*dt)
	 xx=0.5
	 yy=(ddz(1)*h1)**2/(k2(1)*dt)
!UPPER BOUNDARY CONDITION FOR (U,V)-EQUATION.
!1 CASE:WATER SURFACE FREE FROM ICE (ATMOSPHERIC MOMENTUM FLUX ONLY)
	 if (l1==0) then
	  taux=urels/wr*tau_sbl/row0 !wdir+pi
	  tauy=vrels/wr*tau_sbl/row0 !wdir+pi
!	  if(dabs(xx+yy)<dabs(xx).and.debug_out==1)
!     &   write(2114,*) '1',nyear,nmonth,nday,time/month
	  do i=1,2
	   do j=1,2
	    cm(1,i,j)=kron(i,j)*(xx+yy
     &	   +0.5*yy*dt*a_veg*c_veg*dsqrt(u1(1)**2+v1(1)**2)
     &     *veg_sw(1)) !xx
	    bm(1,i,j)=kron(i,j)*xx
         enddo
        enddo
	  cm(1,1,2)=-kor*(h1*ddz(1))**2/(2*k2(1))
        cm(1,2,1)=kor*(h1*ddz(1))**2/(2*k2(1))
	  dm(1,1)=(taux*ddz(1)*h1/k2(1)+yy*u1(1)) 
	  dm(1,1)=dm(1,1)+(u1(2)-u1(1))*xx                                        VS,06.2007
     &   +kor*(h1*ddz(1))**2*v1(1)/(2*k2(1))                                    VS,06.2007
     &   +yy*dt*g*tan(pi*alphax/180.)                                           VS,06.2007
   	  dm(1,2)=(tauy*ddz(1)*h1/k2(1)+yy*v1(1))                                 VS,06.2007
	  dm(1,2)=dm(1,2)+(v1(2)-v1(1))*xx                                        VS,06.2007  
     &   -kor*(h1*ddz(1))**2*u1(1)/(2*k2(1))                                    VS,06.2007
     &   +yy*dt*g*tan(pi*alphay/180.)
       endif
!2 CASE: THIN ICE (COMBINATION OF ATMOSPHERIC MOMENTUM FLUX AND WATER-ICE FRICTION)
	 if (l1>0.and.l1<=L0) then 
	  b0=l1/L0
	  taux=urels/wr*tau_sbl/row0 !wdir+pi
	  tauy=vrels/wr*tau_sbl/row0 !wdir+pi
!	  if(dabs(xx+yy-2*b0*coef1*ddz*h1/k2(2))<dabs(xx)
!     &  .and.debug_out==1)
!     &   write(2114,*) '2',nyear,nmonth,nday,time/month 
	  do i=1,2
	   do j=1,2
	    cm(1,i,j)=kron(i,j)*(xx+yy-b0*coef1*ddz(1)*h1/k2(1)                   VS,06.2007
     &	   +0.5*yy*dt*a_veg*c_veg*dsqrt(u1(1)**2+v1(1)**2)
     &     *veg_sw(1))
	    bm(1,i,j)=kron(i,j)*xx
	   enddo
        enddo
	  cm(1,1,2)=-kor*(h1*ddz(1))**2/(2*k2(1))                                 VS,06.2007
        cm(1,2,1)=kor*(h1*ddz(1))**2/(2*k2(1))                                  VS,06.2007  
	  dm(1,1)=yy*u1(1) !-tau_ix*ddz*h1/k2(1)
        dm(1,1)=dm(1,1)+(u1(2)-u1(1))*xx+2*taux*ddz(1)*h1/k2(1)*(1-b0)          VS,06.2007
     &  +kor*(h1*ddz(1))**2*v1(1)/(2*k2(1))+yy*dt*g*tan(pi*alphax/180.)         VS,06.2007
	  dm(1,2)=yy*v1(1) !-tau_iy*ddz*h1/k2(1)
        dm(1,2)=dm(1,2)+(v1(2)-v1(1))*xx+2*tauy*ddz(1)*h1/k2(1)*(1-b0)          VS,06.2007
     &  -kor*(h1*ddz(1))**2*u1(1)/(2*k2(1))+yy*dt*g*tan(pi*alphay/180.)         VS,06.2007 
	! do i=1,2
	!  do j=1,2
	!   cm(1,i,j)=dble(kron(i,j))
	!   bm(1,i,j)=0
	!  enddo
      ! enddo
	! dm(1,1)=0
	! dm(1,2)=0
       endif
!3 CASE: THICK ICE (WATER-ICE FRICTION ONLY)
	 if (l1>L0) then
!	  if(dabs(xx+yy-2*coef1*ddz*h1/k2(2))<dabs(xx).and.debug_out==1)
!     &   write(2114,*) '3',nyear,nmonth,nday,time/month 
	  do i=1,2
	   do j=1,2
	    cm(1,i,j)=kron(i,j)*(xx+yy-coef1*ddz(1)*h1/k2(1)                      VS,06.2007
     &	   +0.5*yy*dt*a_veg*c_veg*dsqrt(u1(1)**2+v1(1)**2)
     &     *veg_sw(1))
	    bm(1,i,j)=kron(i,j)*xx
	   enddo
        enddo
	  cm(1,1,2)=-kor*(h1*ddz(1))**2/(2*k2(1))                                 VS,06.2007 
        cm(1,2,1)=kor*(h1*ddz(1))**2/(2*k2(1))                                  VS,06.2007
	  dm(1,1)=yy*u1(1) !-tau_ix*ddz*h1/k2(1)
        dm(1,1)=dm(1,1)+(u1(2)-u1(1))*xx
     &   +kor*(h1*ddz(1))**2*v1(1)/(2*k2(1))                                    VS,06.2007
     &   +yy*dt*g*tan(pi*alphax/180.)
	  dm(1,2)=yy*v1(1) !-tau_iy*ddz*h1/k2(1)
        dm(1,2)=dm(1,2)+(v1(2)-v1(1))*xx
     &   -kor*(h1*ddz(1))**2*u1(1)/(2*k2(1))                                    VS,06.2007
     &   +yy*dt*g*tan(pi*alphay/180.)
	 ! do i=1,2
	 !  do j=1,2
	 !   cm(1,i,j)=dble(kron(i,j))
	 !   bm(1,i,j)=0
	 !  enddo
       ! enddo
	 ! dm(1,1)=0
	 ! dm(1,2)=0
	 endif
	 xx=0.5*(1-ddz(M)*h1*dhw/(k2(M)*dt))
	 yy=(ddz(M)*h1)**2/(k2(M)*dt)
!	 if(dabs(xx+yy+2*coef2*ddz*h1/k2(M))<dabs(xx).and.debug_out==1)
!     & write(2114,*) '3',nyear,nmonth,nday,time/month
	  do i=1,2
	   do j=1,2
	    !if (i==1) then
	    ! coef=coef2*(u1(M))/dabs(u1(M))
          !else
	    ! coef=coef2*(v1(M))/dabs(v1(M))
          !endif
	    cm(M+1,i,j)=kron(i,j)*(xx+yy+coef2*ddz(M)*h1/k2(M)                    VS,06.2007
     &	   +0.5*yy*dt*a_veg*c_veg*dsqrt(u1(M+1)**2+v1(M+1)**2)
     &     *veg_sw(M+1) )
	    am(M+1,i,j)=kron(i,j)*xx
	   enddo
        enddo
	  cm(M+1,1,2)=-kor*(h1*ddz(M))**2/(2*k2(M))                               VS,06.2007
        cm(M+1,2,1)=kor*(h1*ddz(M))**2/(2*k2(M))                                VS,06.2007
	  dm(M+1,1)=yy*u1(M+1) !-tau_grx*ddz*h1/k2(M)
	  dm(M+1,1)=dm(M+1,1)-(u1(M+1)-u1(M))*xx
     &   +kor*(h1*ddz(M))**2*v1(M+1)/(2*k2(M))                                  VS,06.2007
     &   +yy*dt*g*tan(pi*alphax/180.)
	  dm(M+1,2)=yy*v1(M+1) !-tau_gry*ddz*h1/k2(M)
	  dm(M+1,2)=dm(M+1,2)-(v1(M+1)-v1(M))*xx
     &   -kor*(h1*ddz(M))**2*u1(M+1)/(2*k2(M))                                  VS,06.2007
     &   +yy*dt*g*tan(pi*alphay/180.)
	 ! do i=1,2
	 !  do j=1,2
	 !   cm(M,i,j)=dble(kron(i,j))
	 !   am(M,i,j)=0
	 !  enddo
       ! enddo
	 ! dm(M,1)=0
	 ! dm(M,2)=0
	 do k=2,M
	  do i=1,2
	   do j=1,2
	    !a(i)=(i-1)*dhw/(2*h1)-k2_mid(i-1)*dt/(h1**2*ddz**2)
	    !b(i)=-(i-1)*dhw/(2*h1)-k2_mid(i)*dt/(h1**2*ddz**2)
	    !c(i)=a(i)+b(i)-1
	    !d(i)=-u1(i)
         am(k,i,j)=kron(i,j)*(dzeta(dble(k))*dhw/                               VS,06.2007
     &   (2*(ddz(k-1)+ddz(k))*h1)                                               VS,06.2007
     &   -k2(k-1)*dt/(h1**2*(ddz(k)+ddz(k-1))*ddz(k-1) ))                       VS,06.2007
         bm(k,i,j)=-kron(i,j)*(dzeta(dble(k))*dhw/                              VS,06.2007
     &   (2*(ddz(k-1)+ddz(k))*h1)                                               VS,06.2007
     &   +k2(k)*dt/(h1**2*(ddz(k)+ddz(k-1))*ddz(k) ))                           VS,06.2007
	   enddo
	  enddo
	  cm(k,1,1)=-(k2(k)/ddz(k)+k2(k-1)/ddz(k-1))*dt/
     &   (h1**2*(ddz(k)+ddz(k-1)) )-1                                           VS,06.2007
     &   -0.5*dt*c_veg*a_veg*dsqrt(u1(k)**2+v1(k)**2)*veg_sw(k)                 VS,06.2007
	  cm(k,1,2)=kor2 !Krank-Nickolson
!       cm(k,1,2)=kor2*2. !implicit scheme
!	  cm(k,1,2)=0. !explicit scheme
	  cm(k,2,1)=-cm(k,1,2)
	  cm(k,2,2)=cm(k,1,1)
	  dm(k,1)=-u1(k)-kor2*v1(k)-dt*g*tan(pi*alphax/180.) !Krank-Nickolson
	  dm(k,1)=dm(k,1)-dt*(h1**2*(ddz(k)+ddz(k-1))*ddz(k) )**(-1)*             VS,06.2007
     &   (k2(k)*(u1(k+1)-u1(k)))
	  dm(k,1)=dm(k,1)+dt*(h1**2*(ddz(k)+ddz(k-1))*ddz(k-1) )**(-1)*           VS,06.2007
     &   (k2(k-1)*(u1(k)-u1(k-1)))
	  dm(k,1)=dm(k,1)-dzeta(dble(k))*dhw*(2*(ddz(k)+ddz(k-1) )*h1)            VS,06.2007
     &   **(-1)*(u1(k+1)-u1(k-1))                                               VS,06.2007
	  dm(k,2)=-v1(k)+kor2*u1(k)-dt*g*tan(pi*alphay/180.) !Krank-Nickolson
	  dm(k,2)=dm(k,2)-dt*(h1**2*(ddz(k)+ddz(k-1))*ddz(k) )**(-1)*             VS,06.2007
     &   (k2(k)*(v1(k+1)-v1(k)))
	  dm(k,2)=dm(k,2)+dt*(h1**2*(ddz(k)+ddz(k-1))*ddz(k-1) )**(-1)*           VS,06.2007
     &   (k2(k-1)*(v1(k)-v1(k-1)))
	  dm(k,2)=dm(k,2)-dzeta(dble(k))*dhw*(2*(ddz(k)+ddz(k-1) )*h1)            VS,06.2007
     &   **(-1)*(v1(k+1)-v1(k-1))
!       dm(k,1)=-u1(k)  !implicit scheme
!	  dm(k,2)=-v1(k)  !implicit scheme
!	  dm(k,1)=-u1(k)-kor2*2*v1(k)  !explicit scheme
!	  dm(k,2)=-v1(k)+kor2*2*u1(k)  !explicit scheme
	 enddo
!      ind_bound=.false.
!	 call ind_stab_fact_db (a,b,c,1,M+1,indstab,ind_bound)
!	 if (indstab==.false.) then
!	  do i=2,M
!	   a(i)=-k2(i)*dt/(h1**2*ddz**2)
!	   b(i)=-k2(i+1)*dt/(h1**2*ddz**2)
!	  enddo
!	 endif  
       call MATRIXPROGONKA(am,bm,cm,dm,ym,M+1)
	 do k=1,M+1
	  u2(k)=ym(k,1)
	  v2(k)=ym(k,2)
	 enddo
	 
!	 print*, (u2(1)-u1(1))/dt*(h1*ddz)**2/k2(1)
!     & -coef1*u2(1)*h1*ddz/k2(1)
!     & - 0.5*(u2(2)-u2(1))-0.5*(u1(2)-u1(1)) -     
!     & 0.5*dhw*h1*ddz/(k2(1)*dt)*(u2(2)-u2(1)) -
!     & 0.5*dhw*h1*ddz/(k2(1)*dt)*(u1(2)-u1(1)) -
!     & kor*(ddz*h1)**2/k2(1)*(v2(1)+v1(1))/2.	 
      
!      do i=2,M
!      print*, (v2(i)-v1(i))/dt
!     &   -0.5/h1/ddz*(k2(i)*(v2(i+1)-v2(i))/h1/ddz+
!     &   k2(i)*(v1(i+1)-v1(i))/h1/ddz -
!     &   k2(i-1)*(v2(i)-v2(i-1))/h1/ddz -
!     &   k2(i-1)*(v1(i)-v1(i-1))/h1/ddz)-
!     &   0.25*(i-1)/h1*dhw/dt*(v2(i+1)-v2(i-1))-
!     &   0.25*(i-1)/h1*dhw/dt*(v1(i+1)-v1(i-1))+
!     &   kor*(u1(i)+u2(i))/2.
!      enddo
	 
!	 print*, (u2(M+1)-u1(M+1))/dt*(h1*ddz)**2/k2(M)
!     & +coef2*u2(M+1)*h1*ddz/k2(M)
!     & + 0.5*(u2(M+1)-u2(M))+ 0.5*(u1(M+1)-u1(M)) -
!     & 0.5*dhw*h1*ddz/(k2(M)*dt)*(u2(M+1)-u2(M)) -
!     & 0.5*dhw*h1*ddz/(k2(M)*dt)*(u1(M+1)-u1(M)) -
!     & kor*(ddz*h1)**2/k2(M)*(v2(M+1)+v1(M+1))/2.	 

!	select case (Turbpar)
!	case (2)
!	case default
!	goto 1
!     end select
      
      ifkeps: if (Turbpar==2) then
      
	if (numEeps==1) then

!Krank-Nikolson numerical scheme for k-eps parameterization
!using simple iterations

	Um=(u2+u1)/2
      Vm=(v2+v1)/2
      
!SOLUTION OF 'E-EPS' PARAMETERIZATION EQUATIONS	
       E_it1=E1
	 eps_it1=eps1
	 iter_t=0
	 k2_mid=0.

	 AL=g/row0 !*drodt
	 DTDZH=(Tw1(2)-Tw1(1))/(h1*ddz(1))
	 !dT_sb=T(M+1)-T(1)	 
       !ALFT=(0.5*30.*(dabs(dT_sb)+dT_sb)+0.5*30.*(dT_sb-dabs(dT_sb)))
     & !/dT_sb
	 !if (l1/=0) ALFT=1.
	 ALFT=1. !1.
 		
10     E12=(E1+E_it1)/2
	 eps12=(eps1+eps_it1)/2
      		
	 do i=1,M
        !E_mid=(E_it1(i)+E_it1(i+1))/2.
	  E_mid=(E12(i)+E12(i+1))/2.
	  !eps_mid=(eps_it1(i)+eps_it1(i+1))/2. 
	  eps_mid=(eps12(i)+eps12(i+1))/2. 
        !k2_mid(i)=dmax1(CE*E_mid*E_mid/(eps_mid+ACC),1.E-6)
	  k2_mid(i)=k2_mid(i)*al_it+
     &  dmax1(E_mid*E_mid/(eps_mid+ACC),1.d-6)*(1-al_it)
	  k5_mid(i)=k2_mid(i)*lam_eps
	  k3_mid(i)=k2_mid(i)*lam_E
!	  k2(i)=dmax1(CE*E12(i)*E12(i)/(eps12(i)+ACC),1.d-6)
	  k2(i)=k2(i)*al_it+
     &  dmax1(E12(i)*E12(i)/(eps12(i)+ACC),1.d-6)*(1-al_it)
!	  k2t(i)=k2(i)/CE*CEt
	  k5(i)=k2(i)*lam_eps
	  k3(i)=k2(i)*lam_E
	  !l(i)=CE*E_it1(i)**(3./2.)/(eps_it1(i)+ACC)
	  l(i)=E12(i)**(3./2.)/(eps12(i)+ACC)
	  !TF(i)=dSQRT(E_it1(i))/(l(i)+ACC)
	  TF(i)=dSQRT(E12(i))/(l(i)+ACC)
	 enddo
	  !k2(M+1)=dmax1(CE*E_it1(M+1)*E_it1(M+1)/(eps_it1(M+1)+ACC),1.E-8)
	  !k2(M)=dmax1(CE*E12(M)*E12(M)/(eps12(M)+ACC),1.E-8)
	  !k5(M)=k2(M)
	  !k3(M)=k2(M)

	 do i=2,M-1
        WU_(i)=(E_it1(i+1)-E_it1(i-1))*(U2(i)-U2(i-1))/
     &   (h1**2*(0.5*ddz(i-1)+0.5*ddz(i+1)+ddz(i))*ddz(i-1) )                   VS,06.2007
        WV_(i)=(E_it1(i+1)-E_it1(i-1))*(V2(i)-V2(i-1))/
     &   (h1**2*(0.5*ddz(i-1)+0.5*ddz(i+1)+ddz(i))*ddz(i-1) )                   VS,06.2007
	 enddo
	 
       do i=2,M
	  GRADT = (Tw1(i+1)-Tw1(i-1))/(h1*(ddz(i-1)+ddz(i) ) )                    VS,06.2007
        GRADU = (U2(i)-U2(i-1))/(h1*ddz(i-1) )                                  VS,06.2007
        GRADV = (V2(i)-V2(i-1))/(h1*ddz(i-1) )                                  VS,06.2007
        GAMUN = - CONUV*(WU_(i+1)-WU_(i-1))/(h1*(ddz(i-1)+ddz(i) ))             VS,06.2007
        GAMVN = - CONUV*(WV_(i+1)-WV_(i-1))/(h1*(ddz(i-1)+ddz(i) ))             VS,06.2007
	 COGRTN = DTDZH * 0.1
	 TFR=E_it1(i)/(l(i)*l(i)+ACC) + acc
	 COGRUN = GAMUN/TFR
       COGRVN = GAMVN/TFR 
	 if(gradT < 0.) then
        gamt(i) = 0.
        gamu(i) = 0.
        gamv(i) = 0.
	  gamt(i) = - AG(3)*cogrtn
        gamu(i) = - AG(1)*cogrun
        gamv(i) = - AG(2)*cogrvn
       else
	    GAMT(i) = - AG(3)*(AL*GRADT**2+CON0*TFR*COGRTN)
     &              /(AL*GRADT+CON0*TFR)
          GAMU(i) = - AG(1)*(AL*(CON2*GRADT+(CON2-1.)*GAMT(i))*GRADU +
     &                CON1*TFR*COGRUN)/(AL*GRADT+CON1*TFR)
          GAMV(i) = - AG(2)*(AL*(CON2*GRADT+(CON2-1.)*GAMT(i))*GRADV +
     &                CON1*TFR*COGRVN)/(AL*GRADT+CON1*TFR)
       end if
	 enddo
	 GAMT(1)=0.;GAMU(1)=0.;GAMV(1)=0.

	 do i=1,M
	  Gen(i)=((Um(i+1)-Um(i))*(Um(i+1)-Um(i)
     &   +h1*ddz(i)*GAMU(i))+                                                   VS,23.06.2007
     &   (Vm(i+1)-Vm(i))*(Vm(i+1)-Vm(i)
     &   +h1*ddz(i)*GAMV(i)))/                                                  VS,23.06.2007
     &   (h1**2*ddz(i)**2)*CE                                                   VS,23.06.2007
        S(i)=-((row(i+1)-row(i))/(h1*ddz(i))+GAMT(i))*ALFT*AL*CEt               VS,23.06.2007
	  F(i)=Gen(i)+S(i)
	 enddo
      
!E-EQUATION SOLUTION      
	 if (l1==0) then 
	  FS=kwe*(tau2/row0)**(3./2.)
	 elseif (l1>0.and.l1<=L0) then
	  FS=((b0*tau_i+(1-b0)*tau2)/row0)**(3./2.)
	 elseif (l1>L0) then
	  FS=(tau_i/row0)**(3./2.)
       endif
	 xx=0.5*(k3_mid(1)*dt/(h1**2*ddz(1)*0.5*(ddz(1)+ddz(2)) )                 VS,23.06.2007
     &  +0.5*ddz(1)*dhw/(h1*0.5*(ddz(1)+ddz(2)) ) )                             VS,23.06.2007
	 yy=dt/(h1*ddz(1))                                                        VS,23.06.2007
	 b(1)=-xx
	 c(1)=-(xx+1)-(dABS(S(1))-S(1))/(TF(1)+ACC)*dt/2-
     & TF(1)*dt	  
	 d(1)=-E1(1)-xx*(E1(2)-E1(1))-yy*FS-
     & k2(1)*(S(1)+dabs(S(1)))*dt/2-k2(1)*Gen(1)*dt !+eps_it1(i)*dt
	 
	 FB=-(tau_gr/row0)**(3./2.)
	 xx=0.5*(k3_mid(M-1)*dt/(h1**2*ddz(M)*0.5*(ddz(M)+ddz(M-1)) )-            VS,23.06.2007
     & dzeta(dble(M+0.5))*dhw/(h1*0.5*(ddz(M)+ddz(M-1)) ) )                     VS,23.06.2007
	 yy=dt/(h1*ddz(M))                                                        VS,23.06.2007
	 a(M)=-xx
	 c(M)=-(xx+1)-(dABS(S(M))-S(M))/(TF(M)+ACC)*dt/2-
     & TF(M)*dt
	 d(M)=-E1(M)-xx*(E1(M-1)-E1(M))+yy*FB
     & -k2(M)*(S(M)+dabs(S(M)))*dt/2-k2(M)*Gen(M)*dt !+eps_it1(i)*dt
	 do i=2,M-1
	  a(i)=dzeta(dble(i+0.5))*dhw/(2*(ddz(i)+0.5*ddz(i-1)                     VS,23.06.2007 
     &   +0.5*ddz(i+1))*h1)-k3_mid(i-1)*dt                                      VS,23.06.2007
     &   /(h1**2*ddz(i)*(ddz(i)+ddz(i-1)) )                                     VS,23.06.2007
	  b(i)=-dzeta(dble(i+0.5))*dhw/(2*(ddz(i)+0.5*ddz(i-1)                    VS,23.06.2007
     &   +0.5*ddz(i+1))*h1)-k3_mid(i)*dt                                        VS,23.06.2007
     &   /(h1**2*ddz(i)*(ddz(i+1)+ddz(i)) )                                     VS,23.06.2007 
	  c(i)=a(i)+b(i)-1-(dABS(S(i))-S(i))/(TF(i)+ACC)*dt/2-
     &  TF(i)*dt	  
     	  d(i)=-E1(i)-k2(i)*(S(i)+dabs(S(i)))*dt/2-k2(i)*Gen(i)*dt !+eps_it1(i)*dt
	  d(i)=d(i)-dt*(h1**2*ddz(i)*(ddz(i)+ddz(i+1)) )**(-1)                    VS,23.06.2007
     &   *(k3_mid(i)*(E1(i+1)-E1(i)))   
	  d(i)=d(i)+dt*(h1**2*ddz(i)*(ddz(i)+ddz(i-1)) )**(-1)                    VS,23.06.2007
     &   *(k3_mid(i-1)*(E1(i)-E1(i-1)))
	  d(i)=d(i)-dzeta(dble(i+0.5))*dhw*                                       VS,23.06.2007
     &   (2*(ddz(i)+0.5*ddz(i-1)+0.5*ddz(i+1))*h1)**(-1)                        VS,23.06.2007
     &   *(E1(i+1)-E1(i-1)) 
	 enddo
	 ind_bound=.false.
	 call ind_stab_fact_db (a,b,c,1,M,indstab,ind_bound)
	 if (indstab==.false.) then
	  do i=2,M-1
	   a(i)=-k3_mid(i-1)*dt/(h1**2*ddz(i)*(ddz(i)+ddz(i-1)) )
	   b(i)=-k3_mid(i)*dt/(h1**2*ddz(i)*(ddz(i+1)+ddz(i)) )
	  enddo
    	 endif 
       call progonka2_db (a,b,c,d,E_it2,1,M)
	 !E_it2(1)=E_it2(2)
	 E_it2(M+1)=0.

		      
	 call CUT_E_DB(E_it2,M,0.d0)

	 dE_it=maxval(dabs(E_it1-E_it2))

      !if (iterat) then
	! if (dE_it>1.E-10.and.iter_t<35) then !.or.deps_it>10.**(-13)
	  !E_it1=E_it1*al_it+E_it2*(1-al_it)
	!  E_it1=E_it1+alp(iter+1)*(E_it2-E_it1)
	! endif
      !endif

!C1 is given following Satyanarayana et al., 1999; Aupoix et al., 1989
       do i=1,M !+1
        Re(i)=ACC+(2*E_it1(i)/3.)**2/(eps_it1(i)*niu+ACC)
	  C1(i)=Co/(1+0.69*(2-Co)/dsqrt(Re(i)))
	  if (S(i)<0.) then
	   ceps3(i)=-0.4
        else
         ceps3(i)= 1.
	  endif
	 enddo

!EPS - EQUATION
20	 if (l1==0) then
	  FS=k5(1)*E_it1(1)**(3./2.)/kar*(roughness+0.2)**(-2) !roughness+ddz/2*h1
	 else
	  FS=k5(1)*E_it1(1)**(3./2.)/kar*(0.01+ddz(1)/2*h1)**(-2) !0.01 - roughness of ice !0.01+ddz/2*h1
       endif 
	 xx=0.5*(k5_mid(1)*dt/(h1**2*ddz(1)*0.5*(ddz(1)+ddz(2)) )                 VS,24.06.2007
     &  +0.5*ddz(1)*dhw/(h1*0.5*(ddz(1)+ddz(2)) ) )                             VS,24.06.2007
	 yy=dt/(h1*ddz(1))                                                        VS,24.06.2007
	 b(1)=-xx
	 c(1)=-(xx+1)-C1(1)*TF(1)*dt
     & +C1(1)*0.5*(-dABS(S(1))+S(1))/(TF(1)+ACC)*dt
	 d(1)=-eps1(1)-xx*(eps1(2)-eps1(1))-yy*FS-0.5*C1(1)*
     & (dABS(S(1))+S(1))*TF(1)*k2(1)*dt-
     & C1(1)*TF(1)*k2(1)*Gen(1)*dt !+eps_it1(i)*dt
       
	 FB=-k5(M)*E_it1(M)**(3./2.)/kar*(0.01+ddz(M)/2*h1)**(-2) !0.01+ddz/2*h1
	 xx=0.5*(k3_mid(M-1)*dt/(h1**2*ddz(M)*0.5*(ddz(M)+ddz(M-1)) )-            VS,24.06.2007
     &  dzeta(dble(M+0.5))*dhw/(h1*0.5*(ddz(M)+ddz(M-1)) ) )                    VS,24.06.2007 
	 yy=dt/(h1*ddz(M))                                                        VS,24.06.2007
	 a(M)=-xx
	 c(M)=-(xx+1)-C1(M)*TF(M)*dt
     & +C1(M)*0.5*(-dABS(S(M))+S(M))/(TF(M)+ACC)*dt
	 d(M)=-eps1(M)-xx*(eps1(M-1)-eps1(M))+yy*FB-0.5*C1(M)*
     & (dABS(S(M))+S(M))*TF(M)*k2(M)*dt-
     & C1(M)*TF(M)*k2(M)*Gen(M)*dt
	 do i=2,M-1
	  a(i)=dzeta(dble(i+0.5))*dhw/(2*(ddz(i)+0.5*ddz(i-1)                     VS,24.06.2007 
     &   +0.5*ddz(i+1))*h1)-k5_mid(i-1)*dt                                      VS,24.06.2007
     &   /(h1**2*ddz(i)*(ddz(i)+ddz(i-1)) )                                     VS,24.06.2007
	  b(i)=-dzeta(dble(i+0.5))*dhw/(2*(ddz(i)+0.5*ddz(i-1)                    VS,24.06.2007
     &   +0.5*ddz(i+1))*h1)-k5_mid(i)*dt                                        VS,24.06.2007
     &   /(h1**2*ddz(i)*(ddz(i+1)+ddz(i)) )                                     VS,24.06.2007 
        c(i)=a(i)+b(i)-1-C1(i)*TF(i)*dt
     &   +C1(i)*0.5*(-dABS(S(i))+S(i))/(TF(i)+ACC)*dt
	  d(i)=-eps1(i)-0.5*C1(i)*(dABS(S(i))+S(i))*TF(i)*k2(i)*dt-
     &   C1(i)*TF(i)*k2(i)*Gen(i)*dt
	  d(i)=d(i)-dt*(h1**2*ddz(i)*(ddz(i)+ddz(i+1)) )**(-1)*
     &   (k5_mid(i)*(eps1(i+1)-eps1(i)))
	  d(i)=d(i)+dt*(h1**2*ddz(i)*(ddz(i)+ddz(i-1)) )**(-1)*
     &   (k5_mid(i-1)*(eps1(i)-eps1(i-1)))
	  d(i)=d(i)-dzeta(dble(i+0.5))*dhw*                                       VS,24.06.2007
     &   (2*(ddz(i)+0.5*ddz(i-1)+0.5*ddz(i+1))*h1)**(-1)*                       VS,24.06.2007
     &   (eps1(i+1)-eps1(i-1))
       enddo
	 ind_bound=.false.
	 call ind_stab_fact_db (a,b,c,1,M,indstab,ind_bound)
	 if (indstab==.false.) then
	  do i=2,M-1
	   a(i)=-k5_mid(i-1)*dt/(h1**2*ddz(i)*(ddz(i)+ddz(i-1)) )                 VS,24.06.2007
	   b(i)=-k5_mid(i)*dt/(h1**2*ddz(i)*(ddz(i+1)+ddz(i)) )                   VS,24.06.2007
	  enddo
    	 endif  
	
	 call progonka2_db (a,b,c,d,eps_it2,1,M)
	 eps_it2(M+1)=0.
	 
       call CUT_EPS_DB(eps_it2,M,0.d0)
	
	 deps_it=maxval(dabs(eps_it1-eps_it2))
	
	 if (iterat) then
        if ((dE_it>1.E-10.or.deps_it>1.E-13).and.iter_t<35) then !.or.deps_it>10.**(-13)
	   E_it1=E_it1*al_it+E_it2*(1-al_it)
	   eps_it1=eps_it1*al_it+eps_it2*(1-al_it)
         !eps_it1=eps_it1+alp(iter+1)*(eps_it2-eps_it1)
	   !E_it1=E_it1+alp(iter+1)*(E_it2-E_it1)
	   !E_it1=E_it1+alp(iter+1)*(E_it2-E_it1)
	   iter=iter+1
	   iter_t=iter_t + 1
	   if (iter==8) iter=0
	   goto 10
	  else
	   if (dE_it<1.E-6.and.deps_it<1.E-7) then
	    E2=E_it2
	    eps2=eps_it2
	    iter=0
	    iter_t=0
	    nl=2
	   else
	    eps_it1=eps1
	    E_it1=E1
	    iterat=.false.
	    nl=1
	    goto 10
         endif
        endif
       else
        E2=E_it2
	  eps2=eps_it2
	  iter=0
	  iter_t=0
	 endif
	 if (dE_it<1.E-10.or.deps_it<1.E-13) nl=3
	 
!END OF ITERATIONAL PROCESS

      elseif (numEeps==2) then
	
	print*, 'This method is not operational at the moment'
	stop

!      include 'numEeps.inc'

      else

	 print*, 'No numerical method associated with numEeps=', numEeps
	 stop

      endif
	
!      call vsmoot(E2,M+1,1,M)
!	call vsmoot(eps2,M+1,1,M)

      E12=(E1+E2)/2
	eps12=(eps1+eps2)/2
	do i=1,M
	 !k2(i)=dmin1(Ck*E2(i)**2/(eps3(i)+ACC),0.01)
	 !E_mid=(E12(i)+E12(i+1))/2.
	 !E_mid=(E2(i)+E2(i+1))/2.
	 !Eps_mid=(Eps12(i)+Eps12(i+1))/2.
       !Eps_mid=(Eps2(i)+Eps2(i+1))/2.
	 !KT(i)=lam_T*CE*E2(i)**2/(eps2(i)+ACC)*row0*cw
	  KT(i)=lam_T*dmax1(CEt*E2(i)**2/(eps2(i)+ACC),1.d-6)*row0*cw
	 !KT(i)=lam_T*dmax1(CE*E_mid**2/(eps_mid+ACC),1.E-6)*row0*cw
	enddo
	
!OUTPUT
      if (turb_out==1) then
	 !if (time/hour+12.-3.*24.>0.and.time/hour+12.-3.*24.<12.) then
	 do i=1,M+1
	  write (2115,7) time/hour+12.-3*24, dzeta(dble(i+0.5))*h1, 
     &   E2(i)*1.E9,eps2(i)*1.E9, KT(i),tau2*1.E3,Tw1(i)
	 enddo
	 !endif
       write (2116,8) time/hour+12.-3*24, E2(1)*1.E9,
     &  eps2(1)*1.E9, KT(1)
	endif
7 	format (f10.4,f8.2,2f12.1,3f8.2)      
8     format (f10.4,2f12.1,f8.2)      
      			
	 
      E1=E2
	eps1=eps2
	
	endif ifkeps
	
1     u1=u2
	v1=v2
	
	select case (Turbpar)
	 case (2)
	 case (1)
	 do i=1,M+1
	  KT(i)=(k2(i)-niu)*cw*row0 !*0.01
       enddo
	 case default
	  KT=lamTM*k2*cw*row0 !*0.01
        if (turb_out==1) then
	   if (firstcall) write (2117,'(a)') tp_name
	   do i=1,M+1
          write (2117,'(F8.2, F5.2, E10.3, F8.1, F5.2, F6.2)')
     &    time/hour+12.-3*24, dzeta(dble(i+0.5))*h1, Ri(i),
     &    KT(i),dzeta(dble(i))*h1,uv(i)
	   enddo
	  endif
      endselect 
     
      if (firstcall) firstcall=.false.

	END SUBROUTINE KT_eq


      SUBROUTINE soilforlake(dt,h1,l1,ls1,hs1,a,b,c,d,Temp)
     
!     SOILFORLAKE calculates profiles of temperature, liquid and solid water 
!     content in the soil under a lake     
     
	use phys_constants2 
	use driving_params
	use arrays
      implicit none

		
	real(8), dimension(1:350):: a,b,c,d,Temp  
	real(8) Hflow,dt,dzmean,ma,mi,pi,
     & wlmax_i,bb1,arg,psi,pf,cr,wflow,wfhigh,lammoist1,lammoist2,h1,
     & l1,surplus,Potphenergy,watavailtofreeze,unfrwat,
     & filtr_low,ls1,hs1,month,Tf_old,
     & dhwfsoil1,dhwfsoil2,dhwfsoil3,dhwfsoil4,wl_max,max1,
     & nyear,nmonth,nday,nhour    
	real(8) alp(10),psiit(10)
	real(8), external:: Meltpnt
	integer(4) i, j, k, iter,iter3
	logical soil_ts_ext_iter,firstcall

	common/ts_ext_iter/ Tf_old,soil_ts_ext_iter
	common /ymdh/ nyear,nmonth,nday,nhour
	data psiit/6,3,7,2,5,4,8,1,0,0/ !/3,2,4,1,0,0,0,0,0,0/  
	data firstcall /.true./
      SAVE

!	open (133, file=dir_out//'\dhwfsoil.dat', status = 'unknown')	
     
	if (firstcall) then
	endif

!	lam = 0.00041*418
	pi = 3.141592654
	cr =  0.2*4180
	rosdry = 1200.
	month = 24*30*60*60
!PARAMETERS OF ITERATIONAL PROCESS
	ma=10.5 !5.5
	mi=4.5
	do i=1,8 
	 alp(i)=2*(mi+ma-(ma-mi)*dcos((2*psiit(i)-1)/16*pi))**(-1)
      enddo
			       
     	do i=1, ns
       wlmax_i = POR(i)*ROW0/(rosdry(i)*(1-por(i))) 
	 BB1 = BH(i) + 2.
	 csoil(i) = cr+WL1(i)*CW+WI1(i)*CI
	 rosoil(i) = rosdry(i)*(1-por(i))*(1+wi1(i)+wl1(i)) 
	 ARG = (WL1(i)+WI1(i))/wlmax_i
	 ARG = dmax1(ARG,1.d-2)
	 PSI = PSIMAX(i)*(ARG)**(-BH(i))
	 PF = dlog(-PSI*100.)/dLOG(10.d0)
	 IF(PF>5.1) THEN
	  lamsoil(i) = 4.1E-4*418.
	 ELSE
	  lamsoil(i) = dexp(-PF-2.7)*418.
	 END IF
	 wlmax_i = POR(i)*ROW0/(rosdry(i)*(1-por(i))) - wi1(i)*row0/roi
	 ARG = (WL1(i))/wlmax_i
	 ARG = dmax1(ARG,1.d-2)
	 lammoist(i) = dlmax(i)*ARG**BB1
	end do
      
      Hflow = 0
      
      do i=1, ns-1
	 wlmax_i=WL_MAX(por(i+1),rosdry(i+1),wi1(i+1))
	 if (wl1(i+1)/wlmax_i>0.98.or.wlmax_i<0.01) then
	  filtr(i) = 0
       else
	  wlmax_i = (POR(i)+POR(i+1))*ROW0/((rosdry(i)+rosdry(i+1))*
     &  (1-(por(i)+por(i+1))/2)) - (wi1(i)+wi1(i+1))/2*row0/roi      
	  filtr(i) = (flwmax(i+1)+flwmax(i))/2*((wl1(i+1)+
     &  wl1(i))/2/wlmax_i)**(bh(i+1)+bh(i)+3)
	 endif
      enddo
	
!SPLITTING-UP METHOD FOR TEMPERATURE EQUATION:
!STEP 1: HEAT DIFFUSION
      
      do i = 1, ns-1                                                            VS,02.10.07
       lammoist1 = 0.5*(lammoist(i)+lammoist(i+1))                              VS,03.10.07
       wsoil(i) = ( - lammoist1*(wl1(i+1)-wl1(i))/dzs(i) +                      VS,03.10.07
     &  filtr(i) ) *0.5*(rosdry(i)+rosdry(i+1))*                                VS,03.10.07
     &  (1-0.5*(por(i)+por(i+1)) )/row0                                         VS,03.10.07
      enddo                                                                     VS,02.10.07
      
      CALL S_diff(dt)                                                           VS,02.10.07
      CALL T_SOLVER(dt)
      
!SPLITTING-UP METHOD FOR MOISTURE EQUATION:
!STEP 1: MOISTURE DIFFUSION
	
	Wflow = 0.
	dhwfsoil=0.
	dhwfsoil1=0.
	dhwfsoil2=0.
	dhwfsoil3=0.
	dhwfsoil4=0.
	
      if ((l1 /= 0.and. h1 == 0.).or.ls1/=0.or.
     & (hs1/=0.and.l1==0.and.h1==0)) then
	 Wfhigh = 0
!	 c(1)=1
!	 b(1)=1
!	 d(1)=0	
       c(1) = -1-dt*(lammoist(1)+lammoist(2))/(dzs(1)**2)                       VS,23.09.07
       b(1) =   -dt*(lammoist(1)+lammoist(2))/(dzs(1)**2)                       VS,23.09.07
       d(1) =   -wl1(1)                                                         VS,23.09.07
	endif
	   
	if (h1 /= 0.and.ls1 == 0) then
	 c(1)=1
	 b(1)=0
	 d(1)=WL_MAX(por(1),rosdry(1),wi1(1))
!	 dhwfsoil1 = dt*(wl1(2)-wl1(1))*rosdry(1)*(1-por(1))/row0
!     & /dzs(1)*(lammoist(1)+lammoist(2))/2 !-
!     &  (WL_MAX(por(1),rosdry(1),wi1(1))-wl1(1))                               VS,23.09.07
!     & *rosdry(1)*(1-por(1))*dzss(1)/row0                                      VS,23.09.07
      endif	 
	  
	lammoist1 = (lammoist(ns-1)+lammoist(ns))/2
!      c(ns)=-lammoist1/dzs(ns-1)
!	 a(ns)=-lammoist1/dzs(ns-1)
!	 d(ns)=Wflow
       c(ns) = -1-2*dt*lammoist1/dzs(ns-1)**2 
       a(ns) = -2*dt*lammoist1/dzs(ns-1)**2 
       d(ns) = -wl1(ns)
           
      do i=2,ns-1
       lammoist2 = (lammoist(i)+lammoist(i+1))/2
       lammoist1 = (lammoist(i-1)+lammoist(i))/2
       dzmean = (dzs(i-1)+dzs(i))/2
       a(i)=-dt/dzmean*lammoist1/dzs(i-1)
	 b(i)=-dt/dzmean*lammoist2/dzs(i)
	 c(i)=-dt/dzmean*(lammoist1/dzs(i-1)+
     & lammoist2/dzs(i))-1
	 d(i)=-wl1(i)
       !wl2(i) = wl1(i)+dt*(lammoist2/dzs(i)*(wl1(i+1)-wl1(i))-
      !& lammoist1/dzs(i-1)*(wl1(i)-wl1(i-1)))/dzmean
	enddo   
		
	call progonka (a,b,c,d,wl2,ns) 	
		
      if (h1 /= 0.and.ls1 == 0) then                                            VS,23.09.07
       lammoist1 = (lammoist(1)+lammoist(2))/2.	                                VS,23.09.07
       dhwfsoil1 = -(wl2(1)*(dzs(1)/(2*dt)+lammoist1/dzs(1))-                   VS,23.09.07
     &  wl2(2)*lammoist1/dzs(1)-wl1(1)*dzs(1)/(2*dt))*                          VS,23.09.07
     &  rosdry(1)*(1-por(1))/row0*dt                                            VS,03.11.07
      endif                                                                     VS,23.09.07
!STEP 2 OF SPLITTING-UP METHOD FOR MOISTURE EQUATION: 
!EVOLUTION OF MOISTURE DUE TO GRAVITATIONAL INFILTRATION 	     	    
	 
      do i=2,ns-1
       wl3(i) = wl2(i) + dt*(filtr(i-1)-filtr(i))/dzss(i)
      enddo
     	 
      if ((h1==0.and.l1/=0).or.ls1 /= 0) then
       wl3(1) = wl2(1) + dt*(-filtr(1)+0)/dzss(1)
      endif

	if (h1/=0.and.ls1 == 0) then
	 wl3(1) = WL_MAX(por(1),rosdry(1),wi1(1))
	 dhwfsoil2 = - filtr(1)*rosdry(1)*(1-por(1))/row0*dt 
      endif
	
	filtr_low = 0
	wl3(ns) = wl2(ns) + dt*(-filtr_low+filtr(ns-1))/dzss(ns)
	
     	do i=1,ns
	 if (wl3(i)>WL_MAX(por(i),rosdry(i),wi1(i))) then
	  surplus=wl3(i)-WL_MAX(por(i),rosdry(i),wi1(i))
	cy1:do j=1,ns
	    if (WL_MAX(por(j),rosdry(j),wi1(j))-wl3(j)>0) then
	     if (WL_MAX(por(j),rosdry(j),wi1(j))
     &	  -wl3(j)>surplus*rosdry(i)*(1-por(i))*dzss(i)/
     &      (rosdry(j)*(1-por(j))*dzss(j))) then
       	  wl3(j)=wl3(j)+surplus*rosdry(i)*(1-por(i))*dzss(i)/
     &	  (rosdry(j)*(1-por(j))*dzss(j))
		  surplus=0
            exit cy1
	     else
	      surplus=surplus-(WL_MAX(por(j),rosdry(j),wi1(j))-wl3(j))*
     &	  (rosdry(j)*(1-por(j))*dzss(j))/
     &      (rosdry(i)*(1-por(i))*dzss(i))
	   	  wl3(j)=WL_MAX(por(j),rosdry(j),wi1(j)) 
           endif
          endif
	   enddo cy1
        dhwfsoil3 = dhwfsoil3 + surplus*rosdry(i)*(1-por(i))*dzss(i)/
     &   row0
	  wl3(i) = wl_max(por(i),rosdry(i),wi1(i))
	 endif
	 if(wl3(i)<0) then
	  if (i/=1) then 
	   do j=i-1,1,-1
	    if (wl3(j)>-wl3(i)*rosdry(i)*(1-por(i))*dzss(i)/
     &  	(rosdry(j)*(1-por(j))*dzss(j))) then
	     wl3(j)=wl3(j)+wl3(i)*rosdry(i)*(1-por(i))*dzss(i)/
     &     (rosdry(j)*(1-por(j))*dzss(j))
	    goto 1
	    endif
         enddo
        endif
	  if (i/=ns) then 
	   do j=i+1,ns
	    if (wl3(j)>-wl3(i)*rosdry(i)*(1-por(i))*dzss(i)/
     &    (rosdry(j)*(1-por(j))*dzss(j))) then
	     wl3(j)=wl3(j)+wl3(i)*rosdry(i)*(1-por(i))*dzss(i)/
     &     (rosdry(j)*(1-por(j))*dzss(j))
	    goto 1
	    endif
         enddo
        endif
       !dhwfsoil = dhwfsoil - abs(wl4(i))*rosdry(i)*(1-por(i))*dzss(i)
     & ! /row
1	  wl3(i)=0.
	 endif
      enddo
      
! STEP 3 OF SPLITTING-UP METHOD : PHASE PROCESSES

20    c2:do i=1, ns
       if (wi1(i)>0.and.Tsoil2(i)>Meltpnt(Sals2(i))+0.01) then
        Potphenergy = (Tsoil2(i)-(Meltpnt(Sals2(i))+0.01))*
     &   csoil(i)*rosoil(i)*dzss(i)
        if (Potphenergy>=wi1(i)*rosdry(i)*dzss(i)*(1-por(i))*Lwi) then
         wl4(i) = wl3(i) + wi1(i)
	   wi2(i) = 0
	   Tsoil3(i) = Meltpnt(Sals2(i))+0.01+
     &	  (Potphenergy-(wi1(i)*rosdry(i)*dzss(i)*(
     &    1-por(i))*Lwi))/(csoil(i)*rosoil(i)*dzss(i))
        else
	   wl4(i) = wl3(i) + Potphenergy/(rosdry(i)*dzss(i)*(1-por(i))*Lwi)
	   wi2(i) = wi1(i) - Potphenergy/(rosdry(i)*dzss(i)*(1-por(i))*Lwi)
	   Tsoil3(i) = Meltpnt(Sals2(i)) + 0.01
	  endif
       else
        if (wl3(i)>=0.and.Tsoil2(i)<Meltpnt(Sals2(i))-0.01) then
         Potphenergy = -(Tsoil2(i)-Meltpnt(Sals2(i))+0.01)
     &    *csoil(i)*rosoil(i)*dzss(i)
         watavailtofreeze = (wl3(i)-unfrwat(Tsoil2(i),i))*rosdry(i)*
     &   dzss(i)*(1-por(i))*Lwi
	   if (watavailtofreeze<0) then 
	    !cc = csoil(i)*rosoil(i)*dzss(i)
	    !cc1 = rosdry(i)*dzss(i)*(1-por(i))*Lwi
	    !bb = (-Potphenergy-0.1*cc-wl3(i)*cc1)/cc
	    !aa = cc1/cc
	    !call phase_iter(aa,bb,i,Tsoil3(i))
	    Tsoil3(i) = Tsoil2(i) +  
     &   (wl3(i)-unfrwat(Tsoil2(i),i))*rosdry(i)*dzss(i)*(1-por(i))*Lwi/
     &    (csoil(i)*rosoil(i)*dzss(i))
	    wi2(i) = wi1(i) + (wl3(i)-unfrwat(Tsoil2(i),i))
          wl4(i) = unfrwat(Tsoil2(i),i)
	    cycle c2  
         endif
	   if (Potphenergy>=watavailtofreeze) then
	    Tsoil3(i) = Meltpnt(Sals2(i)) -0.01 - 
     &	   (Potphenergy-watavailtofreeze)/
     &     (csoil(i)*rosoil(i)*dzss(i))
	    wi2(i) = wi1(i) + (wl3(i)-unfrwat(Tsoil2(i),i))
	    wl4(i) = unfrwat(Tsoil2(i),i)
	   else
	    wl4(i) = wl3(i)-Potphenergy/(rosdry(i)*dzss(i)*(1-por(i))*Lwi)
	    wi2(i) = wi1(i)+Potphenergy/(rosdry(i)*dzss(i)*(1-por(i))*Lwi)
	    Tsoil3(i) = Meltpnt(Sals2(i)) - 0.01
	   endif
	  else
	   wl4(i) = wl3(i)
	   wi2(i) = wi1(i)
	   Tsoil3(i) = Tsoil2(i)
	  endif
       endif
      enddo c2
      
      max1=0.
      do i=1,ns
       if (Tsoil3(i)<Meltpnt(Sals2(i))-0.01.and.
     &  dabs(wl4(i)-unfrwat(Tsoil3(i),i))>max1)
     & then
	   max1=dabs(wl4(i)-unfrwat(Tsoil3(i),i))
	 endif
	enddo    
	  
      if (max1>0.01.and.iter3<20) then
	 !wl3=(wl4+wl3)/2.
	 !wi1=(wi2+wi1)/2.
	 !Tsoil2=(Tsoil3+Tsoil2)/2.
	 wl3=wl3+alp(iter+1)*(wl4-wl3)
	 wi1=wi1+alp(iter+1)*(wi2-wi1)
	 Tsoil2=Tsoil2+alp(iter+1)*(Tsoil3-Tsoil2)
	 iter = iter + 1
	 iter3 = iter3 + 1
	 if(iter==8) iter = 0
	 goto 20
      else
	 iter = 0
	 iter3 = 0 
	endif   
	
	!evol=0

      do i=1,ns
       if (wi2(i)<0) then
	  wl4(i)=wl4(i)+wi2(i)
	  wi2(i)=0
       endif
	enddo

	do i=1,ns
	 if (wl4(i)>POR(i)*ROW0/(rosdry(i)*(1-por(i)))-wi2(i)*row0/roi) then
	  surplus = wl4(i)-(POR(i)*ROW0/(rosdry(i)*(1-por(i))) - wi2(i)*
     &  row0/roi)
	c3:do j=1,ns
	    if (POR(j)*ROW0/(rosdry(j)*(1-por(j))) - wi2(j)*row0/roi
     &	-wl4(j)>0) then
	     if (POR(j)*ROW0/(rosdry(j)*(1-por(j))) - wi2(j)*row0/roi
     &	  -wl4(j)>surplus*rosdry(i)*(1-por(i))*dzss(i)/
     &      (rosdry(j)*(1-por(j))*dzss(j))) then
       	  wl4(j)=wl4(j)+surplus*rosdry(i)*(1-por(i))*dzss(i)/
     &	  (rosdry(j)*(1-por(j))*dzss(j))
		  surplus=0
            exit c3
	     else
	      surplus=surplus-(POR(j)*ROW0/(rosdry(j)*(1-por(j))) -
     & 	  wi2(j)*row0/roi-wl4(j))*(rosdry(j)*(1-por(j))*dzss(j))/
     &      (rosdry(i)*(1-por(i))*dzss(i))
		  wl4(j)=POR(j)*ROW0/(rosdry(j)*(1-por(j))) - wi2(j)*row0/roi 
           endif
          endif
	   enddo c3
	  dhwfsoil4 = dhwfsoil4 + surplus*rosdry(i)*(1-por(i))*dzss(i)/
     &	 row0
	  wl4(i) = POR(i)*ROW0/(rosdry(i)*(1-por(i))) - wi2(i)*row0/roi 
	 endif
	 if(wl4(i)<0) then
	  if (i/=1) then 
	   do j=i-1,1,-1
	    if (wl4(j)>-wl4(i)*rosdry(i)*(1-por(i))*dzss(i)/
     &	   (rosdry(j)*(1-por(j))*dzss(j))) then
	     wl4(j)=wl4(j)+wl4(i)*rosdry(i)*(1-por(i))*dzss(i)/
     &     (rosdry(j)*(1-por(j))*dzss(j))
	    goto 2
	    endif
         enddo
        endif
	  if (i/=ns) then 
	   do j=i+1,ns
	    if (wl4(j)>-wl4(i)*rosdry(i)*(1-por(i))*dzss(i)/
     &	   (rosdry(j)*(1-por(j))*dzss(j))) then
	     wl4(j)=wl4(j)+wl4(i)*rosdry(i)*(1-por(i))*dzss(i)/
     &     (rosdry(j)*(1-por(j))*dzss(j))
	    goto 2
	    endif
         enddo
        endif
       !dhwfsoil = dhwfsoil - abs(wl4(i))*rosdry(i)*(1-por(i))*dzss(i)
     & ! /row
2	  wl4(i)=0
	 endif
      enddo
            
  !!!!! II. FREEZING AND MELTING IN CASE OF TSOIL<MELTING POINT (O C)!!!!!

		!do i=1, ns
		! if (Tsoil3(i)<-1.) then
		!   wi3(i) = wi2(i) + (wl4(i)-unfrwat(Tsoil3(i),i))
        !   wl5(i) = unfrwat(Tsoil3(i),i)
		!   Tsoil4(i) = Tsoil3(i) + & 
		!   (wl4(i)-unfrwat(Tsoil3(i),i))*rosdry(i)*dzss(i)*(1-por(i))*Lwi/&
		!   (csoil(i)*rosoil(i)*dzss(i))
		! else
		!   wi3(i) = wi2(i)
		!   wl5(i) = wl4(i)
		!   Tsoil4(i) = Tsoil3(i)
        ! end if
		! if (wl5(i)>POR(i)*ROW/(rosdry(i)*(1-por(i))) - wi3(i)*row/roi) then
	    !  surplus = wl5(i)-(POR(i)*ROW/(rosdry(i)*(1-por(i))) - wi3(i)*row/roi)
	    !  dhwfsoil = dhwfsoil + surplus*rosdry(i)*(1-por(i))*dzss(i)/row
		!  wl5(i) = POR(i)*ROW/(rosdry(i)*(1-por(i))) - wi3(i)*row/roi 
        ! endif
	    ! if(wl5(i)<0) then 
		!  dhwfsoil = dhwfsoil - abs(wl5(i))*rosdry(i)*(1-por(i))*dzss(i)/row
		!  wl5(i)=0
	    ! endif
        !enddo
   
	   ! evol=0
	!	evol1=0

	!	do i=1,ns
	!	 evol=evol+(Tsoil3(i)-Tsoil1(i))*rosoil(i)*csoil(i)*dzss(i)
	!	 evol1=evol1+(wi2(i)-wi1(i))*rosdry(i)*(1-por(i))*dzss(i)*Lwi
      !   enddo

      dhwfsoil=dhwfsoil1+dhwfsoil2+dhwfsoil3+dhwfsoil4
     	  
  	Tsoil1 = Tsoil3
	wl1 = wl4
	wi1 = wi2

      if (firstcall) firstcall=.false.
      RETURN
	END SUBROUTINE soilforlake 


      SUBROUTINE snowtemp(snowmass,snowmass_init,a,b,c,d,Temp,dt)
      
!     SNOWTEMP calculates temperature profile in the snow cover      
      
      use numeric_params    
	use atmos
	use phys_constants2
	use driving_params
	use arrays
	implicit none


!-----------------------------MAIN VARIABLES-------------------------------
!         arrays: Tsn - snow temperature, C
!                 lams - thermal conductivity of snow
!                
!            rofresh - density of fresh snow, kg/m**3           
!            snowmass - snow mass, kg/m**2

!            Boundary conditions:
!            t2 - air temperature (2 m), C        
       
      real(8)   xx, dzsn(ms)
	real(8)   rofresh,t2,snowmass,pheat,pmelt,snowmass_init,
     & Tf_old1,month,pi,CCT(ML)
      real(8) q(110)
	real(8), dimension(1:ms)::lams,cs,Tsn
      real(8) dt, h1, l1, hs1, ls1
      real(8) dz
      real(8) ST,PGR,TGROLD,QGROLD,RADIAT,PRECIP,WSOLD,SNOLD,
     & ZS,thsoil,whsoil,BOLD,RF1,RF2,SNMELT,HSNOW,
     & HS,ES,TGRNEW,QGRNEW,WSNEW,SNNEW,RUNOFF,
     & ElatOld,HFold,PRSold,shortwave,extinct
      real(8) AL,DLT,DVT,ALLL,DL,ALV,DV,Z,T,WL,WV,WI,dens
	real(8), dimension(1:350)::a,b,c,d,Temp

	integer(4) itop,i,
     & iyear,imonth,iday  

	logical soil_ts_ext_iter

      common /watericesnowarr/ lams, q
	common /layers/ h1,l1,hs1,ls1
	common /BL/ ST,PGR,TGROLD,QGROLD,RADIAT,WSOLD,SNOLD,
     & ZS,thsoil,whsoil,BOLD,RF1,RF2,SNMELT,HSNOW,
     & HS,ES,TGRNEW,QGRNEW,WSNEW,SNNEW,RUNOFF,
     & ElatOld,HFold,PRSold,extinct(ms)
	common /atmos2/ shortwave, precip
	common /SOILSOL/ AL(ML),DLT(ML),DVT(ML),ALLL(ML),DL(ML),
     & ALV(ML),DV(ML),Z(ML),T(ML),WL(ML),WV(ML),WI(ML),dens(ms)
	common /SOILDAT/ dz(ms),itop
	common/ts_ext_iter/ Tf_old1,soil_ts_ext_iter 
	common /snow_char/ Tsn,cs
	
	SAVE
	 
!     BODY	

      pi=3.141592654
	month=30*24*60*60
!	ddz = 1./float(M)
    
	if (flag_snow_init == 1) then
	 itop = ms-2 
	 hs1=0.02
	 klmn=0	
       rofresh = 67.9 + 51.3*dexp(tempair/2.6)
	! snowmass = rofresh*0.012
	 do i = itop, ms-1
	  WL(i) = 0.0
	  dens(i) = rofresh 
	  dz(i) = 0.01
	 end do
	 wl(ms)=0.
	 dens(ms)=rofresh
	 T(ms) = Ti1(1)
	 do i = itop, ms-1
	  T(i) = tempair
       end do
	 do i=1,itop-1
	  T(i)=0
	  wl(i)=0
	  dens(i)=0
	  dz(i)=0
       enddo
       !T(ms)=0
	 !wl(ms)=0
	 !dens(ms)=0
	 !dz(ms)=0
	 totalmelts=0.
	 totalprecips=0.
	 totalevaps=0.
      end if

	
	!ma=8.5
	!mi=1.5
	!data psi/6,3,7,2,5,4,8,1,0,0/   !/3,2,4,1,0,0,0,0,0,0/
	!do i=1,8 
	! alp(i)=2*(mi+ma-(ma-mi)*cos((2*psi(i)-1)/8*pi))**(-1)
      !enddo
	
       do i = itop, ms-2
        dzsn(i) = (dz(i)+dz(i+1))/2
       enddo
       do i = itop, ms-1
        !densav = (dens(i) + dens(i+1))/2
        !lams(i) =  0.419*(6.*(densav/row)**4+1.9*(densav/row)+0.05)
	  lams(i) = 0.419*(6.*(dens(i)/row0)**4+1.9*(dens(i)/row0)+0.05)
        xx = wl(i)*row0/(dens(i)*dz(i)) 
	  cs(i) = ci*(1-xx)+cw*xx
	 end do

      call soilforlake(dt, h1, l1, ls1, hs1, 
     & a,b,c,d,Temp)
	do i=itop,ms
	 T(i)=Tsn(i)	
      enddo
	t2=tempair
	Erad=shortwave*(1-albedoofsnow)
	Elatent=xlew
	call addPrecipToSnow(t2, T(itop), snowmass, iyear, imonth, iday, 
     & CCT, pmelt, pheat,dt)
	call snow_calc(t2,Erad,T(itop),snowmass,iyear, 
     &imonth, iday, CCt, pmelt, pheat,dt)
	
	if (flag_snow_init == 1) then
	! snowmass_init = snowmass  - (precip*dt*row-snmelt*dt*row-
      !& Elatent/Liv*dt)
       flag_snow_init = 0
	endif 
      
	totalevaps = totalevaps+Elatent/(row0*Lwv)*dt
	totalprecips = totalprecips+precip*dt
	totalmelts = totalmelts + snmelt*dt
	      
      hs1=hsnow
	if (hs1==0) hs1=0.00001
            
      END SUBROUTINE snowtemp

	SUBROUTINE snow_calc(t2,Erad_sc,ts,snow_sc,iyear,imonth,
     & iday,CCT,pmelt,pheat,dt)
      use phys_constants
      use phys_parameters
	use numeric_params
	use atmos
	use driving_params
	use arrays
	implicit none

CCCCCCCCCCCCCCC input parameters:
CC      e, cal/(cm2*s), --- surface heat balance: e = Radiation-SIGMA*(TS**4)-Latent-Sensible-InSoil
CC      Elatent, cal/(cm2*s), --- latent heat flux.
CC      ts, deg C, --- surface temperature.
CC      precip, cm --- precipitation during dt.
CC
CCCCCCCCC  some other parameters: CCCCCCCCCCC
CC      densny, g/cm3       --- density of fresh snow, depends on temperature at 2m
CC      dznorm, cm          --- initial thickness of snow layers
CC      dzmin, cm           --- minimal thickness of snow layers
CC      dt, sec             --- timestep
CC      ms                  --- max. number of layers in snowpack
CC      cwat, cal/(gr K)    --- the water specific heat content
CC      csnow, cal/(gr K)   --- the snow specific heat content
CC      rhow, gr/cm3        --- the liquid water density
CC      PLI,  cal/g         --- latent heat for freezing/melting
CC      PL,  cal/g          --- latent heat for evapor./condens.
CC      whc                 --- hydraulic constant
CC      hcond, cm/sec       --- hydraulic conductivity
CC
CCCCCCCCCCC  values to be calculate after every timestep:
CC      wl(i), cm       --- moisture content in snow layer No.i
CC      t(i), deg.C     --- mean temperature of snow layer No.i
CC      dz(i), cm       --- hickness of snow layer No.i
CC      dens(i), g/cm3  --- mean density of snow layer No.i
CC      hsnow, cm       --- snow depth
CC      swe, cm         --- water equivalent snow depth
CC      snmelt, cm/sec  --- mean snow melting speed during dt
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
	
	INTEGER(4) j,i,k,itop,iyear,imonth,iday
	 
	integer(4) nmonth,nyear,nday,nhour
	
	common /time/ nyear,nmonth,nday,nhour
	common /layers/ h1,l1,hs1,ls1
      COMMON /SOILDAT/ dz(ms),itop
	COMMON /BL/       ST,PGR,TGROLD,QGROLD,RADIAT,WSOLD,SNOLD,
     &                  ZS,thsoil,whsoil,BOLD,RF1,RF2,SNMELT,HSNOW,
     &                  HS,ES,TGRNEW,QGRNEW,WSNEW,SNNEW,RUNOFF,
     &                  ElatOld,HFold,PRSold,extinct(ms)
	common /atmos2/ shortwave, precip
	COMMON /SOILSOL/ AL(ML),DLT(ML),DVT(ML),ALLL(ML),DL(ML),
     &                  ALV(ML),DV(ML),Z(ML),T(ML),WL(ML),WV(ML),WI(ML),
     &                  dens(ms)
	COMMON /spin/ Cond,Phase

      common /watericesnowarr/ lams, q
      real(8) dt
	real(8) dz,q(110)
	real(8) lams(ms)
	real(8) ST,PGR,TGROLD,QGROLD,RADIAT,PRECIP,WSOLD,SNOLD, 
     &ZS,thsoil,whsoil,BOLD,RF1,RF2,SNMELT,HSNOW,HS,ES,TGRNEW,QGRNEW,
     &WSNEW,SNNEW,RUNOFF,ElatOld,HFold,PRSold
	real(8) AL,DLT,DVT,ALLL,DL,ALV,DV,Z,T,WL,WV,WI,dens
	real(8) extinct,erad_sc,ts,snow_sc,
     &evap,erain,countrad,qbase,dmass,densev,counter,qin,rf,
     &smelti,ein,dzi,ziz,porosityOfSnow,w0,fukt,dztop,smelt,eout,p1,q0,
     &dzmasi,swe,densny,shortwave,erad1,eflux1,Phase,Ei,sigma,h1,l1,hs1,
     &ls1,Cond
	real(8) EInRad,T2,rho_dry,CCT(ML),a2,eta,P,dens_old,ti(ml)
      real(8) pheat,pmelt

	SAVE
       
	precip = precip*dt
	!hsnow = 0.
	!do i = itop, ms-1
      ! hsnow = hsnow + dz(i)
	!end do
	!eFlux = eFlux - Erad
	eflux1 = eflux !/2 
	erad1 =  erad_sc !/2     
	!eflux1 = eflux1*(1.-exp(-0.02*snow))  ! not -0.02*snow*10.                  
	do i=itop,ms-1
	!	wl(i) = wl(i)*dz(i)*dens(i)/rhow    
		ti(i) = t(i)
	end do
	
	snmelt = 0.
	swe = 0.0
	Phase = 0.
c	if(t2 .le. -10) then
c		densny = 0.05
c	elseif(t2 .le. -2) then
c		densny = 5./800.*(t2+18.)
c		densny = 0.1
c	else
c		densny = (t2+4.)/20.
c	end if
	densny = (67.9 + 51.3 * dexp(t2/2.6))
	if(ts .gt. 0.00001) then
		Elatent = 0
	elseif(snow_sc .lt. 0.000001 .and. Elatent .gt. 0.0) then    
		Elatent = 0
	end if

	dz(ms-1) = dz(ms-1) + dz(ms)
	evap=Elatent/(PLI+PL) 
	!totalevap=totalevap+evap/rhow*dt                              
	hsnow=0

	if(itop.eq.ms) then

	else
CC   		--- first top snow layer ---
		i=itop
		dztop=dz(i)
		Erain = pheat                           
		if(extinct(itop).eq.0.0) then
			Eflux1 = Eflux1 + Erad1 + Erain                            
		else	
			if(itop.eq.ms-1) then
				Eflux1 = Eflux1 + Erain + Erad1            
			else
			   Eflux1 = Eflux1 + Erain + Erad1*(1 - extinct(i)**dz(i))      
			end if
			countrad = Erad1*(1 - extinct(i)**dz(i))
		end if               
CC rain 	heat from rain is added as pheat
CC rain 	the rain water itself is assumed to freeze giving heat to top layer
CC rain 	then this heat will melt snow
		if(Eflux1.gt.0.0)then
			if(itop.eq.ms-1) then
        		    if(Ts.ge.-0.0001)then
				  if(Eflux1/PLI*dt .le. DENS(i)*DZ(i)-WL(i)*rhow) then
						smelt = Eflux1/PLI/rhow                          
						Eout=0.
					else
						smelt = (DZ(i)*DENS(i)/rhow - WL(i))/dt
						Eout = Eflux1 - smelt*rhow*PLI    
					end if
				else
					smelt = 0.0
					Eout = 0.
				end if
			else
				smelt = 0.0
				Eout = Eflux1
			end if
		else
			smelt = 0.
			Eout = 0.
		end if

	
Cc
Cc	PERCOLATION
Cc
		smelt = smelt + wl(i)/dt
		wl(i) = 0.
		qbase = smelt

CC
CC	TOP  LAYER THICKNESS AND DENSITY
CC
		dmass=dens(i)*dztop-(evap+smelt*rhow)*dt     
		if(evap.gt.0.0)then
			densev=dens(i)
		else
			densev=rhoi
		endif
		dztop=dztop-(smelt*rhow/dens(i)+evap/densev)*dt   
	  if(dztop.eq.0.) dztop = 0.000001	
		dens(i)= dmax1(dmass/dztop,0.d0)
		if(dens(i).gt.rhoi) then
	      	dztop = dztop*dens(i)/rhoi
	         	dens(i)=rhoi
			dmass=dens(i)*dztop
		endif
		dz(i)=dztop
		if(itop .eq.ms-1) snmelt=qbase
	end if 


	counter = dz(itop)
	if(itop.lt.ms-1) then
CC
CC INTERNAL LAYERS
CC
		do i=itop+1,ms-1
			qin = qbase                                     
			rf = 0.0                                        
     			smelti = 0.0                                    
			Ein=Eout                                        
			T(i) = T(i)*dz(i)*dens(i)/(qbase*rhow*dt+dz(i)*dens(i))  
			dzi=dz(i)                                       
			wl(i) = wl(i) + qbase*dt
			if(wl(i).lt.0.) wl(i) = 0.
			w0=wl(i)
			if(extinct(i).eq.0.0) then
				Ei = Ein                
			else
				if(i.eq.ms-1) then
					Ei=Ein + (Erad1 - countrad)
				else
					Ei=Ein + Erad1*(extinct(i)**(cou
     &					nter)-extinct(i)**(counter+dz(i)) )
					EInRad = Erad1*(extinct(i)**(counter) -
     &					extinct(i)**(counter+dz(i)) )
				end if                            
				countrad = countrad + Erad1*(extinct(i)**(cou
     &              nter)-extinct(i)**(counter+dz(i)) )
				counter = counter+dz(i)
			end if

			dz(i)=dz(i)+qbase*dt  
			ziz=dzi/dz(i)
			dens(i)=ziz*dens(i)+(1.0-ziz)*rhow

            
			if(Ei.gt.0.0) then
				if(T(i).ge.0.0001) then  ! 0.0001
					Ei = Ei + csnow*dens(i)*T(i)*dz(i)/dt
					T(i) = 0.0
					if(Ei/PLI*dt .le. (DZ(i)*DENS(i)-WL(i)*rhow)) then
						smelti = Ei/PLI/rhow 
						Eout = 0.
					else
						smelti = (DZ(i)*DENS(i)/rhow - WL(i))/dt
						Eout = Ei - smelti*rhow*PLI 
						dzi = 0.
					end if
					wl(i) = wl(i) + smelti*dt

				else
Cc		T<0
					T(i) = T(i) + Ei*dt/(csnow*dens(i)*dz(i))
					if(T(i).lt.-0.0001) then   !0.0001
					if(wl(i) .gt. -T(i)*csnow*dz(i)*dens(i)/PLI/rhow) 
     &							then
							rf = -T(i)*csnow*dz(i)*dens(i)/PLI/dt/rhow 
							T(i) = 0.0
							wl(i) = wl(i) - rf*dt
						else
							rf = wl(i)/dt
			       T(i) = T(i) + rf*rhow*dt*PLI/(csnow*dz(i)*dens(i))    
							wl(i) = 0.
						end if
						Eout = 0.
					end if
					if(T(i).ge.0.0001)then  !0.0001
						if(csnow*dens(i)*T(i)*dz(i)/PLI .le.
     &						dz(i)*dens(i)) then
						smelti = csnow*dens(i)*T(i)*dz(i)/dt/PLI/rhow 
							Eout = 0.
						else
							smelti = dz(i)*dens(i)/dt/rhow  
							Eout = csnow*dens(i)*T(i)*dz(i)/dt
     &                              - smelti*PLI*rhow 
							dzi = 0.
						end if
						T(i) = 0.0
						wl(i) = wl(i) + smelti*dt
					endif
				endif
              else

Cc	E = 0 (E<0 is impossible case)
				if(T(i).ge.0.0001) then     !0.0001
					if(csnow*dens(i)*T(i)*dz(i)/PLI .le.
     &                        dz(i)*dens(i)) then
						smelti = csnow*dens(i)*T(i)*dz(i)/dt/PLI/rhow  
						T(i) = 0.0
						Eout = 0.
						wl(i) = wl(i) + smelti*dt
					else
						smelti = dz(i)*dens(i)/dt/rhow  
						Eout = csnow*dens(i)*T(i)*dz(i)/dt
     &                            - smelti*PLI*rhow  
						T(i) = 0.0
						dzi = 0.
						wl(i) = wl(i) + smelti*dt
					end if
				else
Cc	T<0
				if(wl(i) .gt. -csnow*dens(i)*T(i)*dz(i)/PLI/rhow) then  
						rf = -csnow*dens(i)*T(i)*dz(i)/dt/PLI/rhow   
						T(i) = 0.0
						wl(i) = wl(i) - rf*dt
					else
						rf = wl(i)/dt
						wl(i) = 0.
					T(i) = T(i) + rf*rhow*dt*PLI/(csnow*dz(i)*dens(i))
					end if
					Eout = 0.
					if(T(i).ge.0.0001) then   !0.0001
						if(csnow*dens(i)*T(i)*dz(i)/PLI .le.
     &                         dz(i)*dens(i)) then
						smelti = csnow*dens(i)*T(i)*dz(i)/dt/PLI/rhow 
							Eout = 0.0
						else
							smelti = dz(i)*dens(i)/dt/rhow
							Eout = csnow*dens(i)*T(i)*dz(i)/dt
     &                            - smelti*PLI*rhow 
							dzi = 0.
						end if
						T(i) = 0.0
						wl(i) = wl(i) + smelti*dt
					endif
				endif
			endif
	
Cc
Cc
Cc	internal q
Cc
			rho_dry = (dens(i)-w0/(dz(i))*RHOW)/
     *			(1-w0/(dz(i)))
			if(rho_dry.lt.0.) WRITE(*,*) 'rho_dry = ', rho_dry

!     PorosityOfSnow = 1 - rho_dry/rhoi - wl(i)/dz(i)*(1-rhow/rhoi) wl(i)/rhow/dz(i)
	       PorosityOfSnow = 1 - rho_dry/rhoi - wl(i)/dz(i)
			PorosityOfSnow = dmax1(PorosityOfSnow,whc + 0.1)
			p1 = PorosityOfSnow - whc   ! whc = water holding capacity
			fukt = wl(i)/dz(i)
			if(fukt.gt.whc)then
				fukt = (fukt-whc)/p1
				q0 = hcond*fukt**3.0
	
				qbase = dmin1(q0,wl(i)/dt)
				wl(i) = wl(i) - qbase*dt
			else
				qbase = 0.0
				q0 = 0.0
			end if
			if(dzi .le. 0.000001) then   
				qbase = qbase + wl(i)/dt
				dz(i)=dzi
			else
				dzmasi=dz(i)*dens(i)-qbase*dt*rhow   
	            dzi=dz(i)+rhow*(-qbase/RHOW-rf/RHOW+rf/rhoi-  
     &                    smelti/rho_dry+smelti/RHOW)*dt     
				dens(i)=dmax1(dzmasi/dzi,0.d0)
				dz(i)=dzi
				if(dens(i).gt.rhoi)then
					dz(i)=dz(i)*dens(i)/rhoi
					dens(i)=rhoi
				end if
			end if
			Phase = Phase                                 !diagnostic variable
     &    			+ (t(i)-ti(i))*dens(i)*dz(i)/snow_sc/dt
	
			if(i.eq.ms-1) snmelt=qbase
		end do
      end if
	if (itop==ms-1.and.dz(itop)<0.000001) then
	 dz(itop)=0
	 dens(itop)=0
	 goto 123
	endif
      if(itop .lt. ms .and. dz(itop) .le. 0.000001) then   
		dz(itop)=0.
		dens(itop)=0.
		itop=itop+1
      end if
		
222   do i=ms-1,itop,-1
		if(dz(i).le.0.000001) then   
			do k=i-1,itop,-1
				if(dz(k).gt.0.000001) goto 111   
			end do
	
111			do j=i,itop,-1
				dz(j)=dz(k+j-i)
				dens(j)=dens(k+j-i)
				t(j)=t(k+j-i)
				wl(j)=wl(k+j-i)
			end do
			itop=itop-k+i
			goto 222
		end if
      end do
123   continue
	 
333   do i=itop,ms-1,1
		if(dz(i).lt.dzmin.and.itop.lt.ms-1) then
		if(i.lt.ms-1) then
			dens(i+1)=dens(i+1)*dz(i+1)/(dz(i)+dz(i+1))+
     &				dens(i)*dz(i)/(dz(i)+dz(i+1))
			t(i+1)=t(i+1)*dz(i+1)/(dz(i)+dz(i+1))+
     &				t(i)*dz(i)/(dz(i)+dz(i+1))
			wl(i+1)=wl(i+1)+wl(i)
			dz(i+1)=dz(i+1)+dz(i)
			do j=i,itop+1,-1
				dens(j)=dens(j-1)
				t(j)=t(j-1)
				wl(j)=wl(j-1)
				dz(j)=dz(j-1)
			end do
			itop=itop + 1
			goto 333
		end if
		end if
      end do
	
      swe = 0.
      hsnow = 0.
      do j=ms-1,itop,-1
          swe = swe + dz(j)*dens(j)
          hsnow = hsnow + dz(j)
      end do
       
cccccccccccccc snow densification due to gravity and metamorphism cccccccccccccc
	a2 = 0.00000066
	sigma = 75.                             ! metamorphic processes, Pa
	do j = itop+1,ms-1
		if(dens(j).lt.3*densny) then
			dens_old = dens(j)
			eta =                           ! compactive viscosity of snow 
     &			a2*dexp(19.3*dens(j)/rhoi)*
     &			dexp(67300/8.314/(t(j)+273.15))
			P = 0.0                         ! gravity, Pa
			do i = j,itop,-1
				!P = P + dens(i)/1000.*grav*(dz(i))*10000.
	            P = P + dens(i)*grav*dz(i)
			end do
			dens(j) = dens(j) + 0.001*(dt*dens(j)*1000.*(sigma+P)/eta)
                          dens(j) = dmin1(dens(j),rhoi)
			dz(j) = dz(j) * dens_old/dens(j)
		end if
	end do
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      swe = 0.
      hsnow = 0.
      do j=ms-1,itop,-1
          swe = swe + dz(j)*dens(j)
          hsnow = hsnow + dz(j)
      end do
      !do m = itop,ms-1
	!	por(m) = 1 - dens(m)/rhoi
      !end do
	
      snow_sc = swe

        !! dz(ms) !!
	if(itop.ne.ms-1) then
		dz(ms) = dz(ms-1)/3.
		dz(ms-1) = dz(ms-1) - dz(ms)
	else
		dz(ms) = dz(ms-1)/2.
		dz(ms-1) = dz(ms-1) - dz(ms)
	end if
	if(itop.eq.ms-1) then
		z(ms-1) = z(ms) - dz(ms-1) - dz(ms)
	else
		z(ms-1) = z(ms) - dz(ms-1)/2. - dz(ms)
	end if
	DO j = MS-2,itop+1,-1
		Z(j) = z(j+1)-dz(j)/2.-dz(j+1)/2.
      END DO
	if(itop.lt.ms-1) z(itop) = z(itop+1) - dz(itop+1)/2. - dz(itop)

	!do i=itop,ms-1
!		wl(i) = wl(i)/dz(i)/dens(i)*rhow
!	end do
	!extinct(itop)=0
	do i = itop,ms-1
		!extinct(i) = (0.9 - 0.4/0.45*(dens(i) - 0.05))*100.
	 extinct(i) = dexp(-(0.13*dens(i)+3.4))
	 !extinct(i) = 0.
	end do
	
	do i = 1, itop-1
		dz(i) = 0.
      end do

      
      precip = precip/dt
	!eflux = eflux+Erad
	
      return
	END

      SUBROUTINE addPrecipToSnow(T2,ts,snow,iyear,imonth,
     &	iday,CCT,pmelt,pheat,dt)
      use phys_constants    
	use phys_parameters
	use numeric_params
	use driving_params
	use atmos
	
	implicit none
CCCCCCCCCCCCCCC input parameters:
CC      e, cal/(cm2*s), --- surface heat balance: e = Radiation-SIGMA*(TS**4)-Latent-Sensible-InSoil
CC      Elatent, cal/(cm2*s), --- latent heat flux.
CC      ts, deg C, --- surface temperature.
CC      precip, cm --- precipitation during dt.
CC
CCCCCCCCC  some other parameters: CCCCCCCCCCC
CC      densny, g/cm3       --- density of fresh snow, depends on surface temperature
CC      dznorm, cm          --- initial thickness of snow layers
CC      dzmin, cm           --- minimal thickness of snow layers
CC      dt, sec             --- timestep
CC      ms                  --- max. number of layers in snowpack
CC      cwat, cal/(gr K)    --- the water specific heat content
CC      csnow, cal/(gr K)   --- the snow specific heat content
CC      rhow, gr/cm3        --- the liquid water density
CC      PLI,  cal/g          --- latent heat for freezing/melting
CC      PL,  cal/g          --- latent heat for evapor./condens.
CC      whc                 --- hydraulic constant
CC      hcond, cm/sec       --- hydraulic conductivity
CC
CCCCCCCCCCC  values to be calculate after every timestep:
CC      wl(i), cm       --- moisture content in snow layer No.i
CC      t(i), deg.C     --- mean temperature of snow layer No.i
CC      dz(i), cm       --- thickness of snow layer No.i
CC      dens(i), g/cm3  --- mean density of snow layer No.i
CC      hsnow, cm       --- snow depth
CC      swe, cm         --- water equivalent snow depth
CC      snmelt, cm/sec  --- mean snow melting speed during dt
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

      COMMON /SOILDAT/ dz(ms),itop
	COMMON /BL/       ST,PGR,TGROLD,QGROLD,RADIAT,WSOLD,SNOLD,
     &                  ZS,thsoil,whsoil,BOLD,RF1,RF2,SNMELT,HSNOW,
     &                  HS,ES,TGRNEW,QGRNEW,WSNEW,SNNEW,RUNOFF,
     &                  ElatOld,HFold,PRSold,extinct(ms)
	common /atmos2/ shortwave, precip
	COMMON /SOILSOL/ AL(ML),DLT(ML),DVT(ML),ALLL(ML),DL(ML),
     &                  ALV(ML),DV(ML),Z(ML),T(ML),WL(ML),WV(ML),WI(ML),
     &                  dens(ms)
	COMMON /spin/ Cond,Phase
	common /layers/ h1,l1,hs1,ls1
     	real(8) dz
	real(8) ST,PGR,TGROLD,QGROLD,RADIAT,PRECIP,WSOLD,SNOLD,
     &ZS,thsoil,whsoil,BOLD,RF1,RF2,SNMELT,HSNOW,HS,ES,TGRNEW,QGRNEW,
     &WSNEW,SNNEW,RUNOFF,ElatOld,HFold,PRSold
	real(8) AL,DLT,DVT,ALLL,DL,ALV,DV,Z,T,WL,WV,WI,dens
	real(8) extinct,ts,snow,dz0,pmelt,pheat,
     & z0z,dmass, swe,densny,shortwave,Cond,Phase,h1,l1,hs1,ls1
	INTEGER(4) j,i,k,mn,itop,iyear,imonth,iday
	real(8) T2,CCT(ML)
      real(8) dt 

	SAVE
      
      precip = precip*dt
		
            
	      !if (snow > 0.) then
			! write (*, *) snow
			! STOP
			!end if
	!T2 = T2 - 273.15
c	if(T2 .le. -10) then
c		densny = 0.05
c	elseif(ts .le. -2) then
c		densny = 5./800.*(t2+18.)
cc		densny = 0.1
c	else
c		densny = (t2+4.)/20.
c	end if
	densny = (67.9 + 51.3 * dexp(t2/2.6))
	if(itop.lt.ms) 	dz(ms-1) = dz(ms-1) + dz(ms) 
	
	snow = 0.
	do mn = itop,ms-1
		snow = snow + dz(mn)*dens(mn)
	end do
	
	
	j=itop
555	continue
	if(dz(itop).lt.dznorm+dzmin) goto 444
	dz(j-1) = dz(j) - dznorm
	dz(j) = dznorm
	z(j-1)= z(itop) 
	z(j)= z(j-1) + dz(j-1)*2.
	wl(j-1)=0.0
	dens(j-1)=dens(j)
	t(j-1)=t(j)
	j=j-1
	itop=j
	goto 555
444	continue
      
	if(T2.gt.0.) then
		pmelt = precip/dt                       
		!pmelt = 0.                       
		pheat = T2*precip*cwat*rhow/dt	    
	else	   
	    pmelt=0.0
		pheat=0.0
		if(itop.eq.ms .and. precip.gt.0.0) then
			itop=itop-1
			z(itop) = z(itop+1)
			t(itop)=t(itop+1)
			dens(itop)=densny
			wl(itop)=0
		end if
		if(itop.ne.ms) then
			dz0 = dz(itop)
			z(itop) = z(itop) - precip*rhow/densny   
			dz(itop) = dz(itop) + precip*rhow/densny   
			if(dz(itop) .lt. dznorm+dzmin) then
				z0z=dz0/dz(itop)
				dens(itop)=z0z*dens(itop)+(1.0-z0z)*densny
			end if
	        j=itop
55			continue
			if(dz(itop) .lt. dznorm+dzmin) goto 54
			dz(j-1) = dz(j) - dznorm
			dz(j) = dznorm
			z(j-1)= z(itop) 
			z(j)= z(itop) - dz(itop)*0.5
			z0z=dz0/(dz(itop)+dz(itop-1))
			dens(itop)=z0z*dens(itop)+(1.0-z0z)*densny
			wl(j-1)=0.0
			dz0 = 0.0
			dens(j-1)=densny
			t(j-1)=t(j)
			j=j-1
			itop=j
			goto 55
54			continue
      	end if
	end if
		
	if(pheat.gt.0.0) then
		if(itop.eq.ms-1) then
			dmass = dens(itop)*dz(itop) + precip*rhow    
			dz(itop) = dz(itop) + precip*rhow/rhoi   
			if(dz(itop).eq.0.) dz(itop) = 0.000001   
			dens(itop)= dmax1(dmass/dz(itop),0.d0)
			if(dens(itop).gt.rhoi) then
				dz(itop) = dz(itop)*dens(itop)/rhoi
				dens(itop)=rhoi
			end if
		else
			if(itop.lt.ms-1) then
				mn = itop+1
				dmass = dens(mn)*dz(mn) + precip*rhow  
				dz(mn) = dz(mn) + precip*rhow/rhoi 
				if(dz(mn).eq.0.) dz(mn) = 0.000001    
				dens(mn)= dmax1(dmass/dz(mn),0.d0)
				if(dens(mn).gt.rhoi) then
	                       		dz(mn) = dz(mn)*dens(mn)/rhoi
					dens(mn)=rhoi
				end if
				wl(mn) = wl(mn) + precip
		T(mn) = T(mn)*dz(mn)*dens(mn)/(precip*rhow+dz(mn)*dens(mn)) 
			end if
		end if
	end if

        !!  dz(ms) !!!
	if(itop.ne.ms-1) then
		dz(ms) = dz(ms-1)/3.
		dz(ms-1) = dz(ms-1) - dz(ms)
	else
		dz(ms) = dz(ms-1)/2.
		dz(ms-1) = dz(ms-1) - dz(ms)
	end if
	if(itop.eq.ms-1) then
		z(ms-1) = z(ms) - dz(ms-1) - dz(ms)
	else 
		z(ms-1) = z(ms) - dz(ms-1)/2. - dz(ms)
	end if
	DO Mn = MS-2,itop+1,-1
		Z(Mn) = z(mn+1)-dz(mn)/2.-dz(mn+1)/2.
      END DO
	if(itop.lt.ms-1) z(itop) = z(itop+1) - dz(itop+1)/2. - dz(itop)

	!if (dz(itop).lt.dznorm+dzmin) goto 777 
	
	swe = 0.0
	do mn=ms-1,itop,-1
		swe=swe+dz(mn)*dens(mn)
	end do
	swe = swe + dz(ms)*dens(ms-1)
	snow = swe
	
	hsnow = 0.
	do i = itop, ms
       hsnow = hsnow + dz(i)
	end do
		   
	precip = precip/dt
	
	return
	END


      SUBROUTINE comsoilforlake
      
!     COMSOILFORLAKE specifies parameters of soil according to soil type      
      
      USE numeric_params
	use driving_params
	use arrays
	implicit none


      real(8),dimension(1:Num_Soil):: b,psi_max, 
     & Porosity,gamma_max,lambda_max,W_0,W_m
      real(8) pow
     	  
      integer(4) i,j
	logical firstcall

	data firstcall /.true./

      SAVE
      if (firstcall) then
!       allocate (AL(1:ns+2),DLT(1:ns+2),DVT(1:ns+2),DL(1:ns+2),
!     & ALV(1:ns+2),DV(1:ns+2),Z(1:ns+2),T(1:ns+2),WL(1:ns+2),
!     & WV(1:ns+2),WI(1:ns+2),dens(1:ns+2))
	endif
	
!     I=1       ! SAND
!     I=2       ! LOAMY SAND
!     I=3       ! SANDY LOAM
!     I=4       ! LOAM
!     I=5       ! SILT LOAM
!     I=6       ! SANDY CLAY LOAM
!     I=7       ! CLAY LOAM
!     I=8       ! SILTY CLAY LOAM
!     I=9       ! SANDY CLAY
!     I=10      ! SILTY CLAY
!     I=11      ! CLAY
!----------------------------------------------------------------
!  NN    b     psi_max  Por    gamma_max lambda_max W_0     W_m
!----------------------------------------------------------------
!   1   4.05    3.50    0.395   0.01760   0.29800   0.02    0.01
!   2   4.38    1.78    0.410   0.01560   0.13900   0.05    0.02
!   3   4.90    7.18    0.435   0.00340   0.13700   0.08    0.03
!   4   5.39    14.6    0.451   0.00069   0.06190   0.20    0.08
!   5   5.30    56.6    0.485   0.00072   0.20400   0.18    0.07
!   6   7.12    8.63    0.420   0.00063   0.03070   0.13    0.06
!   7   8.52    36.1    0.476   0.00024   0.03720   0.27    0.12
!   8   7.75    14.6    0.477   0.00017   0.01240   0.24    0.11
!   9   10.4    6.16    0.426   0.00021   0.00641   0.23    0.10
!  10   10.4    17.4    0.492   0.00010   0.00755   0.30    0.15
!  11   11.4    18.6    0.482   0.00013   0.00926   0.40    0.20
!-----------------------------------------------------------------

      zsoil(1) = 0.

	if(UpperLayer > 0.) then
	 pow = dlog(UpperLayer/depth)/dlog(1.d0/float(ns-1))
       do i=2,ns
	  zsoil(i) = (1.*i/(ns-1))**pow*depth
	 end do
	else
	 do i=2,ns
	  zsoil(i) = (1.*i/(ns-1))*depth
	 end do
	end if
	
	do i=1, ns-1
	 dzs(i)=zsoil(i+1)-zsoil(i)
      end do

	do i=1, ns
	 if (i==1) then
        dzss(i) = dzs(i)/2
       elseif (i==ns) then
	  dzss(i) = dzs(i-1)/2
       else
	  dzss(i) = (dzs(i-1)+dzs(i))/2
       endif 
      end do 
	
!      SoilType = 7

      b(1) = 4.05
      b(2) = 4.38
      b(3) = 4.90
      b(4) = 5.39
      b(5) = 5.30
      b(6) = 7.12
      b(7) = 8.52
      b(8) = 7.75
      b(9) = 10.4
      b(10)= 10.4
      b(11)= 11.4

      psi_max( 1) = 3.50/100. 
      psi_max( 2) = 1.78/100. 
      psi_max( 3) = 7.18/100. 
      psi_max( 4) = 14.6/100. 
      psi_max( 5) = 56.6/100. 
      psi_max( 6) = 8.63/100. 
      psi_max( 7) = 36.1/100. 
      psi_max( 8) = 14.6/100. 
      psi_max( 9) = 6.16/100. 
      psi_max(10) = 17.4/100. 
      psi_max(11) = 18.6/100. 

      Porosity( 1) = 0.395
      Porosity( 2) = 0.410
      Porosity( 3) = 0.435
      Porosity( 4) = 0.451
      Porosity( 5) = 0.485
      Porosity( 6) = 0.420
      Porosity( 7) = 0.476
      Porosity( 8) = 0.477
      Porosity( 9) = 0.426
      Porosity(10) = 0.492
      Porosity(11) = 0.482

      gamma_max( 1) = 0.01760/100. 
      gamma_max( 2) = 0.01560/100. 
      gamma_max( 3) = 0.00340/100. 
      gamma_max( 4) = 0.00069/100. 
      gamma_max( 5) = 0.00072/100. 
      gamma_max( 6) = 0.00063/100. 
      gamma_max( 7) = 0.00024/100. 
      gamma_max( 8) = 0.00017/100. 
      gamma_max( 9) = 0.00021/100. 
      gamma_max(10) = 0.00010/100. 
      gamma_max(11) = 0.00013/100. 

      lambda_max( 1) = 0.29800/10000. 
      lambda_max( 2) = 0.13900/10000.
      lambda_max( 3) = 0.13700/10000.
      lambda_max( 4) = 0.06190/10000.
      lambda_max( 5) = 0.20400/10000.
      lambda_max( 6) = 0.03070/10000.
      lambda_max( 7) = 0.03720/10000.
      lambda_max( 8) = 0.01240/10000.
      lambda_max( 9) = 0.00641/10000.
      lambda_max(10) = 0.00755/10000.
      lambda_max(11) = 0.00926/10000.

      W_0( 1) = 0.02
      W_0( 2) = 0.05
      W_0( 3) = 0.08
      W_0( 4) = 0.20
      W_0( 5) = 0.18
      W_0( 6) = 0.13
      W_0( 7) = 0.27
      W_0( 8) = 0.24
      W_0( 9) = 0.23
      W_0(10) = 0.30
      W_0(11) = 0.40

      W_m( 1) = 0.01
      W_m( 2) = 0.02
      W_m( 3) = 0.03
      W_m( 4) = 0.08
      W_m( 5) = 0.07
      W_m( 6) = 0.06
      W_m( 7) = 0.12
      W_m( 8) = 0.11
      W_m( 9) = 0.10
      W_m(10) = 0.15
      W_m(11) = 0.20

	if(SoilType>0) then
	 do j = 1, ns
	  BH(j)     = b(SoilType)               ! PARAMETER B, DIMENSOINLESS
	  PSIMAX(j) = - psi_max(SoilType)       ! SAT. WATER POTENTIAL, M.          
    	  POR(j)    = Porosity(SoilType)        ! POROSITY, DIMENSIONLESS
	  FLWMAX(j) = gamma_max(SoilType)       ! SAT. HYDR. CONDUCTIVITY, M/S
	  DLMAX(j)  = lambda_max(SoilType)      ! SAT. WATER DIFFUSIVITY, ?KG/(M*SEC)
	  WLM0(j)   = W_0(SoilType)             ! MAXIMAL UNFREEZING WATER AT 0C
	  WLM7(j)   = W_m(SoilType)             ! MAXIMAL UNFREEZING WATER AT T<<0C
	 end do
	else
!	 do j = 1, ns
!	  BH(j)     = BH_soil           ! PARAMETER B, DIMENSOINLESS
!	  PSIMAX(j) = PSIMAX_soil       ! SAT. WATER POTENTIAL, M.          
!    	  POR(j)    = POR_soil          ! POROSITY, DIMENSIONLESS
!	  FLWMAX(j) = FLWMAX_soil       ! SAT. HYDR. CONDUCTIVITY, M/S
!	  DLMAX(j)  = DLMAX_soil        ! SAT. WATER DIFFUSIVITY, ?KG/(M*SEC)
!	  WLM0(j)   = WLM0_soil         ! MAXIMAL UNFREEZING WATER AT 0C
!	  WLM7(j)   = WLM7_soil         ! MAXIMAL UNFREEZING WATER AT T<<0C
!	 end do
	end if
      
	if (firstcall) firstcall=.false.   
      RETURN
      END

      SUBROUTINE PBLDAT

!     PBLDAT assigns values to parameters used in surface layer parameterization
!     (see subr. dragvl and subr., called by dragvl)

      IMPLICIT real (A-H,O-Z)
C
C*=====================================================================
C*       INITIALIZATION OF PBL COMMON BLOCKS (FBL,DRAG)               =
C*    FBL(1)     :  FOR LINEAR EXTRAPOLATION OF WIND ONTO SURFACE     =
C*    FBL(2)     :  ---------------------------------------------     =
C*    FBL(3)     :  PRESCRIBED MINIMAL VALUE OF SURFACE WIND MODULE   =
C*    FBL(4)     :  HEIGHT OF CONSTANT FLUX LAYER ( M )               =
C*    FBL(5)     :  CRITICAL VALUE OF RELATIVE HUMIDITY IN            =
C*               :  CALCULATION OF EQUIVALENT POTENTIAL TEMPERATURE   =
C*    FBL(6)     :  SEA WATER TEMPERATURE UNDER ICE ( DEG. K )        =
C*    FBL(7)     :  PRECRIBED MINIMAL VALUE OF SEA ICE SURFACE        =
C*                  TEMPERATURE ( DEG. K )                            =
C*    FBL(8)     :  RESERVED                                          =
C*    FBL(9)     :  PBL WIND TURNING ANGLE OVER OCEAN ( DEG )         =
C*    FBL(10)    :  --------------------------- ICE ----------------- =
C*    FBL(11)    :  --------------------------- SNOW COVERED SOIL --- =
C*    FBL(12)    :  --------------------------- BARED SOIL ---------- =
C*    FBL(13)    :  --------------------------- IN TROPICS ---------- =
C*    FBL(14)    :  BOUNDARIES OF TROPICS ( RADIANS )                 =
C*    FBL(15)    :  RESERVED
C*    FBL(16)    :  (SNOW COVERED SOIL HEAT CONDUCTIVITY)/DEPTH ----- =
C*    FBL(17)    :  (SEA ICE HEAT CONDUCTIVITY)/(SEA ICE DEPTH) ----- =
C*    FBL(18)    :  (SOIL HEAT CONDUCTIVITY)/DEPTH, CAL/(DEG.*M**2*S) =
C*    FBL(19)    :  BULK SOIL HEAT CAPACITY, CAL/(DEG.*M**2)          =
C*    FBL(20)    :  (SOIL MOISTURE CONDUCTIVITY)/DEPTH,SEC**-1        =
C*=====================================================================
      COMMON /FBL/ FBL(20)
      COMMON /BL1/ AZ0P,VEG,WL,DZZ
      COMMON /HYDR/ SNCR,WLMMX,WSMAX,WSL,WSG,CEFF,
     &              CA,DZL,DZG,BMIN,WSDL,WSDG,DMIN,DMAX,D,HR,WIINF,
     &              ZRM,ZRMM,TRM,FLXMIN,TOMIN,HSNold
      COMMON /VEGSW/ CCQ,CK,SWW,TL
      common /p_drag/ vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,
     &                alpha_m,alpha_t,beta_m,beta_t,gamma_m,gamma_t,
     &                zeta_st,a_m,a_t,b_m,b_t
	real FBL,AZ0P,VEG,WL,DZZ,SNCR,WLMMX,WSMAX,WSL,WSG,CEFF,
     &CA,DZL,DZG,BMIN,WSDL,WSDG,DMIN,DMAX,D,HR,WIINF,ZRM,ZRMM,TRM,
     &FLXMIN,TOMIN,CCQ,CK,SWW,TL,
     &vkc,anu,z0min,chc,p_m,p_t,q_m,q_t,alpha_m,alpha_t,beta_m,
     &beta_t,gamma_m,gamma_t,zeta_st,a_m,a_t,b_m,b_t

      FBL(1) = 1.
      FBL(2) = 1.
      FBL(3) = 1.
      FBL(4) = 70.
      FBL(5) = 1.
      FBL(6) = 271.5
      FBL(7) = 243.2
      FBL(8) = 0.
      FBL(9) = 20.
      FBL(10) = 10.
      FBL(11) = 30.
      FBL(12) = 30.
      FBL(13) = 0.
      FBL(14) = 20.*3.1415926536/180.
C     FBL(8)...FBL(14) WILL BE NOT USED IN THE MULTILEVEL PBL
      FBL(15) = 0.
      FBL(16) = 0.04/2.
      FBL(17) = 0.5/3.
      FBL(18) = 0.6/2.
      FBL(19) = 4.0E+04
      FBL(20) = 1.0E-6/5.
c------------------------------------------------------------------
c*    subroutine dragvl constants
c------------------------------------------------------------------
      vkc = 0.4
      anu = 0.000015
      z0min = 1.5e-5
      chc = 0.0132
      zeta_st = -2.
      p_m = - 0.25
      p_t = - 0.5
      q_m = - 1./3.
      q_t = - 1./3.
      alpha_m = 1.
      alpha_t = 1.
      beta_m = 4.7
      beta_t = 4.7
      gamma_m = 16.
      gamma_t = 16.
      a_m = (1. + gamma_m*(p_m/q_m - 1.)*zeta_st)
     &      *(1. - gamma_m*zeta_st)**(p_m/q_m - 1.)
      b_m = a_m*gamma_m*(p_m/q_m)/(1. + gamma_m*(p_m/q_m - 1.)*zeta_st)
      a_t = (1. + gamma_t*(p_t/q_t - 1.)*zeta_st)
     &      *(1. - gamma_t*zeta_st)**(p_t/q_t - 1.)
      b_t = a_t*gamma_t*(p_t/q_t)/(1. + gamma_t*(p_t/q_t - 1.)*zeta_st)

      RETURN
      END
            

      REAL(8) FUNCTION unfrwat(T,i)
	use driving_params
	use arrays
	
!    CALCULATION OF LIQUID WATER CONTENT IN FREEZING SOIL
!    T - DEG. C; WLM0,WLM7,WLMAX - KG/KG

      implicit none
	real(8) T
	integer(4) i

      unfrwat = (WLM0(i)-WLM7(i))*dexp(T/3.) + WLM7(i)
      !unfrwat = 0 
      
      RETURN
      END


	REAL(8) FUNCTION QS(phase,t,p)
	
!     QS - specific humidity, kg/kg, for saturated water vapour

	implicit none
	real(8) t,p,aw,bw,ai,bi,a,b,es
	integer(4) phase
!phase = 1 - liquid, 0 - ice
      aw = 7.6326    
      bw = 241.9    
      ai = 9.5  
      bi = 265.5 
	a  = phase*aw+(1-phase)*ai
	b  = phase*bw+(1-phase)*bi
      es=610.7*10.**(a*t/(b+t))
      QS=0.622*es/p
	END FUNCTION


	REAL(8) FUNCTION ES(phase,t,p)
	
!     ES is pressure of saturated water vapour, Pa	

	implicit none
	real(8) t,p,aw,bw,ai,bi,a,b
	integer(4) phase
!phase = 1 - liquid, 0 - ice
      aw = 7.6326    
      bw = 241.9    
      ai = 9.5  
      bi = 265.5 
	a  = phase*aw+(1-phase)*ai
	b  = phase*bw+(1-phase)*bi
      ES = 610.7*10.**(a*t/(b+t))
      END FUNCTION
      
      REAL(8) FUNCTION MELTPNT(C)
      implicit none
      real(8) C, dtdc
      
      SAVE
      
      dtdc = 66.7
      
      MELTPNT = 0. - C*dtdc
      END FUNCTION MELTPNT

      REAL(8) FUNCTION WL_MAX(por,rosdry,wi)
	implicit none
	real(8) por,rosdry,wi,row0,roi
	row0=1000.
	roi=917.
	wl_max=por*row0/(rosdry*(1-por)) - wi*row0/roi
	END FUNCTION WL_MAX

      REAL(8) FUNCTION WI_MAX(por,rosdry)
	implicit none
	real(8) por,rosdry,roi
	roi=917.
	wi_max=por*roi/(rosdry*(1-por))
	END FUNCTION WI_MAX
	
	SUBROUTINE progonka (a, b, c, f, y, N)
	implicit none
	
!      FACTORIZATION METHOD FOR THE FOLLOWING SYSTEM OF LINEAR EQUATIONS:
!     -a(i)*y(i-1)+c(i)*y(i)-b(i)*y(i+1)=f(i) i=2,N-1
!      c(1)*y(1)-b(1)*y(2)=f(1)
!     -a(N)*y(N-1)+c(N)*y(N)=f(N)
!
      integer(4), parameter:: M=350
      integer(4) N, i
      real(8) a(M), b(M), c(M), f(M), y(M)
	real(8) alpha(M+2), beta(M+2) 

	SAVE
	 	
	!y=0
      alpha(2) = b(1)/c(1)
	beta(2) = f(1)/c(1)
	do i = 3, N+1 
       alpha(i) = b(i-1)/(c(i-1)-a(i-1)*alpha(i-1))
	 beta(i) = (f(i-1)+a(i-1)*beta(i-1))/ 
     & (c(i-1)-a(i-1)*alpha(i-1))
      end do
	y(N) = beta(N+1)
	do i = N-1, 1, -1
	 y(i) = alpha(i+1)*y(i+1)+beta(i+1)
      end do
	!write (*, *) (c(i)*y(i)-b(i)*y(i+1)-a(i)*y(i-1)-f(i),i=2,N-1) !/f(10)
 
      END 
      
	SUBROUTINE progonka2_db (a, b, c, f, y, K, N)
	implicit none
!FACTORIZATION METHOD FOR THE FOLLOWING SYSTEM OF LINEAR EQUATIONS:
!     -a(i)*y(i-1)+c(i)*y(i)-b(i)*y(i+1)=f(i) i=K+1,N-1
!      c(K)*y(K)-b(K)*y(K+1)=f(K)
!     -a(N)*y(N-1)+c(N)*y(N)=f(N)
!
      integer(4), parameter:: M=350
      integer(4) K, N, i
      real(8) a(M), b(M), c(M), f(M), y(M)
	real(8) alpha(M+2), beta(M+2) 
	SAVE
				
      alpha(K+1) = b(K)/c(K)
	beta(K+1) = f(K)/c(K)
	do i = K+2, N+1 
       alpha(i) = b(i-1)/(c(i-1)-a(i-1)*alpha(i-1))
	 beta(i) = (f(i-1)+a(i-1)*beta(i-1))/ 
     & (c(i-1)-a(i-1)*alpha(i-1))
      end do
	y(N) = beta(N+1)
	do i = N-1, K, -1
	 y(i) = alpha(i+1)*y(i+1)+beta(i+1)
      end do
	 
      END 

	SUBROUTINE progonka2(a, b, c, f, y, K, N)
	implicit none
!FACTORIZATION METHOD FOR THE FOLLOWING SYSTEM OF LINEAR EQUATIONS:
!     -a(i)*y(i-1)+c(i)*y(i)-b(i)*y(i+1)=f(i) i=K+1,N-1
!      c(K)*y(K)-b(K)*y(K+1)=f(K)
!     -a(N)*y(N-1)+c(N)*y(N)=f(N)
!
      integer(4), parameter:: M=350
      integer(4) K, N, i
      real(8) a(M), b(M), c(M), f(M), y(M)
	real(8) alpha(M+2), beta(M+2) 
	SAVE
				
      alpha(K+1) = b(K)/c(K)
	beta(K+1) = f(K)/c(K)
	do i = K+2, N+1 
       alpha(i) = b(i-1)/(c(i-1)-a(i-1)*alpha(i-1))
	 beta(i) = (f(i-1)+a(i-1)*beta(i-1))/ 
     & (c(i-1)-a(i-1)*alpha(i-1))
      end do
	y(N) = beta(N+1)
	do i = N-1, K, -1
	 y(i) = alpha(i+1)*y(i+1)+beta(i+1)
      end do
	 
      END 


      SUBROUTINE ind_stab_fact (a,b,c,N,M,ind_stab,ind_bound)
      
!     IND_STAB_FACT checks, if the stability condition for factorization method
!     is satisfied      
	
	implicit none
	real(8), dimension(1:350):: a,b,c
      integer(4) M,i,N
      logical ind_stab, ind_bound 

	SAVE

	ind_stab=.true.
 	if (ind_bound==.true.) then 
	 if (dabs(b(N))>=dabs(c(N)).or.dabs(a(M))>=dabs(c(M))) then
	  ind_stab=.false.
	  RETURN
       endif
      endif
	do i=N+1,M-1
	 if (dabs(a(i))+dabs(b(i))>=dabs(c(i))) then
	  ind_stab=.false.
	  RETURN
	 endif
	enddo
	
	END

	
	SUBROUTINE ind_stab_fact_db (a,b,c,N,M,ind_stab,ind_bound)

	implicit none
	real(8), dimension(1:350):: a,b,c
      integer(4) M,i,N
      logical ind_stab, ind_bound 

	SAVE

	ind_stab=.true.
	if (ind_bound==.true.) then 
	 if (dabs(b(N))>=dabs(c(N)).or.dabs(a(M))>=dabs(c(M))) then
	  ind_stab=.false.
	  RETURN
       endif
      endif
	do i=N+1,M-1
	 if (dabs(a(i))+dabs(b(i))>=dabs(c(i))) then
	  ind_stab=.false.
	  RETURN
	 endif
	enddo
	
	END

	SUBROUTINE CUT_E_DB(E,N,porog)
	implicit none
      real(8) E(350),porog
	integer(4) N,i
!	porog = 10.**(-9) 
	do i=1,N 
	 if (E(i)<porog) E(i)=porog
  	enddo
      END SUBROUTINE CUT_E_DB

      SUBROUTINE CUT_EPS_DB(eps,N,porog)
	implicit none
      real(8) eps(350),porog
	integer(4) N,i
!	porog = 10.**(-11) 
	do i=1,N
	 if (eps(i)<porog) eps(i)=porog
  	enddo
      END SUBROUTINE CUT_EPS_DB
 
      
      SUBROUTINE MATRIXSUM(a,b,c,k)
	implicit none
	
!     MATRIXES: C=A+B
	real(8), dimension (350,2,2)::a,b,c
	integer(4) k,j,i

      do i=1,2
	 do j=1,2
	  c(k,i,j)=a(k,i,j)+b(k,i,j)
       enddo
      enddo
	
	END SUBROUTINE MATRIXSUM

	SUBROUTINE MATRIXMULT(a,b,c,k)
	implicit none
	
!     MATRIXES: C=A*B      

	real(8), dimension (350,2,2)::a,b,c
	integer(4) k

	c(k,1,1)=a(k,1,1)*b(k,1,1)+a(k,1,2)*b(k,2,1)
	c(k,1,2)=a(k,1,1)*b(k,1,2)+a(k,1,2)*b(k,2,2)
	c(k,2,1)=a(k,2,1)*b(k,1,1)+a(k,2,2)*b(k,2,1)
	c(k,2,2)=a(k,2,1)*b(k,1,2)+a(k,2,2)*b(k,2,2)

	END SUBROUTINE MATRIXMULT


	SUBROUTINE MATRIXMULTVECTOR(a,g,f,k)
	implicit none
	
!     MATRIX A, VECTORS g, f: Ag=f

	real(8) a(350,2,2),f(350,2),g(350,2)
	integer(4) k

      f(k,1)=a(k,1,1)*g(k,1)+a(k,1,2)*g(k,2)
	f(k,2)=a(k,2,1)*g(k,1)+a(k,2,2)*g(k,2)

      return
      END SUBROUTINE MATRIXMULTVECTOR

	SUBROUTINE VECTORSUM(a,b,c,k)
	implicit none
	
!     VECTORS: C=A+B

      real(8), dimension(350,2)::a,b,c
	integer(4) k
	c(k,1)=a(k,1)+b(k,1)
      c(k,2)=a(k,2)+b(k,2)
	END SUBROUTINE VECTORSUM


	SUBROUTINE INVERSMATRIX(a,a1,k)
	implicit none 
	
!     MATRIXES: A1=A*(-1)

      real(8), dimension(350,2,2)::a,a1
	integer(4) k
	
	a1(k,1,1)=a(k,2,2)/(a(k,1,1)*a(k,2,2)-a(k,1,2)*a(k,2,1)) 
	a1(k,1,2)=-a(k,1,2)/(a(k,1,1)*a(k,2,2)-a(k,1,2)*a(k,2,1))
	a1(k,2,1)=-a(k,2,1)/(a(k,1,1)*a(k,2,2)-a(k,1,2)*a(k,2,1))
	a1(k,2,2)=a(k,1,1)/(a(k,1,1)*a(k,2,2)-a(k,1,2)*a(k,2,1))  

	END SUBROUTINE INVERSMATRIX

	
	SUBROUTINE MATRIXPROGONKA(a,b,c,d,y,N)
	
!     MATRIXPROGONKA solves the set of MATRIX three-point diference equations	
	
	implicit none
	real(8), dimension(350,2,2):: a,b,c,x3,x32,x31,alpha
	real(8), dimension(350,2):: y,d,x2,beta,x21
	integer(4) N,i,j,k

      call INVERSMATRIX(c,x32,1)
	call MATRIXMULT(x32,b,x3,1)
	do i=1,2
	 do j=1,2
	  alpha(2,i,j)=x3(1,i,j)
	 enddo
      enddo
      call MATRIXMULTVECTOR(x32,d,x2,1)
	do i=1,2
	 beta(2,i)=x2(1,i)
      enddo

	do k=3,N
	 CALL MATRIXMULT(A,ALPHA,X3,k-1)
	 CALL MATRIXSUM(C,-X3,X31,k-1)
	 CALL INVERSMATRIX(X31,X32,k-1)
	 CALL MATRIXMULT(X32,B,X3,k-1)
	 do i=1,2
	  do j=1,2
	   alpha(k,i,j)=X3(k-1,i,j)
	  enddo
       enddo
	 !call matrixmult(x3,x31,x33,k-1)
	 !call matrixsum(-b,x33,x3,k-1)
	 CALL MATRIXMULTVECTOR(A,BETA,X2,K-1)
	 CALL VECTORSUM(D,X2,X21,K-1)
	 CALL MATRIXMULTVECTOR(X32,X21,X2,K-1)
	 do i=1,2
	  beta(k,i)=X2(k-1,i)
       enddo
      enddo

      CALL MATRIXMULT(A,ALPHA,X3,N)
	CALL MATRIXSUM(C,-X3,X31,N)
	CALL INVERSMATRIX(X31,X32,N)
	CALL MATRIXMULTVECTOR(A,BETA,X2,N)
	CALL VECTORSUM(D,X2,X21,N)
	CALL MATRIXMULTVECTOR(X32,X21,X2,N)
	do i=1,2
	 Y(N,i)=X2(N,i)
      enddo

	do k=N-1,1,-1
	 CALL MATRIXMULTVECTOR(ALPHA,Y,X2,K+1)
	 CALL VECTORSUM(X2,BETA,X21,K+1)
	 Y(K,1)=X21(K+1,1)
	 Y(K,2)=X21(K+1,2)
      enddo

	return

	END SUBROUTINE MATRIXPROGONKA


      REAL(8) FUNCTION KRON(i,j)
      implicit none
	integer(4) i,j
	kron=0.
	if(i==j) kron=1.
	END FUNCTION 
	
      
!	SUBROUTINE WaterMix(flag, T1, T2, ro1)
!	use driving_params
!	use arrays
!	implicit none
!                 
!      real(8), allocatable:: T1(:), T2(:), ro1(:), 
!     & ro2(:)
!      integer(4) n, i, k, p, flag
!	logical firstcall

!	data firstcall /.true./

!	SAVE

!	if (firstcall) then
!       allocate (T1(1:M+1), T2(1:M+1), ro1(1:M+1), 
!     & ro2(1:M+1))
!	endif
                            
!      do i = 1, M+1
!       ro1(i) = 800.969d-7+588.194d-7*T1(i)- 
!     & 811.465d-8*(T1(i)**2)+476.600d-10*(T1(i)**3)
!       ro2(i) = ro1(i)
!      end do    
                       
!      do i = 1, M+1
!       T2(i) = T1(i)
!      end do        
                   
!	!if (flag==0) then			      
!      ! do n = 1, M-1 
!      !  k = 1
!      !  do i = 1, M+1-n
!      !   if (ro1(i) > ro1(k)) then
!      !    k = i 
!      !   end if
!      !  end do
!      !  T2(M+1-n) = T1(k)
!      !  T2(k) = T1(M+1-n)
!      !  ro2(M+1-n) = ro1(k)
!      !  ro2(k) = ro1(M+1-n) 
!      !  do p = 1, M+1-n
!      !   ro1(p) = ro2(p)
!      !   T1(p) = T2(p)
!      !  end do
!      ! end do
!	!else       
!       do n = 1, M-2 
!        k = 2
!        do i = 2, M+1-n
!         if (ro1(i) > ro1(k)) then
!          k = i 
!         end if
!        end do
!        T2(M+1-n) = T1(k)
!        T2(k) = T1(M+1-n)
!        ro2(M+1-n) = ro1(k)
!        ro2(k) = ro1(M+1-n) 
!        do p = 1, M+1-n
!         ro1(p) = ro2(p)
!         T1(p) = T2(p)
!        end do
!       end do      
!      !endif
	
!      if (flag==0) then
!       ro1(1) = 800.969d-7+588.194d-7*(T1(1)+T1(2))/2.- 
!     & 811.465d-8*((T1(1)+T1(2))**2)/4.+476.600d-10*
!     & ((T1(1)+T1(2))**3)/8.
!       ro2(1) = ro1(1)
!      c1:do i=M,2,-1
!	  if (ro2(1)>ro2(i)) then
!	   T2(i)=(T1(1)+T1(2))/2
!	   ro2(i)=ro1(1)
!	   do k=2,i
!	    T2(k-1)=T1(k)
!	    ro2(k-1)=ro1(k)
!         enddo
!	   T1=T2
!	   ro1=ro2
!	   exit c1
!	  endif
!	 enddo c1    
!      endif

!	if (firstcall) firstcall=.false.                                       
!      END SUBROUTINE WaterMix 

	function fext2(fname,ext)
      character*80 fname,fext2
      character*3 ext
      parameter(n=80)

      fext2=fname

      do 10 i=30,1,-1
         if(fext2(i:i).eq.'.') then
            fext2(i:n)=char(0)
            go to 11
         endif
10    continue
11    continue

      do 20 i=30,1,-1
         if(fext2(i:i).ne.' ' .and. fext2(i:i).ne.char(0)) then
            fext2(i+1:i+4)='.'//ext
            go to 21
         endif
20    continue
      write(*,*) 'erro em fext'

21    continue
      return
      end
                 
      subroutine readgrd_lake(nunit,var,i0,i1,j0,j1)
c
c reads 2d array from a grd file
c
      implicit real*8(a-h,o-z)
      dimension var(i0:i1,j0:j1)
      character*4 dsaa

      read(nunit,'(a4)') dsaa
      read(nunit,*) nx,ny
      read(nunit,*) xmin,xmax
      read(nunit,*) ymin,ymax
      read(nunit,*) zmin,zmax

      
      do j=j0,j1
         read(nunit,*) (var(i,j),i=i0,i1)
	enddo
      
      return
      end
      
      subroutine wrigrd_lake(nunit,z,dx,dy,i0,i1,j0,j1)

      implicit none
	integer i0,i1,j0,j1,nunit,i,j
      real(8) z(i0:i1,j0:j1)
      real(8) xmin,xmax,ymin,ymax,zmin,zmax,dx,dy
      
	
      zmin=z(i0,j0)
      zmax=z(i0,j0)

      do j=j0,j1
         do i=i0,i1
           zmin=min(zmin,z(i,j))
           zmax=max(zmax,z(i,j))
         enddo
      enddo

	xmin=float(i0)*dx
	xmax=float(i1)*dx
	ymin=float(j0)*dy
	ymax=float(j1)*dy
      
      write(nunit,'(a4)') 'DSAA'
      write(nunit,*) (i1-i0+1),(j1-j0+1)
      write(nunit,*) xmin,xmax
      write(nunit,*) ymin,ymax
      write(nunit,*) zmin,zmax

      do j=j0,j1
        write(nunit,*) (z(i,j),i=i0,i1)
      enddo
     
      return
      end


	SUBROUTINE MON_OUT(dt)
      use arrays
      implicit none
      real(8) year,month,day,hour,dt
      real(8) h1,l1,hs1,ls1
      real(8), allocatable:: Tw_m(:),Tw_sum(:),Sal_m(:),Sal_sum(:),
     & TKE_m(:),TKE_sum(:),eps_m(:),eps_sum(:)
	real(8), external:: dzeta
      integer(4) ndm1,i,ndaym(12),month_old
      character month1*2,year1*4,day1*2,hour1*2,outpath*60
      character*6 timestring
      logical firstcall
      common /ymdh/ year,month,day,hour
      common /layers/ h1,l1,hs1,ls1
      common /out/ outpath
      data firstcall /.true./
      data ndaym /31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31/

      SAVE

      if (firstcall) then
       month_old=int4(month)
       ndm1 = int4(day)
       allocate (Tw_m(1:M+1),Tw_sum(1:M+1))
       allocate (Sal_m(1:M+1),Sal_sum(1:M+1))
	 allocate (TKE_m(1:M+1),TKE_sum(1:M+1))
	 allocate (eps_m(1:M+1),eps_sum(1:M+1))
       Tw_sum = 0.
       Sal_sum = 0.
	 TKE_sum = 0.
	 eps_sum = 0.
      endif

      Tw_sum=Tw_sum+Tw1
      Sal_sum=Sal_sum+Sal1
	TKE_sum=TKE_sum+E1
	eps_sum=eps_sum+eps1
      if (month/=float(month_old)) then
       Tw_m=Tw_sum/(((ndaym(month_old)-ndm1+1)*24.*60.*60.)/dt)
       Sal_m=Sal_sum/(((ndaym(month_old)-ndm1+1)*24.*60.*60.)/dt)
	 TKE_m=TKE_sum/(((ndaym(month_old)-ndm1+1)*24.*60.*60.)/dt)
	 eps_m=eps_sum/(((ndaym(month_old)-ndm1+1)*24.*60.*60.)/dt)
	 call dataminus (1,year1,month1,day1,hour1)
	 call timestr(6,year1,month1,day1,hour1,timestring)
       open(600,file=outpath(1:len_trim(outpath))//'monthly/'//
     &  'Profiles'//timestring//'.dat',status='unknown')
	 write (600,*) '1 - depth, m' 
	 write (600,*) '2 - temperature, C'
       write (600,*) '3 - salinity, kg/kg' 
       write (600,*) '4 - turbulent kinetic energy, m**2/s**2'
	 write (600,*) '5 - disspation rate, m**2/s**3'
       do i=1,M+1 
        write (600,'(f6.2,f9.3,3e12.4)') 
     &   dzeta(dble(i))*h1,Tw_m(i),Sal_m(i),TKE_m(i),eps_m(i)
       enddo
       close(600)
       month_old=int4(month)
       ndm1=int4(day) 
       Tw_sum = 0.
       Sal_sum = 0.
	 TKE_sum = 0.
	 eps_sum = 0.
      endif

      if (firstcall) firstcall=.false.
      END SUBROUTINE MON_OUT


      SUBROUTINE DAY_OUT(dt)
      use arrays
      implicit none
      real(8) year,month,day,hour,dt,day_dur
      real(8) h1,l1,hs1,ls1
      real(8), allocatable:: Tw_m(:),Tw_sum(:),Sal_m(:),Sal_sum(:),
     & TKE_m(:),TKE_sum(:),eps_m(:),eps_sum(:)
	real(8), external:: dzeta
      integer(4) i,nday_old
      character month1*2,year1*4,day1*2,hour1*2,outpath*60
      character*8 timestring
      logical firstcall, firstwrite
      common /ymdh/ year,month,day,hour
      common /layers/ h1,l1,hs1,ls1
      common /out/ outpath
      data firstcall, firstwrite, day_dur /.true., .true., 86400./
      
      SAVE

      if (firstcall) then
       nday_old = int4(day)
       allocate (Tw_m(1:M+1),Tw_sum(1:M+1))
       allocate (Sal_m(1:M+1),Sal_sum(1:M+1))
	 allocate (TKE_m(1:M+1),TKE_sum(1:M+1))
	 allocate (eps_m(1:M+1),eps_sum(1:M+1))
       Tw_sum = 0.
       Sal_sum = 0.
	 TKE_sum = 0.
	 eps_sum = 0.
      endif

      Tw_sum=Tw_sum+Tw1
      Sal_sum=Sal_sum+Sal1
	TKE_sum=TKE_sum+E1
	eps_sum=eps_sum+eps1
      if (int4(day)/=nday_old) then
       Tw_m =Tw_sum /(day_dur/dt)
       Sal_m=Sal_sum/(day_dur/dt)
	 TKE_m=TKE_sum/(day_dur/dt)
	 eps_m=eps_sum/(day_dur/dt)
       if (.not. firstwrite) then
        call dataminus (2,year1,month1,day1,hour1)
        call timestr(8,year1,month1,day1,hour1,timestring)
        open(600,file=outpath(1:len_trim(outpath))//'daily/'// 
     &   'Profiles'//timestring//'.dat',status='unknown')
        write (600,*) '1 - depth, m' 
	  write (600,*) '2 - temperature, C'
        write (600,*) '3 - salinity, kg/kg' 
        write (600,*) '4 - turbulent kinetic energy, m**2/s**2'
	  write (600,*) '5 - disspation rate, m**2/s**3'
        do i=1,M+1 
         write (600,'(f6.2,f9.3,3e12.4)') 
     &    dzeta(dble(i))*h1,Tw_m(i),Sal_m(i),TKE_m(i),eps_m(i)
        enddo
        close(600)
       endif 
       nday_old=int4(day) 
       Tw_sum = 0.
       Sal_sum = 0.
	 TKE_sum = 0.
	 eps_sum = 0.
       if (firstwrite) firstwrite=.false.
      endif

      if (firstcall) firstcall=.false.
      END SUBROUTINE DAY_OUT


	SUBROUTINE HOUR_OUT(dt)
      use arrays
      use driving_params
      implicit none
      real(8) year,month,day,hour,dt,hour_dur
      real(8) h1,l1,hs1,ls1
      real(8), allocatable:: Tw_m(:),Tw_sum(:),Sal_m(:),Sal_sum(:),
     & TKE_m(:),TKE_sum(:),eps_m(:),eps_sum(:)
	real(8), external:: dzeta 
      integer(4) i,nhour_old
      character month1*2,year1*4,day1*2,hour1*2,outpath*60
      character*10 timestring
      logical firstcall, firstwrite
      common /ymdh/ year,month,day,hour
      common /layers/ h1,l1,hs1,ls1
      common /out/ outpath
      data firstcall, firstwrite, hour_dur /.true., .true., 3600./
      
      SAVE

      if (firstcall) then
       nhour_old = int4(hour)
       allocate (Tw_m(1:M+1),Tw_sum(1:M+1))
       allocate (Sal_m(1:M+1),Sal_sum(1:M+1))
	 allocate (TKE_m(1:M+1),TKE_sum(1:M+1))
	 allocate (eps_m(1:M+1),eps_sum(1:M+1))
       Tw_sum = 0.
       Sal_sum = 0.
	 TKE_sum = 0.
	 eps_sum = 0.
      endif

      Tw_sum=Tw_sum+Tw1
      Sal_sum=Sal_sum+Sal1
	TKE_sum=TKE_sum+E1
	eps_sum=eps_sum+eps1
      if (int4(hour)/=nhour_old) then
       Tw_m =Tw_sum/ ((hour-nhour_old)*hour_dur/dt)
       Sal_m=Sal_sum/((hour-nhour_old)*hour_dur/dt)
	 TKE_m=TKE_sum/((hour-nhour_old)*hour_dur/dt)
	 eps_m=eps_sum/((hour-nhour_old)*hour_dur/dt)
       if (.not. firstwrite) then
        call dataminus (3,year1,month1,day1,hour1)
        call timestr(10,year1,month1,day1,hour1,timestring)
        open(600, file=outpath(1:len_trim(outpath))//'hourly/'// 
     &   'Profiles'//timestring//'.dat',
     &    status='unknown')
	  write (600,*) '1 - depth, m' 
	  write (600,*) '2 - temperature, C'
        write (600,*) '3 - salinity, kg/kg' 
        write (600,*) '4 - turbulent kinetic energy, m**2/s**2'
	  write (600,*) '5 - disspation rate, m**2/s**3'
        do i=1,M+1 
         write (600,'(f6.2,f9.3,3e12.4)') 
     &    dzeta(dble(i))*h1,Tw_m(i),Sal_m(i),TKE_m(i),eps_m(i)
        enddo
        close(600)
       endif 
       nhour_old=int4(hour) 
       Tw_sum = 0.
       Sal_sum = 0.
	 TKE_sum = 0.
	 eps_sum = 0.
       if (firstwrite) firstwrite=.false.
      endif

      if (firstcall) firstcall=.false.
      END SUBROUTINE HOUR_OUT          
      
      SUBROUTINE EVERYSTEP_OUT
      use arrays
      implicit none
      real(8) h1,l1,hs1,ls1
	real(8), external:: dzeta
      integer(4) i
      character outpath*60
      logical firstcall
      data firstcall /.true./
      common /out/ outpath
      common /layers/ h1,l1,hs1,ls1
      
      SAVE
      
      if (firstcall) then
       open (605,file=outpath(1:len_trim(outpath))//'everystep/'// 
     &  'Profiles.dat',  status='unknown')
       write (605,*) '1 - depth, m' 
	 write (605,*) '2 - temperature, C'
       write (605,*) '3 - salinity, kg/kg' 
       write (605,*) '4 - turbulent kinetic energy, m**2/s**2'
	 write (605,*) '5 - disspation rate, m**2/s**3'
      endif
      
      write (605,*) 'nstep = ', nstep
      do i=1,M+1 
       write (605,'(f6.2,f9.3,3e12.4)') 
     &  dzeta(dble(i))*h1,Tw1(i),Sal1(i),E1(i),eps1(i)
      enddo
      
      if (firstcall) firstcall=.false.
      END SUBROUTINE EVERYSTEP_OUT
      
      
      SUBROUTINE series_out(tsw)
      use driving_params
      use arrays
      use atmos
      implicit none
      real(8) year,month,day,hour,h1,l1,hs1,ls1,T_mean
      real(8), intent(in):: tsw
      integer(4) out_count
      character*60 outpath
      logical firstcall
      data firstcall /.true./
      common /out/ outpath
      common /ymdh/ year,month,day,hour
      common /layers/ h1,l1,hs1,ls1
      
      SAVE
            
      if (firstcall) then
       open (603,file=outpath(1:len_trim(outpath))//'time_series/'// 
     &  'layers.dat',  status='unknown')
	 write (603,*) '1 - year'
       write (603,*) '2 - month'
	 write (603,*) '3 - day'
	 write (603,*) '4 - hour'
	 write (603,*) '5 - water layer thickness, m'
	 write (603,*) '6 - ice layer thickness,   m'
	 write (603,*) '7 - snow layer thickness,  m'
	 write (603,*) '8 - bottom ice thickness,  m'
       open (604,file=outpath(1:len_trim(outpath))//'time_series/'// 
     &  'T_fluxes.dat',status='unknown')
	 write (604,*) '1 - year'
       write (604,*) '2 - month'
	 write (604,*) '3 - day'
	 write (604,*) '4 - hour'
	 write (604,*) '5 - surface temperature,  C'
	 write (604,*) '6 - bottom temperature of water,  C'
	 write (604,*) '7 - mean temperature of water coloumn,  C'
	 write (604,*) '8 - sensible heat flux,   W/m**2'
	 write (604,*) '9 - latent heat flux,     W/m**2'
	 out_count = int(time/(dt_out*60*60))
      endif
      
      if (int(time/(dt_out*60*60))>=out_count) then
       T_mean = sum(Tw1(1:M+1))/float(M+1)
       write (603,'(4f6.1,4f9.4)') year,month,day,hour,h1,l1,hs1,ls1
       write (604,'(4f6.1,5f9.4)') year,month,day,hour,tsw-273.2,
     &                             Tw1(M+1),T_mean,hw,xlew       
       out_count = out_count + 1 
      endif 
      
      if (firstcall) firstcall=.false.
      END SUBROUTINE series_out           
      
      
      SUBROUTINE dataminus(n,year1,month1,day1,hour1)
      use driving_params
      implicit none
      real(8) year,month,day,hour
      integer(4) n,ndaym(12)
      character year1*4,month1*2,day1*2,hour1*2
      common /ymdh/ year,month,day,hour
      data ndaym /31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31/
      
      SAVE
      
      if (n==1) then
!     Minus 1 month      
       if (int4(month)==1) then
        write (month1,'(i2)') 12
        write (year1, '(i4)') int4(year)-1
       else
        write (month1,'(i2)') int4(month)-1
        write (year1, '(i4)') int4(year)
       endif 
       elseif (n==2) then
!     Minus 1 day      
       if (int4(day)==1) then
        if (int4(month)==1) then
         write (month1,'(i2)') 12
         write (year1, '(i4)') int4(year)-1
         write (day1, '(i2)') 31
        else
         write (month1,'(i2)') int4(month)-1
         write (year1, '(i4)') int4(year)
         write (day1, '(i2)') ndaym(int4(month)-1) 
        endif
       else
        write (month1,'(i2)') int4(month)
        write (year1, '(i4)') int4(year)
        write (day1, '(i2)')  int4(day)-1 
       endif
      elseif (n==3) then
!     Minus 1 hour      
       if (int4(hour)==0) then
        if (int4(day)==1) then
         if (int4(month)==1) then
          write (month1,'(i2)') 12
          write (year1, '(i4)') int4(year)-1
          write (day1, '(i2)')  31
          write (hour1, '(i2)') 24-int4(interval)
         else
          write (month1,'(i2)') int4(month)-1
          write (year1, '(i4)') int4(year)
          write (day1, '(i2)')  ndaym(int4(month)-1) 
          write (hour1, '(i2)') 24-int4(interval)
         endif
        else
         write (month1,'(i2)') int4(month)
         write (year1, '(i4)') int4(year)
         write (day1,  '(i2)') int4(day)-1
         write (hour1, '(i2)') 24-int4(interval)
        endif
       else
        write (month1,'(i2)') int4(month)
        write (year1, '(i4)') int4(year)
        write (day1,  '(i2)') int4(day)
        write (hour1, '(i2)') int4(hour)-int4(interval)
       endif
      else
       print*, 'Invalid meaning of n in dataminus: STOP'
       stop
      endif
      
      END SUBROUTINE dataminus
                  
      SUBROUTINE timestr(n,y,m,d,h,timestring)
      implicit none
      integer n,i
      character y*4,m*2,d*2,h*2
      character(len=n) timestring 
      
      if (n==10) then
       write (timestring, '(a4,3a2)') y,m,d,h 
      elseif (n==8) then
       write (timestring, '(a4,2a2)') y,m,d 
      elseif (n==6) then
       write (timestring, '(a4,a2)') y,m 
      endif 
      
      do i = 1, n
       if (timestring(i:i)==' '.or.timestring(i:i)==char(0))
     &  timestring(i:i)='0'
      enddo
      
      END SUBROUTINE timestr


      real(8) function dzeta(nl)
	use arrays
	implicit none
	real(8) nl,nl_int

      nl_int = int4(nl)
	if (nl_int>1) then
	 dzeta = sum(ddz(1:nl_int-1))
	else
       dzeta = 0.
	endif
	if (nl<M+1)                                                               VS,25.06.2007
     & dzeta = dzeta + (nl-nl_int)*ddz(nl_int)

	end function dzeta


