      MODULE phys_parameters
	REAL(8) z_boundary,AZ0,ALB,whc,hcond,SNCR,CCT_veg,TBEST,FFTMIN,
     &FFQMIN,AA_veg,BB_veg,CC_veg,PSI1_veg,PSI2_veg,DTL_veg,ROOTSM_veg,
     &CCF_veg,DZZ,HR,SoilDensity,BH_soil,PSIMAX_soil,POR_soil,
     &FLWMAX_soil,DLMAX_soil,WLM0_soil,WLM7_soil
      INTEGER(4) VegetType
	contains

	SUBROUTINE DEFINE_phys_parameters()
	implicit none
cccccccccccccccccccccc AIR cccccccccccccccccccccccc
		z_boundary = 1000	     ! height of planetary boundary layer (m)
		AZ0 = 0.035				 ! roughness lenght (m)
		ALB = 0.23               ! albedo (1)

cccccccccccccccccccccc SNOW cccccccccccccccccccccccc
		whc = 0.04               ! water holding capacity of snow
		hcond = 0.01              ! hydraulic conductivity in snow (m/s)
		SNCR = 0.01                ! M OF WATER, FROM WHICH SNOW COVERS ALL THE CELL

cccccccccccccccccccccc VEGETATION cccccccccccccccccccccccc
		CCT_veg = 0.0016     ! DEPEND. CANOPY RESIST. ON TEMPERATURE (1/K)
		TBEST = 298.15       ! LOWEST CANOPY RESIST. TEMPERATURE (K)
		FFTMIN = 0.01        ! MINIMAL RESIST. COEFF. FOR TEMPERATURE (1)
		FFQMIN = 0.01        ! MINIMAL RESIST. COEFF. FOR HUMIDITY (1)

		VegetType = 7        ! 1 -- 13
cc  OR:
        AA_veg = 2582.0                    ! FOR LEAF RESISTANCE, J/M**3
        BB_veg = 1.09                      ! FOR LEAF RESISTANCE, W/M**2
        CC_veg = 110.0                     ! MINIMAL LEAF RESIST., S/M
        PSI1_veg = - 120.                ! WATER POT. OF WILT BEGIN, M
        PSI2_veg = - 230.                ! WATER POT. OF PERMANENT WILT, M
        DTL_veg = 3.0                      ! MAX. LAI, DIMENSIONLESS
        ROOTSM_veg = 0.5                   ! MAX. ROOTS LENGTH, M
        CCF_veg = 0.0238*1000./0.622       ! DEPEND. CANOPY RESIST. ON HUMIDITY (1/mb)

cccccccccccccccccccccc SOIL cccccccccccccccccccccc
		DZZ = 5.             ! ACTUAL ST. DEV. OF TOPOGR. FOR RUNOFF CALC. (m)
		HR = 0.08             ! LAYER OF SOIL FROM WHICH EVAPORATION EXISTS (m)
		SoilDensity = 1200.    ! kg/m^3

!		SoilType = 0         ! 1 -- 11
cc  OR:
		BH_soil     = 7.12             ! PARAMETER B, DIMENSIONLESS
		PSIMAX_soil = - 0.0863         ! SAT. WATER POTENTIAL, M
		POR_soil    = 0.401            ! POROSITY, DIMENSIONLESS
		FLWMAX_soil = 0.0000063          ! SAT. HYDR. CONDUCTIVITY, M/S
		FLWMAX_soil = 0.00002            ! SAT. HYDR. CONDUCTIVITY, M/S
		DLMAX_soil  = 0.000003070          ! SAT. WATER DIFFUSIVITY, M**2/S
		WLM0_soil   = 0.13             ! MAXIMAL UNFREEZING WATER AT 0C, kg/kg
		WLM7_soil   = 0.06             ! MAXIMAL UNFREEZING WATER AT T<<0C, kg/kg

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      END SUBROUTINE DEFINE_phys_parameters
      END MODULE phys_parameters


      MODULE phys_constants
	REAL(8) grav,PL,PLI,GPA,SIGSB,APPA,R,CP,RHOW,RHOI,CR,CWAT,CICE,
     &CSNOW,WMG,WLMMX,CEFF,CA,DZL,DZG,BMIN,DMIN,DMAX,D,WIINF,ZRM,
     &ZRMM,TRM,TOMIN,FLXMIN,CK,cpa
      contains

	SUBROUTINE DEFINE_phys_constants()
	implicit none
cccccccccccccccccccccc COMMON cccccccccccccccccccccccc
		grav = 9.81			      ! acceleration due to gravity (m/s**2)
		PL = 600.*4180.  		  ! latent heat for evapor./condens. (cal/g)	
		PLI = 80.*4180.           ! latent heat for freezing/melting (cal/g)

cccccccccccccccccccccc AIR cccccccccccccccccccccccc
		CPA = 1008.                ! air specific heat content (Dg /(kg deg))
		SIGSB = 5.67E-8    ! Stephan-Boltzman const.(Dg/(m**2 sec K**4))
		APPA = 0.286			  ! kappa = c_v/c_p
		R = 287.				  ! universal gas constant
		CP = 1004.

cccccccccccccccccccccc SOIL and SNOW cccccccccccccccccccccccc
		RHOW = 1000.                ! the liquid water density, kg/m**3
		RHOI = 916.8            ! the ice density, kg/m**3
		CR = 1008.               ! the soil specific heat content, Dg/(kg K)
		CWAT = 4200.             ! the water specific heat content, Dg/(kg K)
		CICE = 2100.               ! the ice specific heat content, Dg/(kg K)
		CSNOW = 2100.              ! the snow specific heat content, Dg/(kg K)
		WMG = 168.

cccccccccccccccccccccc HYDROLOGY cccccccccccccccccccccc
		WLMMX = 5.E-4        ! M. OF WATER. MINIMAL SKIN LAYER CAPACITY
		CEFF = 0.5           ! 1 FRACTION OF INTERCEPTED BY SKIN LAYER PRECIP.
		CA = 1.              ! 1 FRACTION OF THE CELL, COVERED BY PRECIP.
		DZL = 100.           ! M. MINIMAL ST. DEV. OF TOPOGR. FOR RUNOFF CALC.
		DZG = 1000.          ! M. MAXIMAL ST. DEV. OF TOPOGR. FOR RUNOFF CALC.
		BMIN = 0.01          ! 1 MINIMAL B FOR RUNOFF CALC.
		DMIN = 2.8E-10        ! M/S. COEFF. FOR SLOW DRAINAGE
		DMAX = 2.8E-8        ! M/S. COEFF. FOR FAST DRAINAGE
		D = 1.5              ! 1 COEFF. FOR FAST DRAINAGE
		WIINF = 0.30         ! 1 SOIL ICE, FROM WHICH INFILTRATION BEGIN
		ZRM = 0.1           ! M. LAYER OF SOIL WITH LARGE AMOUNT OF ROOTS
		ZRMM = 0.2           ! M. LAYER OF SOIL FOR CALCULATION OF LAI AND VEG
		TRM = 5.0            ! C. CRITICAL SOIL TEMPERATURE FOR LAI AND VEG
		TOMIN = 0.1          ! C. CRITICAL TEMPERATURE FOR WATER ACCESS FOR CANOPY
		FLXMIN = 1.05        ! CAL/(M**2*S) ACCURACY OF BALANCE AT THE SURFACE
		CK = 0.9             ! 1 COEFF. FOR CANOPY RESIST.
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      END SUBROUTINE DEFINE_phys_constants
	END MODULE phys_constants
 
      MODULE phys_constants2
      real(8) g,row0,roi,cw,ci,Lwi,Lwv,Liv,kappa,
     & roughnessofwater,albedoofwater,emissivityofwater, 
     & sigma,cp,aMag,bMag,Rd,gammaadiabat,lami,lamw0,
     & albedoofice,albedoofsnow,emissivityofice, sabs,
     & emissivityofsnow,aMagw,aMagi,bMagw, bMagi, gamst,
     & emissivityofsoil,albedoofsoil,alsal
      
	contains
	SUBROUTINE DEFINE_phys_constants2 
	implicit none
      g = 9.814
      row0 = 1000.
      roi = 917.   
      cw = 3990.   
      ci = 2150. 
      lami = 2.2   
      lamw0 = 0.561 
      Lwi = 267900.
      Lwv = 2501000.
      Liv = 2834600.
      kappa = 0.38
!      roughnessofwater = 0.00001 
      albedoofwater = 0.08   
      albedoofice = 0.35
      albedoofsnow = 0.62  !0.75
      emissivityofice = 0.95 
      emissivityofwater = 0.95 
      emissivityofsnow = 0.99
      sigma = 5.67d-8
      cp = 1005. ! heat capacity of air at constant pressure
      Rd = 287.05 ! gaseous constant for air
      aMagw = 7.6326    ! coefficient for Magnus formula for water surface
      bMagw = 241.9    ! coefficient for Magnus formula for water surface
      aMagi = 9.5  ! coefficient for Magnus formula for ice surface
      bMagi =	265.5 ! coefficient for Magnus formula for ice surface
      gammaadiabat = g/cp
      gamst = 0.006
	sabs = 0.4 ! the part of solar radiation, absorbed by thin layer at the water surface
	emissivityofsoil = 0.9
	albedoofsoil = 0.3 
	alsal = 1.
      END SUBROUTINE DEFINE_phys_constants2
	END MODULE phys_constants2

	MODULE numeric_params
	 integer(4), PARAMETER:: KL = 1,ML = 131,MS = 100,NT = 52592,
     & Num_Soil = 11,Num_Veget = 13 
       real(8), parameter:: dznorm = 0.05, dzmin = 0.01, 
     & UpperLayer = 0.008 

c		ML --- total number of layers
c		MS --- maximal number of layers in snow
c		dznorm --- standard layer depth in snow (m)
c		dzmin --- min layer depth in snow (m)
c		depth --- depth of the lowest layer in soil (m)
c		UpperLayer --- upper soil layer depth (m)

      END MODULE numeric_params
	
	MODULE atmos
	real(8) tempair,relhum,humair,pressure,hw,xlew,cdmw,trib_inflow, 
     & uwind,vwind,wind,longwave,Radbal,hflux,surfrad,hbal,zref,botflux,
     & cdmw2
      END MODULE atmos

	MODULE driving_params
	
c Switches: 1. PBL parameterization
c              PBLpar =1 (Businger-Dayer formulas (Monin-Obukhov theory) for exchange coefficients)
c              PBLpar =11(Businger-Dayer formulas with shallow water correction after Panin(19..))      
c              PBLpar =2 (formulation from NH3d)
c              PBLpar =21(formulation from NH3d with shallow water correction after Panin(19..))
c           2. Relative to water currents wind
c              relwind =1 (relative wind is off)
c              relwind =2 (relative wind is on)
c       	3. Turbulent mixing parameterization
c              Turbpar =1 (analytyical profile of turbulent conductivity)
c	         Turbpar =2 ("E-eps" parameterization of turbulent conductivity)        
c              Turbpar =3 : Nickuradze (NICK) formulation: Rodi (1993)    
c	         Turbpar =4 : Parabolic (PARAB) formulation: Engelund (1976)
c              Turbpar =7 : RNG (re-normalization group) formulation: Simoes (1998)     
      real(8) l10,h10,ls10,hs10,Ts0,Tb0,Tbb0,h_ML0,depth,Sals0,Salb0,
     & Tm,kwe,kor
	real(8) trib_in,tinteg,dtdr,height_T_q,height_u,lam_g,phi_g,
     &	interval,dt_out
	real(8) extwat,alphax,alphay,a_veg,c_veg,h_veg
      integer(4) runmode,PBLpar,relwind,Turbpar,varalb,turb_out,
     & M,ns,spinup_day,form,rad,nscreen,SoilType,init_T,
     & monthly_out,daily_out,hourly_out,time_series,everystep
      integer(4) assim,as_window,error_cov
	character*40 path
	character*60 dataname,setupfile
	
	contains
	SUBROUTINE DEFINE_driving_params()
	
!     DEFINE_driving_params reads files with driving parameters

	implicit none
      character*200 line
      logical firstcall
      data firstcall, line/.true., 'begin file'/
      
      if (firstcall) then            
       open  (96,file='setup_file.dat',status='old')
       read  (96,'(a)') line
       read  (96,'(a)') setupfile
       close (96)
	 open  (99,file=setupfile,status='old')
       do while (line /= 'end')
	  read (99,'(a)') line
	  call readpar(line)
	 enddo
	 close (99)
 	 firstcall = .false. 
	endif

	END SUBROUTINE DEFINE_driving_params

	subroutine readpar(line)
	implicit none
	character*200 line
	integer i,n2,n1
	SAVE

      i=1
      if (line(1:1)==' '.or.line(1:1)==char(1)) then
       do while (line(i:i)==' '.or.line(i:i)==char(1))
	  i=i+1
	 enddo
	 n1=i
       do while (line(i:i)/=' '.and.line(i:i)/=char(1))
	  i=i+1
	 enddo
	 n2=i-1
	else
	 n1=1
	 do while (line(i:i)/=' '.and.line(i:i)/=char(1))
	  i=i+1
	 enddo
	 n2=i-1
	endif
      
	if (line(n1:n1)/='#') then
	 if (line(n1:n2)=='PBLpar') then
        read (line((n2+1):100),*) PBLpar
	  print*, 'PBLpar=', PBLpar
	 elseif (line(n1:n2)=='runmode') then
	  read (line((n2+1):100),*) runmode
        print*, 'runmode=', runmode
	 elseif (line(n1:n2)=='relwind') then
	  read (line((n2+1):100),*) relwind
	  print*, 'relwind=', relwind
       elseif (line(n1:n2)=='Turbpar') then
	  read (line((n2+1):100),*) Turbpar
	  print*, 'Turbpar=', Turbpar
	 elseif (line(n1:n2)=='varalb') then
	  read (line((n2+1):100),*) varalb
	  print*, 'varalb=', varalb
	 elseif (line(n1:n2)=='soiltype') then
	  read (line((n2+1):100),*) SoilType
	  print*, 'SoilType=', SoilType 
       elseif (line(n1:n2)=='path') then
	  read (line((n2+1):100),*) path
        print*, 'path=', path
	 elseif (line(n1:n2)=='M') then
	  read (line((n2+1):100),*) M
	  print*, 'M=', M
	 elseif (line(n1:n2)=='ns') then
	  read (line((n2+1):100),*) ns
	  print*, 'ns=', ns
       elseif (line(n1:n2)=='h10') then
	  read (line((n2+1):100),*) h10
	  write(*,'(a, f10.1, 1x, a)') 'h10=', h10, 'm'
       elseif (line(n1:n2)=='l10') then
	  read (line((n2+1):100),*) l10
	  write(*,'(a, f10.1, 1x, a)') 'l10=', l10, 'm'
       elseif (line(n1:n2)=='ls10') then
	  read (line((n2+1):100),*) ls10
	  write(*,'(a, f10.1, 1x, a)') 'ls10=', ls10, 'm' 
       elseif (line(n1:n2)=='hs10') then
	  read (line((n2+1):100),*) hs10
	  write(*,'(a, f10.1, 1x, a)') 'hs10=', hs10, 'm'
       elseif (line(n1:n2)=='depth') then
	  read (line((n2+1):100),*) depth
	  write(*,'(a, f10.1, 1x, a)') 'Soil thickness=', depth, 'm'
       elseif (line(n1:n2)=='Ts0') then
	  read (line((n2+1):100),*) Ts0
	  write(*,'(a, f10.1, 1x, a)') 'Ts0=', Ts0, 'C'
       elseif (line(n1:n2)=='Tb0') then
	  read (line((n2+1):100),*) Tb0
	  write(*,'(a, f10.1, 1x, a)') 'Tb0=', Tb0, 'C'
       elseif (line(n1:n2)=='Tbb0') then
	  read (line((n2+1):100),*) Tbb0
	  write(*,'(a, f10.1, 1x, a)') 'Tbb0=', Tbb0, 'C'
       elseif (line(n1:n2)=='h_ML0') then
	  read (line((n2+1):100),*) h_ML0
	  write(*,'(a, f10.1, 1x, a)') 'h_ML0=', h_ML0, 'm'
	 elseif (line(n1:n2)=='Tm') then
	  read (line((n2+1):100),*) Tm
	  write(*,'(a, f10.1, 1x, a)') 'Tm = ', Tm, 'C'
	 elseif (line(n1:n2)=='init_T') then
	  read (line((n2+1):100),*) init_T
	  write(*,'(a, i10)') 'init_T = ', init_T
       elseif (line(n1:n2)=='turb_out') then
	  read (line((n2+1):100),*) turb_out
	  print*, 'turb_out=', turb_out
       elseif (line(n1:n2)=='Sals0') then
	  read (line((n2+1):100),*) Sals0
	  write(*,'(a, f12.5, 1x, a)') 'Sals0 = ', Sals0, 'kg/kg'
	 elseif (line(n1:n2)=='Salb0') then
	  read (line((n2+1):100),*) Salb0
	  write(*,'(a, f12.5, 1x, a)') 'Salb0 = ', Salb0, 'kg/kg'
	 elseif (line(n1:n2)=='trib_inflow') then
	  read (line((n2+1):100),*) trib_in
	  write(*,'(a, f10.1, 1x, a)') 'trib_inflow = ',trib_in,'m/yr'
	 elseif (line(n1:n2)=='tinteg') then
	  read (line((n2+1):100),*) tinteg
	  write(*,'(a, f10.1, 1x, a)') 'tinteg = ', tinteg, 'days'
	 elseif (line(n1:n2)=='spinup_day') then
	  read (line((n2+1):100),*) spinup_day
	  write(*,'(a, i10, 1x, a)') 'spinup_day = ', spinup_day,'days'
	 elseif (line(n1:n2)=='dt') then
	  read (line((n2+1):100),*) dtdr
	  write(*,'(a, f10.1, 1x, a)') 'dt = ', dtdr, 'sec'
	 elseif (line(n1:n2)=='form') then
	  read (line((n2+1):100),*) form
	  print*, 'form = ', form 
	 elseif (line(n1:n2)=='dataname') then
	  read (line((n2+1):100),*) dataname
	  print*, 'dataname = ', dataname 
	 elseif (line(n1:n2)=='phi') then
	  read (line((n2+1):100),*) phi_g
	  write(*,'(a, f10.1, 1x, a)') 'phi = ', phi_g, 'deg'  
	 elseif (line(n1:n2)=='lam') then
	  read (line((n2+1):100),*) lam_g
	  write(*,'(a, f10.1, 1x, a)') 'lam = ', lam_g, 'deg'
	 elseif (line(n1:n2)=='height_T_q') then
	  read (line((n2+1):100),*) height_T_q
	  write(*,'(a, f10.1, 1x, a)') 'height_T_q = ',height_T_q,'m'
	 elseif (line(n1:n2)=='height_u') then
	  read (line((n2+1):100),*) height_u
	  write(*,'(a, f10.1, 1x, a)') 'height_u = ', height_u, 'm'     
	 elseif (line(n1:n2)=='interval') then
	  read (line((n2+1):100),*) interval
	  write(*,'(a, f10.1, 1x, a)') 'interval = ',interval,'hours'    
	 elseif (line(n1:n2)=='rad') then
	  read (line((n2+1):100),*) rad
	  print*, 'rad = ', rad     
	 elseif (line(n1:n2)=='monthly') then
	  read (line((n2+1):100),*) monthly_out
	  print*, 'Monthly output = ', monthly_out      
	 elseif (line(n1:n2)=='daily') then
	  read (line((n2+1):100),*) daily_out
	  print*, 'Daily output = ', daily_out      
	 elseif (line(n1:n2)=='hourly') then
	  read (line((n2+1):100),*) hourly_out
	  print*, 'Hourly output = ', hourly_out
	 elseif (line(n1:n2)=='everystep') then
	  read (line((n2+1):100),*) everystep
	  print*, 'Everystep output = ', everystep
	 elseif (line(n1:n2)=='dt_out') then
	  read (line((n2+1):100),*) dt_out
	  write(*,'(a, f10.1, 1x, a)') 
     &   'Interval for output time series = ', dt_out, 'hours'         
       elseif (line(n1:n2)=='time_series') then
	  read (line((n2+1):100),*) time_series
	  print*, 'Output of time series', time_series         
       elseif (line(n1:n2)=='extwat') then
	  read (line((n2+1):100),*) extwat
	  write(*,'(a,f6.3)'), 
     &   'Extinction of solar radiation in water', extwat, ' m**(-1)'
	 elseif (line(n1:n2)=='alphax') then
	  read (line((n2+1):100),*) alphax
	  write(*,'(a,f12.9,a)'), 
     &   'The slope angle in the x-direction', alphax, ' deg' 
	 elseif (line(n1:n2)=='alphay') then
	  read (line((n2+1):100),*) alphay
	  write(*,'(a,f12.9,a)'), 
     &   'The slope angle in the y-direction', alphay, ' deg' 
	 elseif (line(n1:n2)=='c_veg') then
	  read (line((n2+1):100),*) c_veg
	  write(*,'(a,f12.3)'), 
     &   'The friction coefficient of vegetation', c_veg  
	 elseif (line(n1:n2)=='a_veg') then
	  read (line((n2+1):100),*) a_veg
	  write(*,'(a,f12.3)'), 
     &   'a_veg =  ', c_veg  
	 elseif (line(n1:n2)=='h_veg') then
	  read (line((n2+1):100),*) h_veg
	  write(*,'(a,f6.3)'), 
     &   'The height of vegetation in the water body ', h_veg   
       elseif (line(n1:n2)=='kwe') then
	  read (line((n2+1):100),*) kwe
	  write(*,'(a,f6.3)'), 'The parameter of turbulence
     & enhancement by wave breaking is', kwe
       elseif (line(n1:n2)=='kor') then
	  read (line((n2+1):100),*) kor
	  write(*,'(a,f6.3)'), 'The Coriolis parameter is', kor,
     &   ' s**(-1)'
	 elseif (line(n1:n2)=='nscreen') then
	  read (line((n2+1):100),*) nscreen
	  print*, 'Screen output is every ', nscreen, ' timesteps' 
	 elseif (line(n1:n2)=='assim') then
	  read (line((n2+1):100),*) assim
	  print*, 'assim = ', assim  
	 elseif (line(n1:n2)=='as_window') then
	  read (line((n2+1):100),*) as_window
	  print*, 'as_window = ', as_window
	 elseif (line(n1:n2)=='error_cov') then
	  read (line((n2+1):100),*) error_cov
	  print*, 'error_cov = ', error_cov 
	 endif
	endif

	end subroutine readpar

	END MODULE driving_params



	MODULE ARRAYS
	
!     MODULE ARRAYS contains allocatable arrays of the model	

      use driving_params

      integer(4) snow,ice,water,deepice,nstep 
	real(8) time,tempairf,relhumf,humairf,pressuref,
     & windf,shortwavef,longwavef,precipf,time_old,tempair_old, 
     & relhum_old,humair_old,pressure_old,wind_old, 
     & shortwave_old,longwave_old,precip_old,Erad,
     & wdir_old,wdirf,dep_av
	integer(4) time_int,time_int_old,man,klmn,par,flag_snow_init
	real(8) totalevaps,totalmelts,totalprecips
	real(8) dhwfsoil
	real(8) eflux,Elatent,tau,velfrict,velfrict_prev,cdm1,dhw,
     & dhw0,dhi,dhi0,dls,dls0,extice
	real(8), allocatable:: ddz(:) 
	real(8), allocatable:: Tw1(:),Tw2(:),Ti1(:),Ti2(:),Tis1(:),
     & Tis2(:),RS(:),SR(:),lamw(:),lamsal(:),Sal1(:),Sal2(:),
     & Sals1(:),Sals2(:)
	real(8), allocatable:: Tsoil1(:),rosoil(:),csoil(:), 
     & lamsoil(:), dzs(:), dzss(:), wi1(:), wl1(:), Tsoil2(:),
     & zsoil(:), wsoil(:) 
	real(8), allocatable:: S(:),Gen(:),F(:)
	real(8), allocatable:: KT(:),k2(:),u1(:),v1(:),E1(:),eps1(:) 
	real(8), allocatable:: WLM0(:),WLM7(:),bH(:),PSIMAX(:),POR(:),
     & FLWMAX(:),DLMAX(:)
      real(8), allocatable:: KC(:),KLengT(:),RSR(:) 
      real(8), allocatable::l1_2d(:,:),h1_2d(:,:),ls1_2d(:,:),
     & hs1_2d(:,:),time_2d(:,:) 
      real(kind(0.d0)), allocatable:: dep_2d(:,:),cdm2(:,:)
      real(8), allocatable::u_2d(:,:,:),v_2d(:,:,:),Tsoil1_2d(:,:,:),
     & wi1_2d(:,:,:),wl1_2d(:,:,:),Tw1_2d(:,:,:),Ti1_2d(:,:,:),
     & Tis1_2d(:,:,:),dz_2d(:,:,:),T_2d(:,:,:),wl_2d(:,:,:),
     & dens_2d(:,:,:),E_2d(:,:,:),eps_2d(:,:,:),dhwfsoil_2d(:,:),
     & Sal1_2d(:,:,:),Sals1_2d(:,:,:),Elatent_2d(:,:),dhw_2d(:,:),
     & dhw0_2d(:,:),velfrict_2d(:,:)
      integer(4), allocatable::fl_sn_2d(:,:),fl_sn_init_2d(:,:),
     & itop_2d(:,:), init(:,:), num(:)
      real(8), allocatable:: Tsoil4(:), Tsoil3(:), 
     & wl2(:), wl3(:), wl4(:), wl5(:),  
     & wi2(:), wi3(:), lammoist(:), rosdry(:), filtr(:)
      real(8), allocatable::E2(:),eps2(:),eps3(:),
     & u2(:),v2(:),k1(:),k3(:),k4(:),k5(:),u3(:),
     & v3(:),C1(:),Re(:),row(:),Ri(:),uv(:),E_it1(:),
     & E_it2(:),C_num(:),E_it3(:),Feps(:),E_it21(:),
     & eps_it21(:),eps_it1(:),eps_it2(:),
     & l(:),k2_mid(:),k3_mid(:),k4_mid(:),k5_mid(:),
     & Um(:),Vm(:),E12(:),eps12(:),knum(:),ceps3(:),k2t(:),
     & dzeta1(:),Eeps(:),Ecent(:),Epscent(:),res_E(:),
     & res_eps(:),dresd(:,:,:,:),dres2dE(:),dres2deps(:),
     & veg_sw(:) 
	real(8), allocatable:: WU_(:),WV_(:),GAMT(:),GAMU(:),GAMV(:),
     & TF(:),KCu(:),KLengu(:)
!      real(8), allocatable:: AL(:),DLT(:),DVT(:),DL(:),ALV(:),
!     & DV(:),Z(:),T(:),WL(:),WV(:),WI(:),dens(:)
      data nstep /0/

      contains
	SUBROUTINE allocarrays(nx,ny)
	use numeric_params
      implicit none
      integer nx, ny
      SAVE
	  
      allocate (Tw1(1:M+1),Tw2(1:M+1),Ti1(1:M+1),Ti2(1:M+1),
     & Tis1(1:M+1),Tis2(1:M+1),RS(1:M+1),SR(1:M+1),lamw(1:M+1),
     & Sal1(1:M+1),Sal2(1:M+1),lamsal(1:M+1))
	allocate (Tsoil1(1:ns+2), rosoil(1:ns+2), csoil(1:ns+2), 
     & lamsoil(1:ns+2), dzs(1:ns+2), dzss(1:ns+2), wi1(1:ns+2), 
     & wl1(1:ns+2), Tsoil2(1:ns+2), zsoil(1:ns+2)) 
	allocate (S(1:M+1),Gen(1:M+1),F(1:M+1))
	allocate (KT(1:M+1),k2(1:M+1),u1(1:M+1),v1(1:M+1),E1(1:M+1),
     & eps1(1:M+1))
	allocate (WLM0(1:ns+2),WLM7(1:ns+2),bH(1:ns+2),PSIMAX(1:ns+2),
     & POR(1:ns+2),FLWMAX(1:ns+2),DLMAX(1:ns+2))
      allocate (Tsoil4(1:ns+2), Tsoil3(1:ns+2), 
     & wl2(1:ns+2), wl3(1:ns+2), wl4(1:ns+2), wl5(1:ns+2),  
     & wi2(1:ns+2), wi3(1:ns+2), lammoist(1:ns+2), rosdry(1:ns+2),
     & filtr(1:ns+2), wsoil(1:ns), Sals1(1:ns), Sals2(1:ns) )
      allocate (WU_(1:M+1),WV_(1:M+1),GAMT(1:M+1),GAMU(1:M+1),
     &  GAMV(1:M+1),TF(1:M+1),KCu(1:M+1),KLengu(1:M+1)) 
	allocate (num(1:M+1)) 
	allocate ( E2(1:M+1),eps2(1:M+1),eps3(1:M+1),
     & u2(1:M+1),v2(1:M+1),k1(1:M+1),k3(1:M+1),k4(1:M+1),k5(1:M+1),
     & u3(1:M+1),v3(1:M+1),C1(1:M+1),Re(1:M+1),row(1:M+1),Ri(1:M+1),
     & uv(1:M+1),E_it1(1:M+1),dres2dE(1:M+1),dres2deps(1:M+1),  
     & E_it2(1:M+1),C_num(1:M+1),E_it3(1:M+1),Feps(1:M+1),
     & E_it21(1:M+1),eps_it21(1:M+1),eps_it1(1:M+1),
     & eps_it2(1:M+1),l(1:M+1),k2_mid(1:M+1),k3_mid(1:M+1),
     & k4_mid(1:M+1),k5_mid(1:M+1),Um(1:M+1),Vm(1:M+1),E12(1:M+1),
     & eps12(1:M+1),knum(1:M+1),ceps3(1:M+1),k2t(1:M+1),
     & veg_sw(1:M+1) )
	allocate ( dzeta1(1:M+1),Eeps(1:M+1),Ecent(1:M+1),
     & Epscent(1:M+1),res_E(1:M+1),res_eps(1:M+1),
     & dresd(2,2,1:M+1,1:M+1) ) 
      allocate (KC(1:M+1),KLengT(1:M+1),RSR(1:M+1),ddz(1:M),
     & l1_2d(nx,ny), h1_2d(nx,ny), ls1_2d(nx,ny),
     & dep_2d(0:nx-1,0:ny-1), cdm2(0:nx,0:ny))
	allocate (u_2d(M+1,nx,ny),v_2d(M+1,nx,ny),E_2d(M+1,nx,ny),
     & eps_2d(M+1,nx,ny),Tsoil1_2d(ns+1,nx,ny),wi1_2d(ns+1,nx,ny),
     & wl1_2d(ns+1,nx,ny),Tw1_2d(M+1,nx,ny),Sal1_2d(M+1,nx,ny),
     & Ti1_2d(M+1,nx,ny), Tis1_2d(M+1,nx,ny),Sals1_2d(ns,nx,ny),
     & hs1_2d(nx,ny),dz_2d(ms,nx,ny),T_2d(ml,nx,ny),wl_2d(ml,nx,ny),
     & dens_2d(ms,nx,ny),time_2d(nx,ny),dhwfsoil_2d(nx,ny),
     & Elatent_2d(nx,ny), dhw_2d(nx,ny), dhw0_2d(nx,ny),
     & velfrict_2d(nx,nx))
       allocate (fl_sn_2d(nx,ny),
     & fl_sn_init_2d(nx,ny), init(nx,ny), itop_2d(nx,ny))

	Tw1 = 0.;Tw2 = 0.;Ti1 = 0.;Ti2 = 0.
      Tis1 = 0.;Tis2 = 0.;RS = 0.;SR = 0.;lamw = 0.
      Sal1 = 0.; Sal2 = 0.; lamsal = 0.
	Tsoil1 = 0.; rosoil = 0.; csoil = 0. 
      lamsoil = 0.; dzs = 0.; dzss = 0.; wi1 = 0.; 
      wl1 = 0.; Tsoil2 = 0.; 
	S = 0.; Gen = 0; F = 0.
	KT = 0.; k2 = 0.; u1 = 0.; v1 = 0.; E1 = 0
      eps1 = 0.
	WLM0 = 0.; WLM7 = 0.; bH = 0.; PSIMAX = 0.
      POR = 0.; FLWMAX = 0.; DLMAX = 0.;
      KC = 0.; KLengT = 0.; RSR = 0.  
      dhw = 0; dhw0 = 0; dhi = 0; dhi0 = 0;
      dls = 0; dls0 = 0
      init=0

      ddz = 1./float(M)
      velfrict_prev = 1.d-2

	END SUBROUTINE allocarrays

	END MODULE ARRAYS
	
	
	MODULE assim_var
      use driving_params
      integer n_modvar, n_obsvar, size_mod, size_obs, ntim
      
      real(8), allocatable:: y_obs(:,:), x_mod(:,:), b_mod(:,:), 
     & r_obs(:,:), h_obs(:,:), covar(:,:,:,:), c_diff(:,:,:),
     & RpHBHT(:,:), dep_obs(:,:),y_obs_mean(:),x_mod_mean(:),
     & delta(:)
     
      contains
      subroutine alloc_assim_var
      n_modvar = 1
      n_obsvar = 1
      size_mod = (M+1)*n_modvar
      size_obs = (M+1)*n_obsvar
      
      allocate (y_obs(1:ntim, 1:size_obs),
     &          x_mod(1:ntim, 1:size_mod),
     &          dep_obs(1:ntim, 1:size_obs))
      allocate (r_obs(1:size_obs, 1:size_obs),
     &          RpHBHT(1:size_obs,1:size_obs), 
     &          b_mod(1:size_mod, 1:size_mod), 
     &          h_obs(1:size_obs, 1:size_mod))
      allocate (c_diff(1:M+1,1:n_obsvar,1:n_obsvar))
      allocate (covar(1:M+1, 1:M+1, 1:n_obsvar, 1:n_obsvar))
      allocate (y_obs_mean(1:size_obs),
     &          x_mod_mean(1:size_mod),
     &          delta     (1:size_obs))
      
      y_obs=0.
      x_mod=0.
      r_obs=0.
      b_mod=0. 
      h_obs=0.
      c_diff=0.
      covar=0.
      end subroutine alloc_assim_var
      
      END MODULE assim_var